module endf6
!
! FORTRAN module endf6 supplies ENDF-6 format support
!
! ENDF-6 Glosary:
! --------------
! mat    = Material number (integer)
! mf     = File number (integer)
! mt     = Reaction number (Integer)
! ns     = Sequence number (integer)
! text   = 66 characters of information (charcter*66)
! c1,c2  = real parameters (real*8)
! l1,l2  = Integer test (trigger) numbers (integer)
! n1,n2  = Integer count numbers (integer)
! b      = list of real numbers (real*8 array)
! nbt,ibt= Interpolation table (integer arrays)
! x,y    = Tabulated data (real*8 arrays)
! ii,jj  = position locators (integers)
! kij    = integer array (correlation matrix representation)
! ndigit = Number of decimal digits used (2,3,4,5,6)
! nrow   = dimension of array kij (number of columns) as a function of ndigit
!          (nrow(2)=18,nrow(3)=13,nrow(4)=11,nrow(5)=9,nrow(6)=8)
! Z      = Z number: number of protons
! N      = N number: number of neutrons
! A=Z+N  = Mass number: Number of nucleons
! za     = za number of the target: za=1000*Z+A (real*8)
! zap    = zap number of the product: zap=1000*Zp+Ap (real*8)
! awr    = Atomic weight ratio of the target material to neutron mass (real*8)
! awi    = Atomic weight ratio of the incident particle (real*8)
! awp    = Atomic weight ratio of the product (real*8)
!
! ENDF-6 basic types of records:
! -----------------------------
! TEXT record = [mat,mf,mt/text]
! CONT record = [mat,mf,mt/c1,c2,l1,l2,n1,n2]
! LIST record = [mat,mf,mt/c1,c2,l1,l2,n1,n2/
!                          (b(i),i=1,n1)]
! TAB1 record = [mat,mf,mt/c1,c2,l1,l2,n1,n2/
!                          (nbt(k),ibt(k),k=1,n1)/
!                          (x(i),y(i),i=1,n2)]
! TAB2 record = [mat,mf,mt/c1,c2,l1,l2,n1,n2/
!                          (nbt(k),ibt(k),k=1,n1)]
! INTG record = [mat,mf,mt/ii,jj,(kij(k),k=1,nrow(ndigit))]
!
!
! Character arrays containing ENDF-formatted data:
! -----------------------------------------------
! nin(*): Contain input  ENDF-6 formatted data
! nou(*): Contain output ENDF-6 formatted data
!
! jin   : pointer to the current record of nin
! jou   : pointer to the current record of nou
!
! Maximum dimension of ENDF-6 global parameters in the module endf6.f90:
! ---------------------------------------------------------------------
!
 integer,parameter::nrmax=20                   ! Maximum number of TAB1 and TAB2 interpolation ranges
 integer,parameter::nlmax=64                   ! Highest order Legrendre polynomial (MF4,MF6)
 integer,parameter::ncmax=nlmax+1              ! Number of Legendre coefficients (MF4,MF6)
 integer,parameter::ne3max=500000              ! Maximum number of incident energy ranges (MF3)
 integer,parameter::ne4max=2000                ! Maximum number of incident energies (MF4)
 integer,parameter::nmu4max=2501               ! Maximum number of cosines in MF4 (LTT=2,3)
 integer,parameter::ne5max=2000                ! Maximum number of incident energies (MF5)
 integer,parameter::nep5max=3000               ! Maximum number of outgoing energies (MF5)
 integer,parameter::nkmax=2000                 ! Maximum number of subsections (MF6)
 integer,parameter::ne6max=2000                ! Maximum number of incident energies (MF6)
 integer,parameter::nmu6max=1501               ! Maximum number of angular parameters in MF6 (LAW=1,2,5,7)
 integer,parameter::nep6max=2500               ! Maximum number of secondary energies in MF6
 integer,parameter::npmax=175000000            ! Maximum number of elements in an array
!
 contains
!
! MF0.1: Procedures for reading and writing ENDF-6 records:
! --------------------------------------------------------
!
! =================================================================================================================================
! subroutine readtext: Read an ENDF-6 TEXT record
! =================================================================================================================================
  subroutine readtext(nin,jin,text,mat,mf,mt,ns)
    character*66 text
    character*80 nin(*)
    jin=jin+1
    read(nin(jin),'(a66,i4,i2,i3,i5)')text,mat,mf,mt,ns
    return
  end subroutine readtext
!
! =================================================================================================================================
! subroutine readcont: Read an ENDF-6 CONT record
! =================================================================================================================================
!
  subroutine readcont(nin,jin,c1,c2,l1,l2,n1,n2,mat,mf,mt,ns)
    implicit real*8 (a-h, o-z)
    character*80 nin(*)
    jin=jin+1
    read(nin(jin),'(2e11.0,4i11,i4,i2,i3,i5)')c1,c2,l1,l2,n1,n2,mat,mf,mt,ns
    return
  end subroutine readcont
!
! =================================================================================================================================
! subroutine readlist: Read an ENDF-6 LIST record
! =================================================================================================================================
!
  subroutine readlist(nin,jin,c1,c2,l1,l2,npl,n2,b)
    implicit real*8 (a-h, o-z)
    character*80 nin(*)
    dimension b(*)
    jin=jin+1
    read(nin(jin),'(2e11.0,4i11,i4,i2,i3,i5)')c1,c2,l1,l2,npl,n2,mat,mf,mt,ns
    nline=npl/6
    iline=npl-6*nline
    kl=0
    if (nline.gt.0) then
      do i=1,nline
        k0=kl+1
        kl=kl+6
        jin=jin+1
        read(nin(jin),'(6e11.0)')(b(k),k=k0,kl)
      enddo
    endif
    if (iline.gt.0) then
      k0=kl+1
      jin=jin+1
      read(nin(jin),'(6e11.0)')(b(k),k=k0,npl)
    endif
    return
  end subroutine readlist
!
! =================================================================================================================================
! subroutine readtab1: Read an ENDF-6 TAB1 record
! =================================================================================================================================
!
  subroutine readtab1(nin,jin,c1,c2,l1,l2,nr,nbt,ibt,np,x,y)
    implicit real*8 (a-h, o-z)
    character*80 nin(*)
    dimension nbt(*),ibt(*),x(*), y(*)
    jin=jin+1
    read(nin(jin),'(2e11.0,4i11,i4,i2,i3,i5)')c1,c2,l1,l2,nr,np,mat,mf,mt,ns
    nline=nr/3
    iline=nr-3*nline
    kl=0
    if (nline.gt.0) then
      do i=1,nline
        k0=kl+1
        kl=kl+3
        jin=jin+1
        read(nin(jin),'(6i11)')(nbt(k),ibt(k),k=k0,kl)
      enddo
    endif
    if (iline.gt.0) then
      k0=kl+1
      jin=jin+1
      read(nin(jin),'(6i11)')(nbt(k),ibt(k),k=k0,nr)
    endif
    nline=np/3
    iline=np-3*nline
    kl=0
    if (nline.gt.0) then
      do i=1,nline
        k0=kl+1
        kl=kl+3
        jin=jin+1
        read(nin(jin),'(6e11.0)')(x(k),y(k),k=k0,kl)
      enddo
    endif
    if (iline.gt.0) then
      k0=kl+1
      jin=jin+1
      read(nin(jin),'(6e11.0)')(x(k),y(k),k=k0,np)
    endif
    return
  end subroutine readtab1
!
! =================================================================================================================================
! subroutine readtab2: Read an ENDF-6 TAB2 record
! =================================================================================================================================
!
  subroutine readtab2(nin,jin,c1,c2,l1,l2,nr,nbt,ibt,np2)
    implicit real*8 (a-h, o-z)
    character*80 nin(*)
    dimension nbt(*),ibt(*)
    jin=jin+1
    read(nin(jin),'(2e11.0,4i11,i4,i2,i3,i5)')c1,c2,l1,l2,nr,np2,mat,mf,mt,ns
    nline=nr/3
    iline=nr-3*nline
    kl=0
    if (nline.gt.0) then
      do i=1,nline
        k0=kl+1
        kl=kl+3
        jin=jin+1
        read(nin(jin),'(6i11)')(nbt(k),ibt(k),k=k0,kl)
      enddo
    endif
    if (iline.gt.0) then
      k0=kl+1
      jin=jin+1
      read(nin(jin),'(6i11)')(nbt(k),ibt(k),k=k0,nr)
    endif
    return
  end subroutine readtab2
!
! =================================================================================================================================
! subroutine readintg: Read an ENDF-6 INTG record
! =================================================================================================================================
!
  subroutine readintg(nin,jin,ndigit,ii,jj,nrow,kij)
    character*80 nin(*)
    dimension kij(*)
!   nrow   = number of rows of array kij, depends on the number of digits(ndigit)
!            ndigit =  2  3  4 5 6
!            nrow   = 18 13 11 9 8
    jin=jin+1
    if (ndigit.eq.2) then
      nrow=18
      read(nin(jin),'(2i5,1x,18i3,1x,i4,i2,i3,i5)')ii,jj,(kij(k),k=1,nrow),mat,mf,mt,ns
    elseif (ndigit.eq.3) then
      nrow=13
      read(nin(jin),'(2i5,1x,13i4,3x,i4,i2,i3,i5)')ii,jj,(kij(k),k=1,nrow),mat,mf,mt,ns
    elseif (ndigit.eq.4) then
      nrow=11
      read(nin(jin),'(2i5,1x,11i5,i4,i2,i3,i5)')ii,jj,(kij(k),k=1,nrow),mat,mf,mt,ns
    elseif (ndigit.eq.5) then
      nrow=9
      read(nin(jin),'(2i5,1x,9i6,1x,i4,i2,i3,i5)')ii,jj,(kij(k),k=1,nrow),mat,mf,mt,ns
    elseif (ndigit.eq.6) then
      nrow=8
      read(nin(jin),'(2i5,8i7,i4,i2,i3,i5)')ii,jj,(kij(k),k=1,nrow),mat,mf,mt,ns
    else
      write(*,*)' ERROR reading INTG record, NDIGIT=',ndigit,' JIN=',jin
      stop
    endif
    return
  end subroutine readintg
!
! =================================================================================================================================
! subroutine wrtext: Write a TEXT record
! =================================================================================================================================
!
  subroutine wrtext(nou,jou,mat,mf,mt,ns,text)
    character*80 nou(*)
    character*66 text
    ns=ns+1
    if (ns.gt.99999) ns=0
    jou=jou+1
    write(nou(jou),'(a66,i4,i2,i3,i5)')text,mat,mf,mt,ns
    return
  end subroutine wrtext
!
! =================================================================================================================================
! subroutine wrtcont: Write a general ENDF-6 CONT record
! =================================================================================================================================
!
  subroutine wrtcont(nou,jou,mat,mf,mt,ns,c1,c2,l1,l2,n1,n2)
    implicit real*8 (a-h, o-z)
    character*80 nou(*)
    character*11 sc1,sc2
    call chendf(c1,sc1)
    call chendf(c2,sc2)
    ns=ns+1
    if (ns.gt.99999) ns=0
    jou=jou+1
    write(nou(jou),'(2a11,4i11,i4,i2,i3,i5)')sc1,sc2,l1,l2,n1,n2,mat,mf,mt,ns
    return
  end subroutine wrtcont
!
! =================================================================================================================================
! subroutine wrtsend: Write a SEND record (Section END)
! =================================================================================================================================
!
  subroutine wrtsend(nou,jou,mat,mf,ns)
    character*80 nou(*)
    ns=0
    jou=jou+1
    write(nou(jou),'(66x,i4,i2,i3,i5)')mat,mf,0,99999
    return
  end subroutine wrtsend
!
! =================================================================================================================================
! subroutine wrtfend: Write a FEND record (File END)
! =================================================================================================================================
!
  subroutine wrtfend(nou,jou,mat,ns)
    character*80 nou(*)
    ns=0
    jou=jou+1
    write(nou(jou),'(66x,i4,i2,i3,i5)')mat,0,0,ns
    return
  end subroutine wrtfend
!
! =================================================================================================================================
! subroutine wrtmend: Write a MEND record (Material END)
! =================================================================================================================================
!
  subroutine wrtmend(nou,jou,ns)
    character*80 nou(*)
    ns=0
    jou=jou+1
    write(nou(jou),'(66x,i4,i2,i3,i5)')0,0,0,ns
    return
   end subroutine wrtmend
!
! =================================================================================================================================
! subroutine wrtend: Write a TEND record (Tape END)
! =================================================================================================================================
!
  subroutine wrtend(nou,jou,ns)
    character*80 nou(*)
    ns=0
    jou=jou+1
    write(nou(jou),'(66x,i4,i2,i3,i5)')-1,0,0,ns
    return
  end subroutine wrtend
!
! =================================================================================================================================
! subroutine wrtlist: Write a LIST record
! =================================================================================================================================
!
  subroutine wrtlist(nou,jou,mat,mf,mt,ns,c1,c2,l1,l2,npl,n2,b)
    implicit real*8 (a-h, o-z)
    character*80 nou(*)
    character*11 rec(6)
    dimension b(*)
    call wrtcont(nou,jou,mat,mf,mt,ns,c1,c2,l1,l2,npl,n2)
    nline=npl/6
    iline=npl-nline*6
    k=0
    if (nline.gt.0) then
      do i=1,nline
        do j=1,6
          k=k+1
          call chendf(b(k),rec(j))
        enddo
        ns=ns+1
        if (ns.gt.99999) ns=0
        jou=jou+1
        write(nou(jou),'(6a11,i4,i2,i3,i5)')(rec(j),j=1,6),mat,mf,mt,ns
      enddo
    endif
    if (iline.gt.0) then
      do j=1,iline
        k=k+1
        call chendf(b(k),rec(j))
      enddo
      do j=iline+1,6
        rec(j)='           '
      enddo
      ns=ns+1
      if (ns.gt.99999) ns=0
      jou=jou+1
      write(nou(jou),'(6a11,i4,i2,i3,i5)')(rec(j),j=1,6),mat,mf,mt,ns
    endif
    return
  end subroutine wrtlist
!
! =================================================================================================================================
! subroutine wrtab1: Write a TAB1 record
! =================================================================================================================================
!
  subroutine wrtab1(nou,jou,mat,mf,mt,ns,c1,c2,l1,l2,nr,nbt,ibt,np,x,y)
    implicit real*8 (a-h, o-z)
    character*80 nou(*)
    character*11 strx(3),stry(3)
    dimension nbt(*),ibt(*),x(*),y(*)
    dimension nn(3),ii(3)
    call wrtcont(nou,jou,mat,mf,mt,ns,c1,c2,l1,l2,nr,np)
    nline=nr/3
    iline=nr-3*nline
    k=0
    if (nline.gt.0) then
      do i=1,nline
        do j=1,3
          k=k+1
          nn(j)=nbt(k)
          ii(j)=ibt(k)
        enddo
        ns=ns+1
        if (ns.gt.99999) ns=0
        jou=jou+1
        write(nou(jou),'(6i11,i4,i2,i3,i5)')(nn(j),ii(j),j=1,3),mat,mf,mt,ns
      enddo
    endif
    if (iline.gt.0) then
      do j=1,iline
        k=k+1
        nn(j)=nbt(k)
        ii(j)=ibt(k)
      enddo
      ns=ns+1
      if (ns.gt.99999) ns=0
      jou=jou+1
      if (iline.eq.1) then
        write(nou(jou),'(2i11,44x,i4,i2,i3,i5)')nn(1),ii(1),mat,mf,mt,ns
      else
        write(nou(jou),'(4i11,22x,i4,i2,i3,i5)')(nn(j),ii(j),j=1,2),mat,mf,mt,ns
      endif
    endif
    nline=np/3
    ilinen=np-3*nline
    k=0
    if (nline.gt.0) then
      do i=1,nline
        do j=1,3
          k=k+1
          call chendf(x(k),strx(j))
          call chendf(y(k),stry(j))
        enddo
        ns=ns+1
        if (ns.gt.99999) ns=0
        jou=jou+1
        write(nou(jou),'(6a11,i4,i2,i3,i5)')(strx(j),stry(j),j=1,3),mat,mf,mt,ns
      enddo
    endif
    if (iline.gt.0) then
      do j=1,iline
        k=k+1
        call chendf(x(k),strx(j))
        call chendf(y(k),stry(j))
      enddo
      do j=iline+1,3
        strx(j)='           '
        stry(j)='           '
      enddo
      ns=ns+1
      if (ns.gt.99999) ns=0
      jou=jou+1
      write(nou(jou),'(6a11,i4,i2,i3,i5)')(strx(j),stry(j),j=1,3),mat,mf,mt,ns
    endif
    return
  end subroutine wrtab1
!
! =================================================================================================================================
! subroutine wrtab2: Write a TAB2 record
! =================================================================================================================================
!
  subroutine wrtab2(nou,jou,mat,mf,mt,ns,c1,z,l1,l2,nr,nbt,ibt,np2)
    implicit real*8 (a-h, o-z)
    character*80 nou(*)
    dimension nbt(*),ibt(*)
    dimension nn(3),ii(3)
    call wrtcont(nou,jou,mat,mf,mt,ns,c1,z,l1,l2,nr,np2)
    nline=nr/3
    iline=nr-3*nline
    k=0
    if (nline.gt.0) then
      do i=1,nline
        do j=1,3
          k=k+1
          nn(j)=nbt(k)
          ii(j)=ibt(k)
        enddo
        ns=ns+1
        if (ns.gt.99999) ns=0
        jou=jou+1
        write(nou(jou),'(6i11,i4,i2,i3,i5)')(nn(j),ii(j),j=1,3),mat,mf,mt,ns
      enddo
    endif
    if (iline.gt.0) then
      do j=1,iline
        k=k+1
        nn(j)=nbt(k)
        ii(j)=ibt(k)
      enddo
      ns=ns+1
      if (ns.gt.99999) ns=0
      jou=jou+1
      if (iline.eq.1) then
        write(nou(jou),'(2i11,44x,i4,i2,i3,i5)')nn(1),ii(1),mat,mf,mt,ns
      else
        write(nou(jou),'(4i11,22x,i4,i2,i3,i5)')(nn(j),ii(j),j=1,2),mat,mf,mt,ns
      endif
    endif
    return
  end subroutine wrtab2
!
! =================================================================================================================================
! subroutine wrtintg: Write an ENDF-6 INTG record
! =================================================================================================================================
!
  subroutine wrtintg(nou,jou,mat,mf,mt,ns,ii,jj,nrow,kij)
    character*80 nou(*)
    dimension kij(*)
!   nrow   = number of rows of array kij, depends on the number of digits(ndigit)
!            ndigit =  2  3  4 5 6
!            nrow   = 18 13 11 9 8
    jou=jou+1
    if (nrow.eq.18) then
      write(nou(jou),'(2i5,1x,18i3,1x,i4,i2,i3,i5)')ii,jj,(kij(k),k=1,nrow),mat,mf,mt,ns
    elseif (nrow.eq.13) then
      write(nou(jou),'(2i5,1x,13i4,3x,i4,i2,i3,i5)')ii,jj,(kij(k),k=1,nrow),mat,mf,mt,ns
    elseif (nrow.eq.11) then
      write(nou(jou),'(2i5,1x,11i5,i4,i2,i3,i5)')ii,jj,(kij(k),k=1,nrow),mat,mf,mt,ns
    elseif (nrow.eq.9) then
      write(nou(jou),'(2i5,1x,9i6,1x,i4,i2,i3,i5)')ii,jj,(kij(k),k=1,nrow),mat,mf,mt,ns
    elseif (nrow.eq.8) then
      write(nou(jou),'(2i5,8i7,i4,i2,i3,i5)')ii,jj,(kij(k),k=1,nrow),mat,mf,mt,ns
    else
      write(*,*)' ERROR writing INTG record, NROW=',nrow,' JOU=',jou
      stop
    endif
    return
  end subroutine wrtintg
!
! =================================================================================================================================
!
! MF0.2: Auxiliary procedures for preparing ENDF-6 records:
! --------------------------------------------------------
!
! =================================================================================================================================
! subroutine chendf: Convert a real*8 to ENDF-6 11-characters numeric string
! =================================================================================================================================
!   convert +n.nnnnnnnnE+mm  to  +.nnnnnnnnn
!                                +n.nnnnnnnn
!                                +nn.nnnnnnn
!                                +nnn.nnnnnn
!                                +nnnn.nnnnn
!                                +nnnnn.nnnn
!                                +nnnnnn.nnn
!                                +nnnnnnn.nn
!                                +nnnnnnnn.n
!                                +nnnnnnnnn.
!                                +n.nnnnnn+m
!                                +n.nnnnn+mm
!
  subroutine chendf(ffin,str11)
    implicit real*8 (a-h,o-z)
    character*11 str11
    character*12 str12
    character*13 str13
    character*15 str15
    write(str15,'(1pe15.8)')ffin
    read(str15,'(e15.0)')ff
    aff=abs(ff)
    if (aff.lt.1.19d-38) then
      str11=' 0.0       '
    elseif (aff.gt.3.39d+38) then
      if (ff.lt.0.0d0) then
        str11='-3.90001+38'
      else
        str11=' 3.90001+38'
      endif
    else
      write(str15,'(1pe15.8)')ff
      read(str15,'(12x,i3)')iex
      select case (iex)
        case (-9,-8,-7,-6,-5,-4, 9)
          write(str13,'(1pe13.6)')ff
          str11(1:9)=str13(1:9)
          str11(10:10)=str13(11:11)
          str11(11:11)=str13(13:13)
        case (-3,-2,-1)
          write(str12,'(f12.9)')ff
          str11(1:1)=str12(1:1)
          str11(2:11)=str12(3:12)
        case (0)
          write(str11,'(f11.8)')ff
        case (1)
          write(str11,'(f11.7)')ff
        case (2)
          write(str11,'(f11.6)')ff
        case (3)
          write(str11,'(f11.5)')ff
        case (4)
          write(str11,'(f11.4)')ff
        case (5)
          write(str11,'(f11.3)')ff
        case (6)
          write(str11,'(f11.2)')ff
        case (7)
          write(str11,'(f11.1)')ff
        case (8)
          write(str11,'(f11.0)')ff
        case default
          write(str12,'(1pe12.5)')ff
          str11(1:8)=str12(1:8)
          str11(9:11)=str12(10:12)
      end select
    endif
    return
  end subroutine chendf
!
! =================================================================================================================================
! subroutine packibt: Pack interpolation table for TAB1 and TAB2 records
! =================================================================================================================================
!   np         = number of points of the X-Y table (np-1=number of intervals)
!   ibtu       = array containing the interpolation law between [x(i),x(i+1)]
!   nr,nbt,ibt = Packed interpolation table for TAB1 and TAB2 records
!   ierr       = error trigger (nr > 20 is not allowed in ENDF-6 format)
!
  subroutine packibt(np,ibtu,nr,nbt,ibt)
    dimension ibtu(*),nbt(*),ibt(*)
    nr=0
    ilaw=ibtu(1)
    nru=np-1
    do i=2,nru
      if (ilaw.ne.ibtu(i)) then
        nr=nr+1
        nbt(nr)=i
        ibt(nr)=ilaw
        ilaw=ibtu(i)
      endif
    enddo
    nr=nr+1
    nbt(nr)=np
    ibt(nr)=ibtu(nru)
    if (nr.gt.nrmax) then
      write(*,*)' Error: Too many interpolation ranges nr=',nr,' nrmax=',nrmax
      stop
    endif
    return
  end subroutine packibt
!
! =================================================================================================================================
! subroutine unpackibt: Unpack interpolation table from TAB1 and TAB2 records
! =================================================================================================================================
!   nr,nbt,ibt = Packed interpolation table from TAB1 and TAB2 records (nr > 20 is not allowed in ENDF-6 format)
!   np         = np number of points of the X-Y table (np-1=number of intervals)
!   ibtu       = array containing the interpolation law between [x(i),x(i+1)]
!
  subroutine unpackibt(nr,nbt,ibt,np,ibtu)
    dimension nbt(*),ibt(*),ibtu(*)
    if (nr.gt.nrmax) then
      write(*,*)' Error: Too many ENDF-6 interpolation ranges nr=',nr,' nrmax=',nrmax
      stop
    else
      jf=0
      do i=1,nr
        j0=jf+1
        jf=nbt(i)-1
        ilaw=ibt(i)
        do j=j0,jf
          ibtu(j)=ilaw
        enddo
      enddo
      np1=np-1
      if (jf.lt.np1) then
        write(*,*)' Warning: Interpolation law could be incorrent NR<(NP-1) ',nr,' < ',np1
        do i=jf+1,np1
          ibt(i)=ibt(jf)
        enddo
      elseif (jf.gt.np1) then
        write(*,*)' Warning: Interpolation law could be incorrent NR>(NP-1) ',nr,' > ',np1
      endif
      return
    endif
  end subroutine unpackibt
!
! =================================================================================================================================
!
! MF0.3: Procedures for reading/writing ENDF-6 tapes and searching [mat,mf,mt] structures:
! ---------------------------------------------------------------------------------------
!
! =================================================================================================================================
! subroutine readtape: Read a complete ENDF-6 formatted tape
! =================================================================================================================================
!
  subroutine readtape(endftape,maxdim,tapehead,nin,ninrec,ierr)
    character*120 endftape
    character*80 nin(*),tapehead
    data inp/20/
    open(inp,file=endftape)
    read(inp,'(a80)',iostat=iosinp)tapehead
    if (iosinp.lt.0) then
      ierr=-2
    else
      ninrec=0
      mat=10000
      iosinp=0
      do while (mat.ge.0.and.ninrec.lt.maxdim.and.iosinp.ge.0)
        ninrec=ninrec+1
        read(inp,'(a80)',iostat=iosinp)nin(ninrec)
        if (iosinp.lt.0) then
          exit
        else
          read(nin(ninrec),'(66x,i4)')mat
        endif
      enddo
      if (mat.lt.0) then
        ierr=0
      elseif (ninrec.ge.maxdim) then
        ierr=1
      else
        ierr=-1
      endif
    endif
    close(inp)
    return
  end subroutine readtape
!
! =================================================================================================================================
! subroutine wrt2tape: write (joumax-joumin+1) records to an ENDF-6 formatted tape opened on unit=iou
! =================================================================================================================================
!
  subroutine wrt2tape(iou,nou,joumin,joumax)
    character*80 nou(*)
    do jou=joumin,joumax
      write(iou,'(a80)')nou(jou)
    enddo
    return
  end subroutine wrt2tape
!
! =================================================================================================================================
! subroutine findmat: Find the 1st record of a [mat]-structure on an ENDF-6 tape(array)
! =================================================================================================================================
!
  subroutine findmat(nin,jin,mat,ierr)
    character*80 nin(*)
    if (jin.le.0) jin=1
    read(nin(jin),'(66x,i4)',iostat=iosnin)mati
    if (iosnin.lt.0) then
      ierr=2
    else
      if (mati.eq.0) then
        jin=jin+1
        read(nin(jin),'(66x,i4)',iostat=iosnin)mati
      endif
      if (mati.ge.mat.or.mati.le.0.or.iosnin.lt.0) jin=0
      mati=0
      iosnin=0
      do while(mati.lt.mat.and.mati.ge.0.and.iosnin.ge.0)
        jin=jin+1
        read(nin(jin),'(66x,i4)',iostat=iosnin)mati
      enddo
      if (mati.eq.mat) then
        ierr=0
        jin=jin-1
      else
        ierr=1
      endif
    endif
    return
  end subroutine findmat
!
! =================================================================================================================================
! subroutine findmf: Find the 1st record of [mat,mf]-structure on an ENDF-6 tape(array)
! =================================================================================================================================
!
  subroutine findmf(nin,jin,mat,mf,ierr)
    character*80 nin(*)
    mkey=mat*100+mf
    if (jin.le.0) jin=1
    read(nin(jin),'(66x,i4,i2)',iostat=iosnin)mati,mfi
    if (iosnin.lt.0) then
      ierr=2
    else
      if (mati.eq.0) then
        jin=jin+1
        read(nin(jin),'(66x,i4,i2)',iostat=iosnin)mati,mfi
      elseif(mfi.eq.0) then
        jin=jin-1
        read(nin(jin),'(66x,i4,i2)',iostat=iosnin)mati,mfi
      endif
      ikey=mati*100+mfi
      if (ikey.ge.mkey.or.ikey.le.0.or.iosnin.lt.0) jin=0
      ikey=0
      iosnin=0
      do while(ikey.lt.mkey.and.ikey.ge.0.and.iosnin.ge.0)
        jin=jin+1
        read(nin(jin),'(66x,i4,i2)',iostat=iosnin)mati,mfi
        ikey=mati*100+mfi
      enddo
      if (ikey.eq.mkey) then
        ierr=0
        jin=jin-1
      else
        ierr=1
      endif
    endif
    return
  end subroutine findmf
!
! =================================================================================================================================
! subroutine findmt: Find the 1st record of [mat,mf,mf]-structure on an ENDF-6 tape(array)
! =================================================================================================================================
!
  subroutine findmt(nin,jin,mat,mf,mt,ierr)
    character*80 nin(*)
    mkey=1000*(mat*100+mf)+mt
    if (jin.le.0) jin=1
    read(nin(jin),'(66x,i4,i2,i3)',iostat=iosnin)mati,mfi,mti
    if (iosnin.lt.0) then
      ierr=2
    else
      if (mati.eq.0) then
        jin=jin+1
        read(nin(jin),'(66x,i4,i2,i3)',iostat=iosnin)mati,mfi,mti
      elseif(mfi.eq.0) then
        jin=jin-2
        read(nin(jin),'(66x,i4,i2,i3)',iostat=iosnin)mati,mfi,mti
      elseif(mti.eq.0) then
        jin=jin-1
        read(nin(jin),'(66x,i4,i2,i3)',iostat=iosnin)mati,mfi,mti
      endif
      ikey=1000*(mati*100+mfi)+mti
      if (ikey.ge.mkey.or.ikey.le.0.or.iosnin.lt.0) jin=0
      ikey=0
      iosnin=0
      do while(ikey.lt.mkey.and.ikey.ge.0.and.iosnin.ge.0)
        jin=jin+1
        read(nin(jin),'(66x,i4,i2,i3)',iostat=iosnin)mati,mfi,mti
        ikey=1000*(mati*100+mfi)+mti
      enddo
      if (ikey.eq.mkey) then
        ierr=0
        jin=jin-1
      else
        ierr=1
      endif
    endif
    return
  end subroutine findmt
!
! =================================================================================================================================
! subroutine getnextmt: find next [mt] section for the current [mat, mf]
! =================================================================================================================================
!
  subroutine getnextmt(nin,jin,mf,mt)
    character*80 nin(*)
    read(nin(jin),'(70x,i2,i3)',iostat=iosnin)mfi,mti
    if (iosnin.lt.0) then
      mt=-2
    elseif (mfi.ne.mf) then
      mt=-1
    elseif (mti.eq.0) then
      jin=max(jin-1,0)
    endif
    mti=1000
    iosnin=0
    do while (mti.ne.0.and.iosnin.ge.0)
      jin=jin+1
      read(nin(jin),'(70x,i2,i3)',iostat=iosnin)mfi,mti
    enddo
    if (iosnin.lt.0) then
      mt=-2
    else
      jin=jin+1
      read(nin(jin),'(70x,i2,i3)',iostat=iosnin)mfi,mti
      if (mfi.eq.mf.and.mti.ne.0.and.iosnin.ge.0) then
        mt=mti
      else
        mt=-1
      endif
    endif
    return
  end subroutine getnextmt
!
! =================================================================================================================================
! subroutine nextsub6: find next subsection for [mat=current, mf=6, mt=current]
! =================================================================================================================================
!
  subroutine nextsub6(nin,jin,law,nbt,ibt,x,b)
    implicit real*8 (a-h,o-z)
    character*80 nin(*)
    dimension nbt(*),ibt(*),x(*),b(*)
    if (law.eq.1.or.law.eq.2.or.law.eq.5) then
      call readtab2(nin,jin,c1,c2,l1,l2,n1,nbt,ibt,ne)
      do i=1,ne
        call readlist(nin,jin,c1,c2,l1,l2,n1,n2,b)
      enddo
    elseif (law.eq.6) then
      call readcont(nin,jin,c1,c2,l1,l2,n1,n2,mat,mf,mt,ns)
    elseif (law.eq.7) then
      call readtab2(nin,jin,c1,c2,l1,l2,n1,nbt,ibt,ne)
      do i=1,ne
        call readtab2(nin,jin,c1,c2,l1,l2,n1,nbt,ibt,nmu)
        do j=1,nmu
          call readtab1(nin,jin,c1,c2,l1,l2,n1,nbt,ibt,n2,x,b)
        enddo
      enddo
    elseif ((law.ne.-15.and.law.ne.-5.and.law.ne.0.and.law.ne.3.and.law.ne.4).or.law.gt.7) then
      write(*,*)' ERROR: unknown MF6/LAW=',law
      stop
    endif
    return
  end subroutine nextsub6
!
! =================================================================================================================================
!
! MF3: Procedures for reading and writing ENDF-6/MF3/MT sections:
! --------------------------------------------------------------
!
! =================================================================================================================================
! subroutine getmf3mt_form: Get number of incident energy points.
! =================================================================================================================================
!
!   np: number of incident energy points. Dimension of the incident energy index.
!
  subroutine getmf3mt_form(nin,jin,np)
    implicit real*8 (a-h, o-z)
    character*80 nin(*)
    jin0=jin
    call readcont(nin,jin,c1,c2,l1,l2,n1,n2,mat,mf,mt,ns)
    call readcont(nin,jin,c1,c2,l1,l2,nr,np,mat,mf,mt,ns)
    if (np.gt.ne3max) then
      write(*,*)' MF3: Warning number incident energy points ne=',np,' > ',ne3max
    endif
    if (nr.gt.nrmax) then
      write(*,*)' Error: Too many ENDF-6 interpolation ranges nr=',nr,' nrmax=',nrmax
      stop
    endif
    jin=jin0
    return
  end subroutine getmf3mt_form
!
! =================================================================================================================================
! subroutine readmf3mt_std: Read one MF3 section MT(cross section data). Use packed ibt
! =================================================================================================================================
  subroutine readmf3mt_std(nin,jin,za,awr,qm,qi,lr,np,x,y,nr,nbt,ibt)
    implicit real*8 (a-h, o-z)
    character*80 nin(*)
    dimension nbt(*),ibt(*),x(*),y(*)
    call readcont(nin,jin,za,awr,l1,l2,n1,n2,mat,mf,mt,ns)
    call readtab1(nin,jin,qm,qi,l1,lr,nr,nbt,ibt,np,x,y)
    return
  end subroutine readmf3mt_std
!
! =================================================================================================================================
! subroutine readmf3mt_ext: Read one MF3 section MT(cross section data). Use unpacked ibt
! =================================================================================================================================
!
  subroutine readmf3mt_ext(nin,jin,za,awr,qm,qi,lr,np,x,y,ibt)
    implicit real*8 (a-h, o-z)
    character*80 nin(*)
    dimension ibt(*),x(*),y(*)
    dimension nbt(nrmax),jbt(nrmax)
    call readmf3mt_std(nin,jin,za,awr,qm,qi,lr,np,x,y,nrj,nbt,jbt)
    call unpackibt(nrj,nbt,jbt,np,ibt)
    return
  end subroutine readmf3mt_ext
!
! =================================================================================================================================
! subroutine wrtmf3mt_std: Write one MF3 section MT(cross section data). Use packed ibt
! =================================================================================================================================
!
  subroutine wrtmf3mt_std(nou,jou,mat,mt,za,awr,qm,qi,lr,np,x,y,nr,nbt,ibt)
    implicit real*8 (a-h, o-z)
    character*80 nou(*)
    dimension nbt(*),ibt(*),x(*),y(*)
    mf=3
    l1=0
    l2=0
    n1=0
    n2=0
    ns=0
    call wrtcont(nou,jou,mat,mf,mt,ns,za,awr,l1,l2,n1,n2)
    call wrtab1(nou,jou,mat,mf,mt,ns,qm,qi,l1,lr,nr,nbt,ibt,np,x,y)
    call wrtsend(nou,jou,mat,mf,ns)
    return
  end subroutine wrtmf3mt_std
!
! =================================================================================================================================
! subroutine wrtmf3mt_ext: Write one MF3 section MT(cross section data). Use packed ibt
! =================================================================================================================================
!
  subroutine wrtmf3mt_ext(nou,jou,mat,mt,za,awr,qm,qi,lr,np,x,y,ibt)
    implicit real*8 (a-h, o-z)
    character*80 nou(*)
    dimension ibt(*),x(*),y(*)
    dimension nbt(nrmax),jbt(nrmax)
    call packibt(np,ibt,nrj,nbt,jbt)
    call wrtmf3mt_std(nou,jou,mat,mt,za,awr,qm,qi,lr,np,x,y,nrj,nbt,jbt)
    return
  end subroutine wrtmf3mt_ext
!
! =================================================================================================================================
! subroutine wrtmf3mt_law: Write one MF3 section MT(cross section data). Use lin-lin
! =================================================================================================================================
!
  subroutine wrtmf3mt_law(nou,jou,mat,mt,za,awr,qm,qi,lr,np,x,y,ilaw)
    implicit real*8 (a-h, o-z)
    character*80 nou(*)
    dimension x(*),y(*)
    dimension nbt(1),ibt(1)
    nr=1
    nbt(1)=np
    ibt(1)=ilaw
    call wrtmf3mt_std(nou,jou,mat,mt,za,awr,qm,qi,lr,np,x,y,nr,nbt,ibt)
    return
  end subroutine wrtmf3mt_law
!
! =================================================================================================================================
! subroutine wrtmf3mt_lin: Write one MF3 section MT(cross section data). Use lin-lin
! =================================================================================================================================
!
  subroutine wrtmf3mt_lin(nou,jou,mat,mt,za,awr,qm,qi,lr,np,x,y)
    implicit real*8 (a-h, o-z)
    character*80 nou(*)
    dimension x(*),y(*)
    dimension nbt(1),ibt(1)
    nr=1
    nbt(1)=np
    ibt(1)=2
    call wrtmf3mt_std(nou,jou,mat,mt,za,awr,qm,qi,lr,np,x,y,nr,nbt,ibt)
    return
  end subroutine wrtmf3mt_lin
!
!=====================================================================================================
!
! MF4: Procedures for reading and writing ENDF-6/MF4/MT sections:
! --------------------------------------------------------------
!
! =================================================================================================================================
! subroutine getmf4mt_form: To get the type and dimension of the MF4 section (angular distribution)
! =================================================================================================================================
!
!   ltt: Type of section
!        ltt=0 (iso)= Isotropic angular distribution at all energies
!        ltt=1 (leg)= Legendre expansion coefficients.
!        ltt=2 (tab)= Tabulated probability distributions. Dimension xm
!        ltt=3 (mix)= Lower energy range, Legendre expansion coefficients
!                     Higher energy range, Tabulated probability distributions
!   nel: number of incident energies for Legendre expansion (ltt=1,3)
!   nal: number of Legendre coefficient (ltt=1,3)
!   neh: number of incident energies for tabular data (ltt=2,3)
!   nah: number of cosines for tabular data (ltt=2,3)
!
!   Isotropic representation (ltt=0): (null dimensioning = nal=nel=nah=neh=0)
!   Legendre  representation (ltt=1): (angular dimension=nal, energy dimension=nel, neh=nah=0)
!   Tabular   representation (ltt=2): (angular dimension=nah, energy dimension=neh, nel=nal=0)
!   Mixed     representation (ltt=3): Legendre expansion(angular dimension=nal, energy dimension=nel)
!                                     Tabular data(angular dimension=nah, energy dimension=neh)
!
  subroutine getmf4mt_form(nin,jin,ltt,nel,nal,neh,nah)
    implicit real*8 (a-h,o-z)
    character*80 nin(*)
    dimension nbt(nrmax),ibt(nrmax),b(nlmax),x1(nmu4max),y1(nmu4max)
    jin0=jin
    nel=0
    neh=0
    nal=0
    nah=0
    n0=0
    call readcont(nin,jin,c1,c2,l1,ltt,n1,n2,mat,mf,mt,ns)
    call readcont(nin,jin,c1,c2,li,ltt,n1,nm,mat,mf,mt,ns)
    if (ltt.eq.1.or.ltt.eq.3) then
      call readtab2(nin,jin,c1,c2,l1,l2,n1,nbt,ibt,nel)
      if (n1.gt.nrmax) then
        write(*,*)' MF4 Error: Too many TAB2 interpolation ranges nr=',n1,' nrmax=',nrmax
        stop
      endif
      do i=1,nel
        call readlist(nin,jin,t,c2,lt,l2,n1,n2,b)
        if (n1.gt.nlmax) then
          write(*,*)' MF4 Error: Order of Legendre expansion greater than ',nlmax,' nl=',n1,' at E=',c2
          stop
        elseif (n1.gt.nal) then
          nal=n1
        endif
        if (t.ne.0.0d0.or.lt.ne.0) n0=n0+1
      enddo
      nal=nal+1
      if (nal*nel.gt.npmax) then
        write(*,*)' MF4 Error: nal*nel=',nal*nel,' > npmax=',npmax
        stop
      endif
    endif
    if (ltt.eq.2.or.ltt.eq.3) then
      call readtab2(nin,jin,c1,c2,l1,l2,n1,nbt,ibt,neh)
      if (n1.gt.nrmax) then
        write(*,*)' MF4 Error: Too many TAB2 interpolation ranges nr=',n1,' nrmax=',nrmax
        stop
      endif
      do i=1,neh
        call readtab1(nin,jin,t,c2,lt,l2,n1,nbt,ibt,n2,x1,y1)
        if (n1.gt.nrmax) then
          write(*,*)' MF4 Error: Too many ENDF-6 interpolation ranges nr=',n1,' nrmax=',nrmax,' at E=',c2
          stop
        endif
        if (n2.gt.nmu4max) then
          write(*,*)' MF4 Warning: Number of cosines greater than ',nmu4max,' nah=',n2,' at E=',c2
        elseif (n2.gt.nah) then
          nah=n2
        endif
        if (t.ne.0.0d0.or.lt.ne.0) n0=n0+1
      enddo
      if (nah*neh.gt.npmax) then
        write(*,*)' MF4 Error: nah*neh=',nah*neh,' > npmax=',npmax
        stop
      endif
    endif
    if (n0.gt.0) then
      write(*,*)' Warning: Temperature T and flag LT not equal to 0 for ',n0,' incident energies'
    endif
    jin=jin0
    return
  end subroutine getmf4mt_form
!
! =================================================================================================================================
! subroutine readmf4mt_iso: Read purely isotropic angular distribution (ltt=0, li=1)
! =================================================================================================================================
!
!    za: ZA number (real*8)
!   awr: Atomic weight ratio to neutron mass (real*8)
!   lct: Reference system (integer)
!
  subroutine readmf4mt_iso(nin,jin,za,awr,lct)
    implicit real*8 (a-h,o-z)
    character*80 nin(*)
    call readcont(nin,jin,za,awr,l1,ltt,n1,n2,mat,mf,mt,ns)
    call readcont(nin,jin,c1,awr,li,lct,n1,n2,mat,mf,mt,ns)
    return
  end subroutine readmf4mt_iso
!
! =================================================================================================================================
! subroutine readmf4mt_leg: Read one MF4 section given in term of Legendre expansion (ltt=1,li=0)
! =================================================================================================================================
!
!    za: ZA number (real*8)
!   awr: Atomic weight ratio to neutron mass (real*8)
!   lct: Reference system (integer)
!    ne: number of incident energies (integer)
!     e: incident energies
!        real*8 array e=[e(i),i=1..ne]
!   ile: interpolation law by incident energy interval
!        integer array ile=[ile(i),1..ne-1]
!    na: Maximum expected number of Legendre coefficients (integer)
!    nl: Number of Legrende coefficients by incident energy
!        integer array nl=[nl(i),i=1..ne]
!        (max(nl(i)) cannot be greater than na)
!     a: Legendre coefficients
!        real*8 array a=[a(l,i), l=1..nl(i), i=1..ne]
!
  subroutine readmf4mt_leg(nin,jin,za,awr,lct,ne,e,ile,na,nl,a)
    implicit real*8 (a-h,o-z)
    character*80 nin(*)
    dimension e(*),ile(*),a(na,*),nl(*)
    dimension nbt(nrmax),ibt(nrmax),b(na)
    call readcont(nin,jin,za,awr,l1,ltt,n1,n2,mat,mf,mt,ns)
    call readcont(nin,jin,c1,awr,li,lct,n1,n2,mat,mf,mt,ns)
    call readtab2(nin,jin,c1,c2,l1,l2,nr,nbt,ibt,ne)
    call unpackibt(nr,nbt,ibt,ne,ile)
    do i=1,ne
      call readlist(nin,jin,t,e(i),lt,l2,n1,n2,b)
      nl(i)=n1+1
      a(1,i)=1.0d0
      do j=1,n1
        a(j+1,i)=b(j)
      enddo
    enddo
    return
  end subroutine readmf4mt_leg
!
! =================================================================================================================================
! subroutine readmf4mt_tab: Read one MF4 section given in term of tabulated distributions (ltt=2,li=0)
! =================================================================================================================================
!
!     za: ZA number (real*8)
!    awr: Atomic weight ratio to neutron mass (real*8)
!    lct: Reference system (integer)
!     ne: number of incident energies (integer = second dimension of 2D-arrays)
!      e: incident energies
!         real*8 array e=[e(i),i=1..ne]
!    ile: interpolation law by incident energy interval
!         integer array ile=[ile(i),1..ne-1]
!     na: Maximum number of cosines (integer = first dimension 2D-arrays)
!    nmu: number of cosine by incident energy
!         integer array nmu=[nmu(i),1..ne] nmu(i)<=na
!    xmu: set of cosine values by incident energy
!         real*8 xmu=[xmu(k,i),k=1..nmu(i),i=1..ne]
!    ymu: angle-distribution values by incident energy
!         real*8 ymu=[ymu(k,i),k=1,nmu(i),i=1..ne]
!   ilmu: cosine interpolation law by incident energy
!         integer ilmu=[ilmu(k,i),k=1..nmu(i)-1,i=1..ne]
!
  subroutine readmf4mt_tab(nin,jin,za,awr,lct,ne,e,ile,na,nmu,xmu,ymu,ilmu)
    implicit real*8 (a-h,o-z)
    character*80 nin(*)
    dimension e(*),ile(*),nmu(*),xmu(na,*),ymu(na,*),ilmu(na,*)
    dimension nbt(nrmax),ibt(nrmax),jbt(na),x(na),y(na)
    call readcont(nin,jin,za,awr,l1,ltt,n1,n2,mat,mf,mt,ns)
    call readcont(nin,jin,c1,awr,li,lct,n1,n2,mat,mf,mt,ns)
    call readtab2(nin,jin,c1,c2,l1,l2,nr,nbt,ibt,ne)
    call unpackibt(nr,nbt,ibt,ne,ile)
    do i=1,ne
      call readtab1(nin,jin,t,e(i),lt,l2,nr,nbt,ibt,np,x,y)
      nmu(i)=np
      do k=1,np
        xmu(k,i)=x(k)
        ymu(k,i)=y(k)
      enddo
      call unpackibt(nr,nbt,ibt,np,jbt)
      do k=1,np-1
        ilmu(k,i)=jbt(k)
      enddo
    enddo
    return
  end subroutine readmf4mt_tab
!
! =================================================================================================================================
! subroutine readmf4mt_mix: Read one MF4 mixed section (ltt=3,li=0)
! =================================================================================================================================
!
!     1. Legendre expansion is given at lower energies, and
!     2. Probability distributions at higher energies
!
!     za: ZA number (real*8)
!    awr: Atomic weight ratio to neutron mass (real*8)
!    lct: Reference system (integer)
!    nel: number of incident energies in the lower energy range (integer)
!     el: incident energies in the lower energy range
!         real*8 array el=[el(i),i=1..nel]
!   ilel: interpolation law by incident energy interval
!         integer array ilel=[ilel(i),1..nel-1]
!    nal: Maximum number of Legendre coefficients <=ncmax (integer = first dimension of array a)
!     nl: Number of Legrende coefficients by incident energy
!         integer array nl=[nl(i),i=1..nel]
!         (max(nl(i)) cannot be greater than nal)
!      a: Legendre coefficients
!         real*8 array a=[a(l,i), l=1..nl(i), i=1..nel] (nl(i)<=nal)
!    neh: number of incident energies in the higher energy range (integer)
!     eh: incident energies in the higher energy range
!         real*8 array eh=[eh(i),i=1..neh]
!   ileh: interpolation law by incident energy interval
!         integer array ileh=[ileh(i),1..neh-1]
!    nah: Maximum number of cosines (integer = first dimension of arrays xmu,ymu,ilmu)
!    nmu: number of cosines by incident energy
!         integer array nmu=[nmu(i),1..neh]
!         (max(nmu(i)) cannot be greater than nah)
!    xmu: set of cosine values by incident energy
!         real*8 xmu=[xmu(k,i),k=1..nmu(i),i=1..neh]
!    ymu: angle-distribution by incident energy
!         real*8 ymu=[ymu(k,i),k=1,nmu(i),i=1..neh]
!   ilmu: cosine interpolation law by incident energy
!         integer ilmu=[ilmu(k,i),k=1..nmu(i)-1,i=1..neh]
!
  subroutine readmf4mt_mix(nin,jin,za,awr,lct,nel,el,ilel,nal,nl,a,neh,eh,ileh,nah,nmu,xmu,ymu,ilmu)
    implicit real*8 (a-h,o-z)
    parameter (eps=1.0d-6)
    character*80 nin(*)
    dimension el(*),ilel(*),nl(*),a(nal,*)
    dimension eh(*),ileh(*),nmu(*),xmu(nah,*),ymu(nah,*),ilmu(nah,*)
    dimension nbt(nrmax),ibt(nrmax),jbt(nah),b(nal),x(nah),y(nah)
    call readcont(nin,jin,za,awr,l1,ltt,n1,n2,mat,mf,mt,ns)
    call readcont(nin,jin,c1,awr,li,lct,n1,n2,mat,mf,mt,ns)
!
!   Legendre expansion data
!
    call readtab2(nin,jin,c1,c2,l1,l2,nr,nbt,ibt,nel)
    call unpackibt(nr,nbt,ibt,nel,ilel)
    do i=1,nel
      call readlist(nin,jin,t,el(i),lt,l2,n1,n2,b)
      nl(i)=n1+1
      a(1,i)=1.0d0
      do j=1,n1
        a(j+1,i)=b(j)
      enddo
    enddo
!
!   Tabulated probability data
!
    call readtab2(nin,jin,c1,c2,l1,l2,nr,nbt,ibt,neh)
    call unpackibt(nr,nbt,ibt,neh,ileh)
    do i=1,neh
      call readtab1(nin,jin,t,eh(i),lt,l2,nr,nbt,ibt,np,x,y)
      call unpackibt(nr,nbt,ibt,np,jbt)
      nmu(i)=np
      do k=1,np-1
        xmu(k,i)=x(k)
        ymu(k,i)=y(k)
        ilmu(k,i)=jbt(k)
      enddo
      xmu(np,i)=x(np)
      ymu(np,i)=y(np)
    enddo
    if (abs(eh(1)-el(nel)).gt.(eps*el(nel))) then
      write(*,*)' MF4 Warning: Energy discontinuity between Legendre expansion and Tabulated data'
    else
      eh(1)=el(nel)
    endif
    return
  end subroutine readmf4mt_mix
!
! =================================================================================================================================
! subroutine wrtmf4mt_iso: Write purely isotropic angular distribution (ltt=0,li=1)
! =================================================================================================================================
!
!    za: ZA number (real*8)
!   awr: Atomic weight ratio to neutron mass (real*8)
!   lct: Reference system (integer)
!
  subroutine wrtmf4mt_iso(nou,jou,mat,mt,za,awr,lct)
    implicit real*8 (a-h, o-z)
    character*80 nou(*)
    mf=4
    ns=0
    ltt=0
    li=1
    c1=0.0d0
    l1=0
    l2=0
    n1=0
    n2=0
    call wrtcont(nou,jou,mat,mf,mt,ns,za,awr,l1,ltt,n1,n2)
    call wrtcont(nou,jou,mat,mf,mt,ns,c1,awr,li,lct,n1,n2)
    call wrtsend(nou,jou,mat,mf,ns)
    return
  end subroutine wrtmf4mt_iso
!
! =================================================================================================================================
! subroutine wrtmf4mt_leg:  Write one MF4 section given in term of Legendre expansion (ltt=1,li=0)
! =================================================================================================================================
!
!    za: ZA number (real*8)
!   awr: Atomic weight ratio to neutron mass (real*8)
!   lct: Reference system (integer)
!    ne: number of incident energies (integer)
!     e: incident energies
!        real*8 array e=[e(i),i=1..ne]
!   ile: interpolation law by incident energy interval
!        integer array ile=[ile(i),1..ne-1]
!    na: Maximum number of Legendre coefficients <=ncmax (first dimension of array a)
!    nl: Number of Legrende coefficient by incident energy
!        integer array nl=[nl(i),i=1..ne]
!        (max(nl(i)) cannot be greater than na)
!     a: Legendre coefficients
!        real*8 array a=[a(l,i), l=1..nl(i), i=1..ne]
!
  subroutine wrtmf4mt_leg(nou,jou,mat,mt,za,awr,lct,ne,e,ile,na,nl,a)
    implicit real*8 (a-h,o-z)
    character*80 nou(*)
    dimension e(*),ile(*),a(na,*),nl(*)
    dimension nbt(nrmax),ibt(nrmax),b(na)
    mf=4
    ns=0
    ltt=1
    li=0
    c0=1.0d0
    c1=0.0d0
    c2=0.0d0
    l1=0
    l2=0
    n1=0
    n2=0
    call wrtcont(nou,jou,mat,mf,mt,ns,za,awr,l1,ltt,n1,n2)
    call wrtcont(nou,jou,mat,mf,mt,ns,c1,awr,li,lct,n1,n2)
    if (ne.gt.ne4max) then
      write(*,*)' MF4 Warning: ne=',ne,' > ne4max=',ne4max,' ltt=',ltt
    endif
    if (na.gt.ncmax) then
      write(*,*)' MF4 Error: Legendre expansion order greater than ',nlmax,' nl=',na-1,' ltt=',ltt
      stop
    endif
    if (na*ne.gt.npmax) then
      write(*,*)' MF4 Error: na*ne=',na*ne,' > npmax=',npmax,' ltt=',ltt
      stop
    endif
    call packibt(ne,ile,nr,nbt,ibt)
    call wrtab2(nou,jou,mat,mf,mt,ns,c1,c2,l1,l2,nr,nbt,ibt,ne)
    do i=1,ne
      n=nl(i)
      if (n.gt.na) then
        write(*,*)' Error: Legendre expansion order greater than ',na-1,' nl=',n-1,' at E=',e(i)
        stop
      else
        do k=1,n
          b(k)=a(k,i)
        enddo
        call renorm_leg(n,b,c0,f)
        n=n-1
        do k=1,n
          b(k)=b(k+1)
        enddo
        call wrtlist(nou,jou,mat,mf,mt,ns,c1,e(i),l1,l2,n,n2,b)
      endif
    enddo
    call wrtsend(nou,jou,mat,mf,ns)
    return
  end subroutine wrtmf4mt_leg
!
! =================================================================================================================================
! subroutine wrtmf4mt_leg_law: Similar to subroutine wrtmf4mt_leg, but the incident energy
!                              interpolation is described by a single law=ilaw.
!                              The rest of parameters have the same meaning as before
! =================================================================================================================================
!
  subroutine wrtmf4mt_leg_law(nou,jou,mat,mt,za,awr,lct,ne,e,ilaw,na,nl,a)
    implicit real*8 (a-h,o-z)
    character*80 nou(*)
    dimension e(*),a(na,*),nl(*)
    dimension ile(ne)
    do i=1,ne-1
      ile(i)=ilaw
    enddo
    call wrtmf4mt_leg(nou,jou,mat,mt,za,awr,lct,ne,e,ile,na,nl,a)
    return
  end subroutine wrtmf4mt_leg_law
!
! =================================================================================================================================
! subroutine wrtmf4mt_leg_lin: Similar to subroutine wrtmf4mt_leg, but the incident energy
!                              interpolation is described by the single law=2 (lin-lin).
!                              The rest of parameters have the same meaning as before
! =================================================================================================================================
!
  subroutine wrtmf4mt_leg_lin(nou,jou,mat,mt,za,awr,lct,ne,e,na,nl,a)
    implicit real*8 (a-h,o-z)
    character*80 nou(*)
    dimension e(*),a(na,*),nl(*)
    dimension ile(ne)
    do i=1,ne-1
      ile(i)=2
    enddo
    call wrtmf4mt_leg(nou,jou,mat,mt,za,awr,lct,ne,e,ile,na,nl,a)
    return
  end subroutine wrtmf4mt_leg_lin
!
! =================================================================================================================================
! subroutine wrtmf4mt_tab:  Write one MF4 section given in term of tabulated distributions (ltt=2, li=0)
! =================================================================================================================================
!
!     za: ZA number (real*8)
!    awr: Atomic weight ratio to neutron mass (real*8)
!    lct: Reference system (integer)
!     ne: number of incident energies (integer)
!      e: incident energies
!         real*8 array e=[e(i),i=1..ne]
!    ile: interpolation law by incident energy interval
!         integer array ile=[ile(i),1..ne-1]
!     na: Maximum number of cosines (first dimension of arrays xmu, ymu, ilmu)
!    nmu: number of cosine by incident energy (nmu(i)<=na)
!         integer array nmu=[nmu(i),1..ne]
!    xmu: set of cosine values by incident energy
!         real*8 xmu=[xmu(k,i),k=1..nmu(i),i=1..ne]
!    ymu: angle-distribution values by incident energy
!         real*8 ymu=[ymu(k,i),k=1,nmu(i),i=1..ne]
!   ilmu: cosine interpolation law by incident energy
!         integer ilmu=[ilmu(k,i),k=1..nmu(i)-1,i=1..ne]
!
  subroutine wrtmf4mt_tab(nou,jou,mat,mt,za,awr,lct,ne,e,ile,na,nmu,xmu,ymu,ilmu)
    implicit real*8 (a-h,o-z)
    character*80 nou(*)
    dimension e(*),ile(*),nmu(*),xmu(na,*),ymu(na,*),ilmu(na,*)
    dimension nbt(nrmax),ibt(nrmax),jbt(na),x(na),y(na)
    mf=4
    ns=0
    ltt=2
    li=0
    c0=1.0d0
    c1=0.0d0
    c2=0.0d0
    l1=0
    l2=0
    n1=0
    n2=0
    call wrtcont(nou,jou,mat,mf,mt,ns,za,awr,l1,ltt,n1,n2)
    call wrtcont(nou,jou,mat,mf,mt,ns,c1,awr,li,lct,n1,n2)
    if (ne.gt.ne4max) then
      write(*,*)' MF4 Warning: ne=',ne,' > ne4max=',ne4max,' ltt=',ltt
    endif
    if (na.gt.nmu4max) then
      write(*,*)' MF4 Warning: na=',na,' > nmu4max=',nmu4max,' ltt=',ltt
    endif
    if (na*ne.gt.npmax) then
      write(*,*)' MF4 Error: na*ne=',na*ne,' > npmax=',npmax,' ltt=',ltt
      stop
    endif
    call packibt(ne,ile,nr,nbt,ibt)
    call wrtab2(nou,jou,mat,mf,mt,ns,c1,c2,l1,l2,nr,nbt,ibt,ne)
    do i=1,ne
      np=nmu(i)
      if (np.gt.na) then
        write(*,*)' Error: Too many cosines nmu=',np,' na=',na,' at E=',e(i)
        stop
      else
        do k=1,np-1
          x(k)=xmu(k,i)
          y(k)=ymu(k,i)
          jbt(k)=ilmu(k,i)
        enddo
        x(np)=xmu(np,i)
        y(np)=ymu(np,i)
        call renorm_tab(np,x,y,jbt,c0,f)
        call packibt(np,jbt,nr,nbt,ibt)
        call wrtab1(nou,jou,mat,mf,mt,ns,c1,e(i),l1,l2,nr,nbt,ibt,np,x,y)
      endif
    enddo
    call wrtsend(nou,jou,mat,mf,ns)
    return
  end subroutine wrtmf4mt_tab
!
! =================================================================================================================================
! subroutine wrtmf4mt_tab_law: Similar to subroutine wrtmf4mt_tab, but the incident energy
!                              and the angular distribution interpolation at all energies
!                              are described by a single law, ilaw and mulaw respectively.
!                              The rest of parameters have the same meaning as before
! =================================================================================================================================
!
  subroutine wrtmf4mt_tab_law(nou,jou,mat,mt,za,awr,lct,ne,e,ilaw,na,nmu,xmu,ymu,mulaw)
    implicit real*8 (a-h, o-z)
    character*80 nou(*)
    dimension e(*),nmu(*),xmu(na,*),ymu(na,*)
    dimension ile(ne),ilmu(na,ne)
    do i=1,ne
      ile(i)=ilaw
      n=nmu(i)-1
      do j=1,n
        ilmu(j,i)=mulaw
      enddo
    enddo
    call wrtmf4mt_tab(nou,jou,mat,mt,za,awr,lct,ne,e,ile,na,nmu,xmu,ymu,ilmu)
    return
  end subroutine wrtmf4mt_tab_law
!
! =================================================================================================================================
! subroutine wrtmf4mt_tab_lin: Similar to subroutine wrtmf4mt_tab, but the incident energy
!                              and the angular distribution interpolation at all energies
!                              are described by the lin-lin interpolation law.
!                              The rest of parameters have the same meaning as before
! =================================================================================================================================
!
  subroutine wrtmf4mt_tab_lin(nou,jou,mat,mt,za,awr,lct,ne,e,na,nmu,xmu,ymu)
    implicit real*8 (a-h, o-z)
    character*80 nou(*)
    dimension e(*),nmu(*),xmu(na,*),ymu(na,*)
    dimension ile(ne),ilmu(na,ne)
    do i=1,ne
      ile(i)=2
      n=nmu(i)-1
      do j=1,n
        ilmu(j,i)=2
      enddo
    enddo
    call wrtmf4mt_tab(nou,jou,mat,mt,za,awr,lct,ne,e,ile,na,nmu,xmu,ymu,ilmu)
    return
  end subroutine wrtmf4mt_tab_lin
!
! =================================================================================================================================
! subroutine wrtmf4_mix: Write a mixed (Legrendre/Tabulated) MF4 section (ltt=3,li=0)
! =================================================================================================================================
!
!     1. Legendre expansion at lower energies
!     2. Probability distributions at higher energies
!
!     za: ZA number (real*8)
!    awr: Atomic weight ratio to neutron mass (real*8)
!    lct: Reference system (integer)
!    nel: number of incident energies in the lower range (integer)
!     el: incident energies in the lower energy range
!         real*8 array el=[el(i),i=1..nel]
!   ilel: interpolation law by incident energy interval
!         integer array ilel=[ilel(i),1..nel-1]
!    nal: Maximun number of Legendre coefficients nal<=ncmax (first dimension of array a)
!     nl: Number of Legrende coefficient by incident lower energy
!         integer array nl=[nl(i),i=1..nel]
!         (max(nl(i)) cannot be greater than nal)
!      a: Legendre coefficients
!         real*8 array a=[a(l,i), l=1..nl(i), i=1..nel]
!    neh: number of incident energies in the higher energy range (integer)
!     eh: incident energies in the higher energy range
!         real*8 array eh=[eh(i),i=1..neh]
!   ileh: interpolation law by incident energy interval
!         integer array ileh=[ileh(i),1..neh-1]
!    nah: Maximum number of cosines (first dimension of arrays xmu, ymu, ilmu)
!    nmu: number of cosine by incident energy
!         integer array nmu=[nmu(i),1..neh]
!         (max(nmu(i)) cannot be greater than nah)
!    xmu: set of cosine values by incident energy
!         real*8 xmu=[xmu(k,i),k=1..nmu(i),i=1..neh]
!    ymu: angle-distribution values by incident energy
!         real*8 ymu=[ymu(k,i),k=1,nmu(i),i=1..neh]
!   ilmu: cosine interpolation law by incident energy
!         integer ilmu=[ilmu(k,i),k=1..nmu(i)-1,i=1..neh]
!
  subroutine wrtmf4_mix(nou,jou,mat,mt,za,awr,lct,nel,el,ilel,nal,nl,a,neh,eh,ileh,nah,nmu,xmu,ymu,ilmu)
    implicit real*8 (a-h, o-z)
    parameter (eps=1.0d-6)
    character*80 nou(*)
    dimension el(*),ilel(*),nl(*),a(nal,*)
    dimension eh(*),ileh(*),nmu(*),xmu(nah,*),ymu(nah,*),ilmu(nah,*)
    dimension nbt(nrmax),ibt(nrmax),jbt(nah),x(nah),y(nah)
    if (abs(eh(1)-el(nel)).gt.(eps*el(nel))) then
      write(*,*)' MF4 Warning: Energy discontinuity between Legendre expansion and Tabulated data'
    endif
    eh(1)=el(nel)
    mf=4
    ns=0
    ltt=3
    li=0
    c0=1.0d0
    c1=0.0d0
    c2=0.0d0
    l1=0
    l2=0
    n1=0
    n2=0
    call wrtcont(nou,jou,mat,mf,mt,ns,za,awr,l1,ltt,n1,n2)
    call wrtcont(nou,jou,mat,mf,mt,ns,c1,awr,li,lct,n1,n2)
    if ((nel+neh).gt.ne4max) then
      write(*,*)' MF4 Warning: ne=',nel+neh,' > ne4max=',ne4max,' ltt=',ltt
    endif
    if (nal.gt.ncmax) then
      write(*,*)' MF4 Error: Legendre expansion order greater than ',nlmax,' nl=',nal-1,' ltt=',ltt
      stop
    endif
    if (nal*nel.gt.npmax) then
      write(*,*)' MF4 Error: nal*nel=',nal*nel,' > npmax=',npmax,' ltt=',ltt
      stop
    endif
    call packibt(nel,ilel,nr,nbt,ibt)
    call wrtab2(nou,jou,mat,mf,mt,ns,c1,c2,l1,l2,nr,nbt,ibt,nel)
    do i=1,nel
      na=nl(i)
      if (na.gt.nal) then
        write(*,*)' Error: Legendre expansion order greater than ',nal-1,' nl=',na-1,' at E=',el(i)
        stop
      else
        do k=1,na
          y(k)=a(k,i)
        enddo
        call renorm_leg(na,y,c0,f)
        na=na-1
        do k=1,na
          y(k)=y(k+1)
        enddo
        call wrtlist(nou,jou,mat,mf,mt,ns,c1,el(i),l1,l2,na,n2,y)
      endif
    enddo
    if (nah.gt.nmu4max) then
      write(*,*)' MF4 Warning: Too many cosine values ',nah,' nmu4max=',nmu4max,' ltt=',ltt
    endif
    if (nah*neh.gt.npmax) then
      write(*,*)' MF4 Error: nah*neh=',nah*neh,' > npmax=',npmax,' ltt=',ltt
      stop
    endif
    call packibt(neh,ileh,nr,nbt,ibt)
    call wrtab2(nou,jou,mat,mf,mt,ns,c1,c2,l1,l2,nr,nbt,ibt,neh)
    do i=1,neh
      np=nmu(i)
      if (np.gt.nah) then
        write(*,*)' MF4 Error: Too many cosines nmu=',np,' nah=',nah
        stop
      else
        do k=1,np-1
          x(k)=xmu(k,i)
          y(k)=ymu(k,i)
          jbt(k)=ilmu(k,i)
        enddo
        x(np)=xmu(np,i)
        y(np)=ymu(np,i)
        call renorm_tab(np,x,y,jbt,c0,f)
        call packibt(np,jbt,nr,nbt,ibt)
        call wrtab1(nou,jou,mat,mf,mt,ns,c1,eh(i),l1,l2,nr,nbt,ibt,np,x,y)
      endif
    enddo
    call wrtsend(nou,jou,mat,mf,ns)
    return
  end subroutine wrtmf4_mix
!
! =================================================================================================================================
! subroutine wrtmf4_mix_law: Similar to subroutine wrtmf4_mix, but using a single interpolation law
!                            for each of the following parameters:
!                             1. Lower incident energy for Legendre expansion (lawl)
!                             2. Higher incident energy for Tabulated data (lawh)
!                             3. Angular distributions at all energies for tabulated data (lawmu)
!                            The rest of parameters have the same meaning as before.
! =================================================================================================================================
!
  subroutine wrtmf4_mix_law(nou,jou,mat,mt,za,awr,lct,nel,el,lawl,nal,nl,a,neh,eh,lawh,nah,nmu,xmu,ymu,lawmu)
    implicit real*8 (a-h, o-z)
    character*80 nou(*)
    dimension el(*),nl(*),a(nal,*)
    dimension eh(*),nmu(*),xmu(nah,*),ymu(nah,*)
    dimension ilel(nel),ileh(neh),ilmu(nah,neh)
    do i=1,nel-1
      ilel(i)=lawl
    enddo
    do i=1,neh
      ileh(i)=lawh
      n=nmu(i)-1
      do j=1,n
        ilmu(j,i)=lawmu
      enddo
    enddo
    call wrtmf4_mix(nou,jou,mat,mt,za,awr,lct,nel,el,ilel,nal,nl,a,neh,eh,ileh,nah,nmu,xmu,ymu,ilmu)
    return
  end subroutine wrtmf4_mix_law
!
! =================================================================================================================================
! subroutine wrtmf4_mix_lin: Similar to subroutine wrtmf4_mix, but using lin-lin interpolation
!                            each of the following parameters
!                             1. Lower incident energy for Legendre expansion (lawl)
!                             2. Higher incident energy for Tabulated data (lawh)
!                             3. Angular distributions at all energies for tabulated data (lawmu)
!                            The rest of parameters have the same meaning as before.
! =================================================================================================================================
!
  subroutine wrtmf4_mix_lin(nou,jou,mat,mt,za,awr,lct,nel,el,nal,nl,a,neh,eh,nah,nmu,xmu,ymu)
    implicit real*8 (a-h, o-z)
    character*80 nou(*)
    dimension el(*),nl(*),a(nal,*)
    dimension eh(*),nmu(*),xmu(nah,*),ymu(nah,*)
    dimension ilel(nel),ileh(neh),ilmu(nah,neh)
    do i=1,nel-1
      ilel(i)=2
    enddo
    do i=1,neh
      ileh(i)=2
      n=nmu(i)-1
      do j=1,n
        ilmu(j,i)=2
      enddo
    enddo
    call wrtmf4_mix(nou,jou,mat,mt,za,awr,lct,nel,el,ilel,nal,nl,a,neh,eh,ileh,nah,nmu,xmu,ymu,ilmu)
    return
  end subroutine wrtmf4_mix_lin
!
! =================================================================================================================================
!
! MF5: Procedures for reading and writing ENDF-6/MF5/MT sub-sections:
! ------------------------------------------------------------------
!
! =================================================================================================================================
! subroutine readmf5mt_head: Read head record of MF5/MT to get za, awr, and the number of sub-sections nk
! =================================================================================================================================
!
!    za: ZA number of the target (real*8)
!   awr: Atomic weight ratio of the target (real*8)
!    nk: number of sub-sections in a section (integer)
!
  subroutine readmf5mt_head(nin,jin,za,awr,nk)
    implicit real*8 (a-h, o-z)
    character*80 nin(*)
    call readcont(nin,jin,za,awr,l1,l2,nk,n2,mat,mf,mt,ns)
    if (nk.gt.nkmax) then
      write(*,*)' MF5 Warning: Number of subsection greater than nkmax=',nkmax
    endif
    return
  end subroutine readmf5mt_head
!
! =================================================================================================================================
! subroutine mf5mtsub_form: Get general information about all sub-sections of one MF5/MT section (energy distribution)
! =================================================================================================================================
!
!   fave(E' <- E) = fave(E',E) = p1(E)*f1(E',E)+p2(E)*f2(E',E)+ .. +pnk(E)*fnk(E',E)
!
!   where  p1(E)+p2(E)+ .. +pnk(E)=1.0
!
!    nk: number of sub-sections in a section (integer)
!    lf: array containing the types of sub-sections
!        integer array lf=[lf(i),i=1..nk]
!          lf(i)= 1: Arbitrary tabulated Function
!          lf(i)= 5: General Evaporation Spectrum
!          lf(i)= 7: Simple Maxwellian Fission Spectrum
!          lf(i)= 9: Evaporation Spectrum
!          lf(i)=11: Energy-dependent Watt Spectrum
!          lf(i)=12: Energy-dependent Fission Neutron Spectrum (Madland and Nix spectrum)
!   jlf: array containing the pointer to the first line of each sub-section in the input array nin(*)
!        integer array jlf=[jlf(i),i=1..nk]
!        the first line of sub-section i is located at nin(jlf(i))
!    np: array containing the number of incident energy points at which the probability p(E) is given by sub-section
!        integer array np=[np(i),i=1..nk]
!    ne: array containing the number of incident energy points at which the distribution data are given by sub-section
!        integer array ne=[ne(i),i=1..nk]
!   nep: array containing the maximum number of outgoing energies (or second incident energy grid for lf=11) by sub-section
!        integer array nep=[nep(i),i=1..nk]
!        for lf(i)=11(Energy dependent Watt Spectrum) nep(i) contains the number of incident energies for the parameter b(E)
!        nep(i)=0 means that this dimension is not needed for the i-th representation
!
  subroutine mf5mtsub_form(nin,jin,nk,lf,jlf,np,ne,nep)
    implicit real*8 (a-h, o-z)
    character*80 nin(*)
    dimension lf(*),jlf(*),np(*),ne(*),nep(*)
    dimension x(ne5max),y(ne5max),x1(nep5max),y1(nep5max)
    dimension nbt(nrmax),ibt(nrmax)
    jin0=jin
    do k=1,nk
      jlf(k)=jin
      call readtab1(nin,jin,c1,c2,l1,lfk,nr,nbt,ibt,npk,x,y)
      if (npk.gt.ne5max) then
        write(*,*)' MF5 Warning: Number of incident energy points for p(E) equal to ',npk,' > ne5max=',ne5max,' lf=',lfk
      endif
      if (nr.gt.nrmax) then
        write(*,*)' MF5 Error: Number of interpolation ranges for p(E) equal to ',nr,' > nrmax=',nrmax,' lf=',lfk
        stop
      endif
      lf(k)=lfk
      np(k)=npk
      if (lfk.eq.1) then
        call readtab2(nin,jin,c1,c2,l1,l2,nr,nbt,ibt,nek)
        if (nek.gt.ne5max) then
          write(*,*)' MF5 Warning: Incident energy points for lf=1 equal to ',nek,' > ne5max=',ne5max
        endif
        if (nr.gt.nrmax) then
          write(*,*)' MF5 Error: Incident energy interpolation ranges for lf=1 equal to ',nr,' > nrmax=',nrmax
          stop
        endif
        ne(k)=nek
        nepk=0
        do i=1,nek
          call readtab1(nin,jin,c1,c2,l1,l2,nr,nbt,ibt,nepi,x1,y1)
          if (nepi.gt.nep5max) then
            write(*,*)' MF5 Warning: Outgoing energy points for lf=1 equal to ',nepi,' > nep5max=',nep5max,' at E=',c2
          endif
          if (nr.gt.nrmax) then
            write(*,*)' MF5 Error: Outgoing energy interpolation ranges for lf=1 equal to ',nr,' > nrmax=',nrmax,' at E=',c2
            stop
          endif
          if (nepi.gt.nepk) nepk=nepi
        enddo
        nep(k)=nepk
      elseif (lfk.eq.5) then
        call readtab1(nin,jin,c1,c2,l1,l2,nr,nbt,ibt,nek,x,y)
        if (nek.gt.ne5max) then
          write(*,*)' MF5 Warning: Incident energy points for lf=5 equal to ',nek,' > ne5max=',ne5max
        endif
        if (nr.gt.nrmax) then
          write(*,*)' MF5 Error: Incident energy interpolation ranges for lf=5 equal to ',nr,' > nrmax=',nrmax
          stop
        endif
        ne(k)=nek
        call readtab1(nin,jin,c1,c2,l1,l2,nr,nbt,ibt,nepk,x1,y1)
        if (nepk.gt.nep5max) then
          write(*,*)' MF5 Warning: Outgoing energy points for lf=5 equal to ',nepk,' > nep5max=',nep5max
        endif
        if (nr.gt.nrmax) then
          write(*,*)' MF5 Error: Outgoing energy interpolation ranges for lf=5 equal to ',nr,' > nrmax=',nrmax
          stop
        endif
        nep(k)=nepk
      elseif (lfk.eq.7.or.lfk.eq.9.or.lfk.eq.12) then
        call readtab1(nin,jin,c1,c2,l1,l2,nr,nbt,ibt,nek,x,y)
        if (nek.gt.ne5max) then
          write(*,*)' MF5 Warning: Incident energy points for lf=',lfk,' equal to ',nek,' > ne5max=',ne5max
        endif
        if (nr.gt.nrmax) then
          write(*,*)' MF5 Error: Incident energy interpolation ranges for lf=',lfk,' equal to ',nr,' > nrmax=',nrmax
          stop
        endif
        ne(k)=nek
        nep(k)=0
      elseif (lfk.eq.11) then
        call readtab1(nin,jin,c1,c2,l1,l2,nr,nbt,ibt,nek,x,y)
        if (nek.gt.ne5max) then
          write(*,*)' MF5 Warning: Incident energy points for lf=11 a(E) equal to ',nek,' > ne5max=',ne5max
        endif
        if (nr.gt.nrmax) then
          write(*,*)' MF5 Error: Incident energy interpolation ranges for lf=11 a(E) equal to ',nr,' > nrmax=',nrmax
          stop
        endif
        ne(k)=nek
        call readtab1(nin,jin,c1,c2,l1,l2,nr,nbt,ibt,nepk,x,y)
        if (nepk.gt.ne5max) then
          write(*,*)' MF5 Warning: Incident energy points for lf=11 b(E) equal to ',nek,' > ne5max=',ne5max
        endif
        if (nr.gt.nrmax) then
          write(*,*)' MF5 Error: Incident energy interpolation ranges for lf=11 b(E) equal to ',nr,' > nrmax=',nrmax
          stop
        endif
        nep(k)=nepk
      else
        write(*,*)' MF5 Error: Incorrect representation lf=',lfk
        stop
      endif
    enddo
    jin=jin0
    return
  end subroutine mf5mtsub_form
!
! =================================================================================================================================
! subroutine readmf5_tabular: Read an arbitrary tabulated spectrum (lf=1 MF5/MT subsection)
! =================================================================================================================================
!
!   fk(E' <- E) = fk(E',E) = p(E)*f(E',E)
!
!      np: number of incident energy points at which p(E) is given (integer)
!      ep: incident energy grid for p(E)
!          real*8 array ep=[ep(i),i=1..np]
!       p: fractional probability p(E) of the partial distribution
!          real*8 array p=[p(i),i=1..np]
!    ilep: (p(E),E) interpolation law by intervals
!          integer array ilep=[ilep(i),i=1..np-1]
!      ne: number of incident energy points at which f(E',E) is given (integer)
!       e: incident energy grid for f(E',E)
!          real*8 array e=[e(i),i=1..ne]
!     ile: incident energy interpolation law by intervals
!          integer array ile=[ile(i),i=1..ne-1]
!     nep: maximun number of outgoin energies (integer)
!      nf: number of outgoing energy points at which f(E',E) is given (max(nf) <= nep)
!          integer array nf=[nf(i),i=1..ne]
!      ef: outgoing energy grid at E for f(E',E)
!          real*8 array ef=[ef(j,i),j=1..nf(i),i=1..ne]
!       f: tabulated energy distribution f(E',E)
!          real*8 array f=[f(j,i),j=1..nf(i),i=1..ne]
!    ilef: (f(E',E),E') interpolation law by E' intervals at E
!          integer array ilef=[ilef(j,i),j=1..nf(i)-1,i=1..ne]
!
  subroutine readmf5_tabular(nin,jin,np,ep,p,ilep,ne,e,ile,nep,nf,ef,f,ilef)
    implicit real*8 (a-h, o-z)
    character*80 nin(*)
    dimension ep(*),p(*),ilep(*),e(*),ile(*),nf(*),ef(nep,*),f(nep,*),ilef(nep,*)
    dimension nbt(nrmax),ibt(nrmax),x(nep),y(nep),jbt(nep)
    call readtab1(nin,jin,c1,c2,l1,lf,nr,nbt,ibt,np,ep,p)
    call unpackibt(nr,nbt,ibt,np,ilep)
    call readtab2(nin,jin,c1,c2,l1,l2,nr,nbt,ibt,ne)
    call unpackibt(nr,nbt,ibt,ne,ile)
    do i=1,ne
      call readtab1(nin,jin,c1,e(i),l1,l2,nr,nbt,ibt,nfi,x,y)
      call unpackibt(nr,nbt,ibt,nfi,jbt)
      nf(i)=nfi
      do j=1,nfi-1
        ef(j,i)=x(j)
        f(j,i)=y(j)
        ilef(j,i)=jbt(j)
      enddo
      ef(nfi,i)=x(nfi)
      f(nfi,i)=y(nfi)
    enddo
    return
  end subroutine readmf5_tabular
!
! =================================================================================================================================
! subroutine readmf5_gen_eva: Read a general evaporation spectrum (lf=5 MF5/MT subsection)
! =================================================================================================================================
!
!   fk(E' <- E) = fk(E',E) = p(E)*f(x),  where x=E'/theta(E) with E' in the interval [0,E-u]
!
!      np: number of incident energy points at which p(E) is given
!      ep: incident energy grid for p(E)
!          real*8 array ep=[ep(i),i=1..np]
!       p: fractional probability p(E) of the partial distribution
!          real*8 array p=[p(i),i=1..np]
!    ilep: (p(E),E) interpolation law by intervals
!          integer array ilep=[ilep(i),i=1..np-1]
!      ne: number of incident energy points at which theta(E) is given
!       e: incident energy grid for theta(E)
!          real*8 array e=[e(i),i=1..ne]
!   theta: Effective temperature theta(E) in eV
!          real*8 array theta=[theta(i),i=1..ne]
!     ile: (theta(E),E) interpolation law by intervals
!          integer array ile=[ile(i),i=1..ne-1]
!      nf: number of points at which f(x) is given
!       x: x grid for g(x)
!          real*8 array ef=[ef(j),j=1..nf]
!       f: general evaporation spectrum f(x)
!          real*8 array f=[f(j),j=1..nf]
!     ilx: (f(x),x) interpolation law by E' intervals at E
!          integer array ilx=[ilx(j),j=1..nf-1]
!       u: Constant that defines the upper energy limit for E' (real*8)
!          E' lies in the closed interval [0,E-u] if u=0, then max(E')=E
!
  subroutine readmf5_gen_eva(nin,jin,np,ep,p,ilep,ne,e,theta,ile,nf,x,f,ilx,u)
    implicit real*8 (a-h, o-z)
    character*80 nin(*)
    dimension ep(*),p(*),ilep(*),e(*),theta(*),ile(*),x(*),f(*),ilx(*)
    dimension nbt(nrmax),ibt(nrmax)
    call readtab1(nin,jin,u,c2,l1,lf,nr,nbt,ibt,np,ep,p)
    call unpackibt(nr,nbt,ibt,np,ilep)
    call readtab1(nin,jin,c1,c2,l1,l2,nr,nbt,ibt,ne,e,theta)
    call unpackibt(nr,nbt,ibt,ne,ile)
    call readtab1(nin,jin,c1,c2,l1,l2,nr,nbt,ibt,nf,x,f)
    call unpackibt(nr,nbt,ibt,nf,ilx)
    return
  end subroutine readmf5_gen_eva
!
! =================================================================================================================================
! subroutine readmf5_maxwell: Read a Maxwellian fission spectrum (lf=7 MF5/MT subsection)
! =================================================================================================================================
!
!    fk(E' <- E) = fk(E',E) = p(E)*(1/I)*sqrt(E')*exp(E'/theta(E))
!
!    where I=c(E,theta(E),u) is a normalization constant and E' is within the interval [0,E-u]
!
!      np: number of incident energy points at which p(E) is given (integer)
!      ep: incident energy grid for p(E)
!          real*8 array ep=[ep(i),i=1..np]
!       p: fractional probability p(E) of the partial distribution
!          real*8 array p=[p(i),i=1..np]
!    ilep: (p(E),E) interpolation law by intervals
!          integer array ilep=[ilep(i),i=1..np-1]
!      ne: number of incident energy points at which theta(E) is given (integer)
!       e: incident energy grid for theta(E)
!          real*8 array e=[e(i),i=1..ne]
!   theta: Effective temperature theta(E) in eV
!          real*8 array theta=[theta(i),i=1..ne]
!     ile: (theta(E),E) interpolation law by intervals
!          integer array ile=[ile(i),i=1..ne-1]
!       u: Constant that defines the upper energy limit for E' (real*8)
!          E' lies in the closed interval [0,E-u] if u=0, then max(E')=E
!
  subroutine readmf5_maxwell(nin,jin,np,ep,p,ilep,ne,e,theta,ile,u)
    implicit real*8 (a-h, o-z)
    character*80 nin(*)
    dimension ep(*),p(*),ilep(*),e(*),theta(*),ile(*)
    dimension nbt(nrmax),ibt(nrmax)
    call readtab1(nin,jin,u,c2,l1,lf,nr,nbt,ibt,np,ep,p)
    call unpackibt(nr,nbt,ibt,np,ilep)
    call readtab1(nin,jin,c1,c2,l1,l2,nr,nbt,ibt,ne,e,theta)
    call unpackibt(nr,nbt,ibt,ne,ile)
    return
  end subroutine readmf5_maxwell
!
! =================================================================================================================================
! subroutine readmf5_eva: Read an evaporation spectrum (lf=9 MF5/MT subsection)
! =================================================================================================================================
!
!    fk(E' <- E) = fk(E',E) = p(E)*(1/I)*E'*exp(E'/theta(E))
!
!    where I=c(E,theta(E),u) normalization constant and E' is within in the interval [0,E-u]
!
!      np: number of incident energy points at which p(E) is given
!      ep: incident energy grid for p(E)
!          real*8 array ep=[ep(i),i=1..np]
!       p: fractional probability p(E) of the partial distribution
!          real*8 array p=[p(i),i=1..np]
!    ilep: (p(E),E) interpolation law by intervals
!          integer array ilep=[ilep(i),i=1..np-1]
!      ne: number of incident energy points at which theta(E) is given
!       e: incident energy grid for theta(E)
!          real*8 array e=[e(i),i=1..ne]
!   theta: Effective temperature theta(E) in eV
!          real*8 array theta=[theta(i),i=1..ne]
!     ile: (theta(E),E) interpolation law by intervals
!          integer array ile=[ile(i),i=1..ne-1]
!       u: Constant that defines the upper energy limit for E' (real*8)
!          E' lies in the closed interval [0,E-u] if u=0, then max(E')=E
!
  subroutine readmf5_eva(nin,jin,np,ep,p,ilep,ne,e,theta,ile,u)
    implicit real*8 (a-h, o-z)
    character*80 nin(*)
    dimension ep(*),p(*),ilep(*),e(*),theta(*),ile(*)
    call readmf5_maxwell(nin,jin,np,ep,p,ilep,ne,e,theta,ile,u)
    return
  end subroutine readmf5_eva
!
! =================================================================================================================================
! subroutine readmf5_watt: Read energy-dependent Watt spectrum (lf=11 MF5/MT subsection)
! =================================================================================================================================
!
!   fk(E' <- E) = fk(E',E) = p(E)*(1/I)*exp(E'/a(E))*sinh(sqrt(b(E)*E'))
!
!   where I=c(E,a(E),b(E),u) normalization constant and E' is within in the interval [0,E-u]
!
!     np: number of incident energy points at which p(E) is given (integer)
!     ep: incident energy grid for p(E)
!         real*8 array ep=[ep(i),i=1..np]
!      p: fractional probability p(E) of the partial distribution
!         real*8 array p=[p(i),i=1..np]
!   ilep: (p(E),E) interpolation law by intervals
!         integer array ilep=[ilep(i),i=1..np-1]
!    nea: number of incident energy points at which a(E) is given (integer)
!     ea: incident energy grid for a(E)
!         real*8 array ea=[ea(i),i=1..nea]
!      a: Effective parameter a(E) in eV
!         real*8 array a=[a(i),i=1..nea]
!   ilea: (a(E),E) interpolation law by intervals
!         integer array ile=[ilea(i),i=1..nea-1]
!    neb: number of incident energy points at which b(E) is given (integer)
!     eb: incident energy grid for b(E)
!         real*8 array eb=[eb(i),i=1..neb]
!      b: Effective parameter b(E) in 1/eV
!         real*8 array b=[b(i),i=1..neb]
!   ileb: (b(E),E) interpolation law by intervals
!         integer array ile=[ileb(i),i=1..neb-1]
!      u: Constant that defines the upper energy limit for E' (real*8)
!         E' lies in the closed interval [0,E-u] if u=0, then max(E')=E
!
  subroutine readmf5_watt(nin,jin,np,ep,p,ilep,nea,ea,a,ilea,neb,eb,b,ileb,u)
    implicit real*8 (a-h, o-z)
    character*80 nin(*)
    dimension ep(*),p(*),ilep(*),ea(*),a(*),ilea(*),eb(*),b(*),ileb(*)
    call readmf5_gen_eva(nin,jin,np,ep,p,ilep,nea,ea,a,ilea,neb,eb,b,ileb,u)
    return
  end subroutine readmf5_watt
!
! =================================================================================================================================
! subroutine readmf5_madland_nix: Read Madland and Nix fission neutron spectrum (lf=12 MF5/MT subsection)
! =================================================================================================================================
!
!   fk(E' <- E) = fk(E',E) = p(E)*0.5*(g(E',tm(E),efl)+g(E',tm(E),efh))
!
!   where g(E',tm,Ef) is an analytical function given by Madland and Nix, and E' is whitin the interval [0, infinity)
!
!     np: number of incident energy points at which p(E) is given (integer)
!     ep: incident energy grid for p(E)
!         real*8 array ep=[ep(i),i=1..np]
!      p: fractional probability p(E) of the partial distribution
!         real*8 array p=[p(i),i=1..np]
!   ilep: (p(E),E) interpolation law by intervals
!         integer array ilep=[ilep(i),i=1..np-1]
!     ne: number of incident energy points at which tm(E) is given
!      e: incident energy grid for tm(E)
!         real*8 array e=[e(i),i=1..ne]
!     tm: Effective parameter tm(E) in eV
!         real*8 array tm=[tm(i),i=1..ne]
!    ile: (tm(E),E) interpolation law by intervals
!         integer array ile=[ile(i),i=1..ne-1]
!    efl: average kinetic energy per nucleon of the average
!         light fission fragment (real*8)
!    efh: average kinetic energy per nucleon of the average
!         heavy fission fragment (real*8)
!
  subroutine readmf5_madland_nix(nin,jin,np,ep,p,ilep,ne,e,tm,ile,efl,efh)
    implicit real*8 (a-h, o-z)
    character*80 nin(*)
    dimension ep(*),p(*),ilep(*),e(*),tm(*),ile(*)
    dimension nbt(nrmax),ibt(nrmax)
    call readtab1(nin,jin,c1,c2,l1,lf,nr,nbt,ibt,np,ep,p)
    call unpackibt(nr,nbt,ibt,np,ilep)
    call readtab1(nin,jin,efl,efh,l1,l2,nr,nbt,ibt,ne,e,tm)
    call unpackibt(nr,nbt,ibt,ne,ile)
    return
  end subroutine readmf5_madland_nix
!
! =================================================================================================================================
! subroutine wrtmf5mt_head: Write MF5/MT section head
! =================================================================================================================================
!
!   mat: material number (integer)
!    mt: section number (integer)
!    ns: ENDF-6 sequence number (integer)
!    za: ZA number (real*8)
!   awr: Atomic weigth ratio (real*8)
!    nk: number of subsection to write for this section (integer)
!
  subroutine wrtmf5mt_head(nou,jou,mat,mt,ns,za,awr,nk)
    implicit real*8 (a-h, o-z)
    character*80 nou(*)
    mf=5
    l1=0
    l2=0
    n2=0
    if (nk.gt.nkmax) then
      write(*,*)' MF5 Error: Too many sub-sections for MF5/MT=',mt,' nk=',nk,' nkmax=',nkmax
      stop
    else
      call wrtcont(nou,jou,mat,mf,mt,ns,za,awr,l1,l2,nk,n2)
      return
    endif
  end subroutine wrtmf5mt_head
!
! =================================================================================================================================
! subroutine wrtmf5_tabular: Write an arbitrary tabulated spectrum (lf=1 MF5/MT subsection)
! =================================================================================================================================
!
!   fk(E' <- E) = fk(E',E) = p(E)*f(E',E)
!
!      np: number of incident energy points at which p(E) is given (integer)
!      ep: incident energy grid for p(E)
!          real*8 array ep=[ep(i),i=1..np]
!       p: fractional probability p(E) of the partial distribution
!          real*8 array p=[p(i),i=1..np]
!    ilep: (p(E),E) interpolation law by intervals
!          integer array ilep=[ilep(i),i=1..np-1]
!      ne: number of incident energy points at which f(E',E) is given (integer)
!       e: incident energy grid for f(E',E)
!          real*8 array e=[e(i),i=1..ne]
!     ile: incident energy interpolation law by intervals
!          integer array ile=[ile(i),i=1..ne-1]
!     nep: maximun number of outgoin energies (integer)
!      nf: number of outgoing energy points at which f(E',E) is given (max(nf) <= nep)
!          integer array nf=[nf(i),i=1..ne]
!      ef: outgoing energy grid at E for f(E',E)
!          real*8 array ef=[ef(j,i),j=1..nf(i),i=1..ne]
!       f: tabulated energy distribution f(E',E)
!          real*8 array f=[f(j,i),j=1..nf(i),i=1..ne]
!    ilef: (f(E',E),E') interpolation law by E' intervals at E
!          integer array ilef=[ilef(j,i),j=1..nf(i)-1,i=1..ne]
!
  subroutine wrtmf5_tabular(nou,jou,mat,mt,ns,np,ep,p,ilep,ne,e,ile,nep,nf,ef,f,ilef)
    implicit real*8 (a-h,o-z)
    character*80 nou(*)
    dimension ep(*),p(*),ilep(*),e(*),ile(*),nf(*),ef(nep,*),f(nep,*),ilef(nep,*)
    dimension nbt(nrmax),ibt(nrmax),x(nep),y(nep),jbt(nep)
    if (np.gt.ne5max) then
      write(*,*)' MF5/lf=1 Warning: Too many incident energies for p(E) np=',np,' ne5max=',ne5max
    endif
    if (ne.gt.ne5max) then
      write(*,*)' MF5/lf=1 Warning: Too many incident energies for p(E) np=',np,' ne5max=',ne5max
    endif
    if (nep.gt.nep5max) then
      write(*,*)' MF5/lf=1 Warning: Too many incident energies for p(E) np=',np,' ne5max=',ne5max
    endif
    if (ne*nep.gt.npmax.or.np.gt.npmax) then
      write(*,*)' MF5/lf=1 Error: Too many array elements. np=',np,' ne*nep=',ne*nep,' npmax=',npmax
      stop
    endif
    mf=5
    lf=1
    c1=0.0d0
    c2=0.0d0
    l1=0
    l2=0
    n1=0
    n2=0
    call packibt(np,ilep,nr,nbt,ibt)
    call wrtab1(nou,jou,mat,mf,mt,ns,c1,c2,l1,lf,nr,nbt,ibt,np,ep,p)
    call packibt(ne,ile,nr,nbt,ibt)
    call wrtab2(nou,jou,mat,mf,mt,ns,c1,c2,l1,l2,nr,nbt,ibt,ne)
    do i=1,ne
      ei=e(i)
      nfi=nf(i)
      if (nfi.gt.nep) then
        write(*,*)' MF5/lf=1 Error: Too many outgoing energies nf=',nfi,' at E=',ei,' nep=',nep
        stop
      endif
      do j=1,nfi-1
        x(j)=ef(j,i)
        y(j)=f(j,i)
        jbt(j)=ilef(j,i)
      enddo
      x(nfi)=ef(nfi,i)
      y(nfi)=f(nfi,i)
      call packibt(nfi,jbt,nr,nbt,ibt)
      call wrtab1(nou,jou,mat,mf,mt,ns,c1,ei,l1,l2,nr,nbt,ibt,nfi,x,y)
    enddo
    return
  end subroutine wrtmf5_tabular
!
! =================================================================================================================================
! subroutine wrtmf5_tabular_law: Similar to subroutine wrtmf5_tabular, but each interpolation scheme
!                                is represented by only one interpolation law
! =================================================================================================================================
!
!   ilepk: interpolation law for [E, p(E)]
!    ilek: interpolation law for [E,f(E',E)]
!   ilefk: interpolation law for [E',f(E',E)]
!   The rest of parameters have the same meaning as before
!
  subroutine wrtmf5_tabular_law(nou,jou,mat,mt,ns,np,ep,p,ilepk,ne,e,ilek,nep,nf,ef,f,ilefk)
    implicit real*8 (a-h,o-z)
    character*80 nou(*)
    dimension ep(*),p(*),e(*),nf(*),ef(nep,*),f(nep,*)
    dimension ilep(np),ile(ne),ilef(nep,ne)
    do i=1,np-1
      ilep(i)=ilepk
    enddo
    do i=1,ne
      ile(i)=ilek
      nfi=nf(i)-1
      do j=1,nfi
        ilef(j,i)=ilefk
      enddo
    enddo
    call wrtmf5_tabular(nou,jou,mat,mt,ns,np,ep,p,ilep,ne,e,ile,nep,nf,ef,f,ilef)
    return
  end subroutine wrtmf5_tabular_law
!
! =================================================================================================================================
! subroutine wrtmf5_tabular_lin: Similar to subroutine wrtmf5_tabular, but all interpolation schema
!                                are lin-lin, therefore they are not input parameters.
!                                The rest of parameters have the same meaning as before
! =================================================================================================================================
!
  subroutine wrtmf5_tabular_lin(nou,jou,mat,mt,ns,np,ep,p,ne,e,nep,nf,ef,f)
    implicit real*8 (a-h,o-z)
    character*80 nou(*)
    dimension ep(*),p(*),e(*),nf(*),ef(nep,*),f(nep,*)
    dimension ilep(np),ile(ne),ilef(nep,ne)
    do i=1,np-1
      ilep(i)=2
    enddo
    do i=1,ne
      ile(i)=2
      nfi=nf(i)-1
      do j=1,nfi
        ilef(j,i)=2
      enddo
    enddo
    call wrtmf5_tabular(nou,jou,mat,mt,ns,np,ep,p,ilep,ne,e,ile,nep,nf,ef,f,ilef)
    return
  end subroutine wrtmf5_tabular_lin
!
! =================================================================================================================================
! subroutine wrtmf5_gen_eva: Write a general evaporation spectrum (lf=5 MF5/MT subsection)
! =================================================================================================================================
!
!   fk(E' <- E) = fk(E',E) = p(E)*f(x),  where x=E'/theta(E) with E' in the interval [0,E-u]
!
!      np: number of incident energy points at which p(E) is given
!      ep: incident energy grid for p(E)
!          real*8 array ep=[ep(i),i=1..np]
!       p: fractional probability p(E) of the partial distribution
!          real*8 array p=[p(i),i=1..np]
!    ilep: (p(E),E) interpolation law by intervals
!          integer array ilep=[ilep(i),i=1..np-1]
!      ne: number of incident energy points at which theta(E) is given
!       e: incident energy grid for theta(E)
!          real*8 array e=[e(i),i=1..ne]
!   theta: Effective temperature theta(E) in eV
!          real*8 array theta=[theta(i),i=1..ne]
!     ile: (theta(E),E) interpolation law by intervals
!          integer array ile=[ile(i),i=1..ne-1]
!      nf: number of points at which f(x) is given
!       x: x grid for g(x)
!          real*8 array ef=[ef(j),j=1..nf]
!       f: general evaporation spectrum f(x)
!          real*8 array f=[f(j),j=1..nf]
!     ilx: (f(x),x) interpolation law by E' intervals at E
!          integer array ilx=[ilx(j),j=1..nf-1]
!       u: Constant that defines the upper energy limit for E' (real*8)
!          E' lies in the closed interval [0,E-u] if u=0, then max(E')=E
!
  subroutine wrtmf5_gen_eva(nou,jou,mat,mt,ns,np,ep,p,ilep,ne,e,theta,ile,nf,x,f,ilx,u)
    implicit real*8 (a-h, o-z)
    character*80 nou(*)
    dimension ep(*),p(*),ilep(*),e(*),theta(*),ile(*),x(*),f(*),ilx(*)
    dimension nbt(nrmax),ibt(nrmax)
    if (np.gt.ne5max) then
      write(*,*)' MF5/lf=5 Warning: Too many incident energies for p(E) np=',np,' ne5max=',ne5max
    endif
    if (np.gt.ne5max) then
      write(*,*)' MF5/lf=5 Warning: Too many incident energies for theta(E) ne=',ne,' ne5max=',ne5max
    endif
    if (nf.gt.nep5max) then
      write(*,*)' MF5/lf=5 Warning: Too many outgoing energies for f(x) nf=',nf,' nep5max=',nep5max
    endif
    if (np.gt.npmax.or.ne.gt.npmax.or.nf.gt.npmax) then
      write(*,*)' MF5/lf=5 Error: Too many array elements. np=',np,' ne=',ne,' nf=',nf,' npmax=',npmax
      stop
    endif
    mf=5
    lf=5
    c1=0.0d0
    c2=0.0d0
    l1=0
    l2=0
    n1=0
    n2=0
    call packibt(np,ilep,nr,nbt,ibt)
    call wrtab1(nou,jou,mat,mf,mt,ns,u,c2,l1,lf,nr,nbt,ibt,np,ep,p)
    call packibt(ne,ile,nr,nbt,ibt)
    call wrtab1(nou,jou,mat,mf,mt,ns,c1,c2,l1,l2,nr,nbt,ibt,ne,e,theta)
    call packibt(nf,ilx,nr,nbt,ibt)
    call wrtab1(nou,jou,mat,mf,mt,ns,c1,c2,l1,l2,nr,nbt,ibt,nf,x,f)
    return
  end subroutine wrtmf5_gen_eva
!
! =================================================================================================================================
! subroutine wrtmf5_gen_eva_law: Similar to subroutine wrtmf5_gen_eva, but each interpolation scheme
!                                is represented by only one interpolation law
! =================================================================================================================================
!
!   ilepk: interpolation law for [E, p(E)]
!    ilek: interpolation law for [E,theta(E)]
!    ilxk: interpolation law for [x,f(x)], x=E'/theta(E)
!   The rest of parameters have the same meaning as before
!
  subroutine wrtmf5_gen_eva_law(nou,jou,mat,mt,ns,np,ep,p,ilepk,ne,e,theta,ilek,nf,x,f,ilxk,u)
    implicit real*8 (a-h, o-z)
    character*80 nou(*)
    dimension ep(*),p(*),e(*),theta(*),x(*),f(*)
    dimension ilep(np),ile(ne),ilx(nf)
    do i=1,np-1
      ilep(i)=ilepk
    enddo
    do i=1,ne-1
      ile(i)=ilek
    enddo
    do i=1,nf-1
      ilx(i)=ilxk
    enddo
    call wrtmf5_gen_eva(nou,jou,mat,mt,ns,np,ep,p,ilep,ne,e,theta,ile,nf,x,f,ilx,u)
    return
  end subroutine wrtmf5_gen_eva_law
!
! =================================================================================================================================
! subroutine wrtmf5_gen_eva_lin: Similar to subroutine wrtmf5_gen_eva, but all interpolation schema
!                                are lin-lin, therefore they are not input parameters.
!                                The rest of parameters have the same meaning as before
! =================================================================================================================================
!
  subroutine wrtmf5_gen_eva_lin(nou,jou,mat,mt,ns,np,ep,p,ne,e,theta,nf,x,f,u)
    implicit real*8 (a-h, o-z)
    character*80 nou(*)
    dimension ep(*),p(*),e(*),theta(*),x(*),f(*)
    dimension ilep(np),ile(ne),ilx(nf)
    do i=1,np-1
      ilep(i)=2
    enddo
    do i=1,ne-1
      ile(i)=2
    enddo
    do i=1,nf-1
      ilx(i)=2
    enddo
    call wrtmf5_gen_eva(nou,jou,mat,mt,ns,np,ep,p,ilep,ne,e,theta,ile,nf,x,f,ilx,u)
    return
  end subroutine wrtmf5_gen_eva_lin
!
! =================================================================================================================================
! subroutine wrtmf5_maxwell: Write a Maxwellian fission spectrum (lf=7 MF5/MT subsection)
! =================================================================================================================================
!
!   fk(E' <- E) = fk(E',E) = p(E)*(1/I)*sqrt(E')*exp(E'/theta(E))
!
!   where I=c(E,theta(E),u) is a normalization constant and E' is within the interval [0,E-u]
!
!      np: number of incident energy points at which p(E) is given
!      ep: incident energy grid for p(E)
!          real*8 array ep=[ep(i),i=1..np]
!       p: fractional probability p(E) of the partial distribution
!          real*8 array p=[p(i),i=1..np]
!    ilep: (p(E),E) interpolation law by intervals
!          integer array ilep=[ilep(i),i=1..np-1]
!      ne: number of incident energy points at which theta(E) is given
!       e: incident energy grid for theta(E)
!          real*8 array e=[e(i),i=1..ne]
!   theta: Effective temperature theta(E) in eV
!          real*8 array theta=[theta(i),i=1..ne]
!     ile: (theta(E),E) interpolation law by intervals
!          integer array ile=[ile(i),i=1..ne-1]
!       u: Constant that defines the upper energy limit for E' (real*8)
!          E' lies in the closed interval [0,E-u] if u=0, then max(E')=E
!
  subroutine wrtmf5_maxwell(nou,jou,mat,mf,ns,np,ep,p,ilep,ne,e,theta,ile,u)
    implicit real*8 (a-h, o-z)
    character*80 nou(*)
    dimension ep(*),p(*),ilep(*),e(*),theta(*),ile(*)
    dimension nbt(nrmax),ibt(nrmax)
    if (np.gt.ne5max) then
      write(*,*)' MF5/lf=7 Warning: Too many incident energies for p(E) np=',np,' ne5max=',ne5max
    endif
    if (ne.gt.ne5max) then
      write(*,*)' MF5/lf=7 Warning: Too many incident energies for theta(E) ne=',ne,' ne5max=',ne5max
    endif
    if (np.gt.npmax.or.ne.gt.npmax) then
      write(*,*)' MF5/lf=7 Error: Too many array elements. np=',np,' ne=',ne,' npmax=',npmax
      stop
    endif
    mf=5
    lf=7
    c1=0.0d0
    c2=0.0d0
    l1=0
    l2=0
    n1=0
    n2=0
    call packibt(np,ilep,nr,nbt,ibt)
    call wrtab1(nou,jou,mat,mf,mt,ns,u,c2,l1,lf,nr,nbt,ibt,np,ep,p)
    call packibt(ne,ile,nr,nbt,ibt)
    call wrtab1(nou,jou,mat,mf,mt,ns,c1,c2,l1,l2,nr,nbt,ibt,ne,e,theta)
    return
  end subroutine wrtmf5_maxwell
!
! =================================================================================================================================
! subroutine wrtmf5_maxwell_law: Similar to subroutine wrtmf5_maxwell, but each interpolation scheme
!                                is represented by only one interpolation law
! =================================================================================================================================
!
!   ilepk: interpolation law for [E, p(E)]
!    ilek: interpolation law for [E,theta(E)]
!   The rest of parameters have the same meaning as before
!
  subroutine wrtmf5_maxwell_law(nou,jou,mat,mf,ns,np,ep,p,ilepk,ne,e,theta,ilek,u)
    implicit real*8 (a-h, o-z)
    character*80 nou(*)
    dimension ep(*),p(*),e(*),theta(*)
    dimension ilep(np),ile(ne)
    do i=1,np-1
      ilep(i)=ilepk
    enddo
    do i=1,ne-1
      ile(i)=ilek
    enddo
    call wrtmf5_maxwell(nou,jou,mat,mf,ns,np,ep,p,ilep,ne,e,theta,ile,u)
    return
  end subroutine wrtmf5_maxwell_law
!
! =================================================================================================================================
! subroutine wrtmf5_maxwell_lin: Similar to subroutine wrtmf5_maxwell, but all interpolation schema
!                                are lin-lin, therefore they are not input parameters.
!                                The rest of parameters have the same meaning as before
! =================================================================================================================================
!
  subroutine wrtmf5_maxwell_lin(nou,jou,mat,mf,ns,np,ep,p,ne,e,theta,u)
    implicit real*8 (a-h, o-z)
    character*80 nou(*)
    dimension ep(*),p(*),e(*),theta(*)
    dimension ilep(np),ile(ne)
    do i=1,np-1
      ilep(i)=2
    enddo
    do i=1,ne-1
      ile(i)=2
    enddo
    call wrtmf5_maxwell(nou,jou,mat,mf,ns,np,ep,p,ilep,ne,e,theta,ile,u)
    return
  end subroutine wrtmf5_maxwell_lin
!
! =================================================================================================================================
! subroutine wrtmf5_eva: Write an evaporation spectrum (lf=9 MF5/MT subsection)
! =================================================================================================================================
!
!   fk(E' <- E) = fk(E',E) = p(E)*(1/I)*E'*exp(E'/theta(E))
!
!   where I=c(E,theta(E),u) normalization constant and E' is within in the interval [0,E-u]
!
!      np: number of incident energy points at which p(E) is given
!      ep: incident energy grid for p(E)
!          real*8 array ep=[ep(i),i=1..np]
!       p: fractional probability p(E) of the partial distribution
!          real*8 array p=[p(i),i=1..np]
!    ilep: (p(E),E) interpolation law by intervals
!          integer array ilep=[ilep(i),i=1..np-1]
!      ne: number of incident energy points at which theta(E) is given
!       e: incident energy grid for theta(E)
!          real*8 array e=[e(i),i=1..ne]
!   theta: Effective temperature theta(E) in eV
!          real*8 array theta=[theta(i),i=1..ne]
!     ile: (theta(E),E) interpolation law by intervals
!          integer array ile=[ile(i),i=1..ne-1]
!       u: Constant that defines the upper energy limit for E' (real*8)
!          E' lies in the closed interval [0,E-u] if u=0, then max(E')=E
!
  subroutine wrtmf5_eva(nou,jou,mat,mt,ns,np,ep,p,ilep,ne,e,theta,ile,u)
    implicit real*8 (a-h, o-z)
    character*80 nou(*)
    dimension ep(*),p(*),ilep(*),e(*),theta(*),ile(*)
    dimension nbt(nrmax),ibt(nrmax)
    if (np.gt.ne5max) then
      write(*,*)'  MF5/lf=9 Warning: Too many incident energies for p(E) np=',np,' ne5max=',ne5max
    endif
    if (ne.gt.ne5max) then
      write(*,*)'  MF5/lf=9 Warning: Too many incident energies for theta(E) ne=',ne,' ne5max=',ne5max
    endif
    if (np.gt.npmax.or.ne.gt.npmax) then
      write(*,*)' MF5/lf=9 Error: Too many array elements. np=',np,' ne=',ne,' npmax=',npmax
      stop
    endif
    mf=5
    lf=9
    c1=0.0d0
    c2=0.0d0
    l1=0
    l2=0
    n1=0
    n2=0
    call packibt(np,ilep,nr,nbt,ibt)
    call wrtab1(nou,jou,mat,mf,mt,ns,u,c2,l1,lf,nr,nbt,ibt,np,ep,p)
    call packibt(ne,ile,nr,nbt,ibt)
    call wrtab1(nou,jou,mat,mf,mt,ns,c1,c2,l1,l2,nr,nbt,ibt,ne,e,theta)
    return
  end subroutine wrtmf5_eva
!
! =================================================================================================================================
! subroutine wrtmf5_eva_law: Similar to subroutine wrtmf5_maxwell, but each interpolation scheme
!                            is represented by only one interpolation law
! =================================================================================================================================
!
!   ilepk: interpolation law for [E, p(E)]
!    ilek: interpolation law for [E,theta(E)]
!   The rest of parameters have the same meaning as before
!
  subroutine wrtmf5_eva_law(nou,jou,mat,mt,ns,np,ep,p,ilepk,ne,e,theta,ilek,u)
    implicit real*8 (a-h, o-z)
    character*80 nou(*)
    dimension ep(*),p(*),e(*),theta(*)
    dimension ilep(np),ile(ne)
    do i=1,np-1
      ilep(i)=ilepk
    enddo
    do i=1,ne-1
      ile(i)=ilek
    enddo
    call wrtmf5_eva(nou,jou,mat,mf,ns,np,ep,p,ilep,ne,e,theta,ile,u)
    return
  end subroutine wrtmf5_eva_law
!
! =================================================================================================================================
! subroutine wrtmf5_eva_lin: Similar to subroutine wrtmf5_maxwell, but all interpolation schema
!                            are lin-lin, therefore they are not input parameters.
!                            The rest of parameters have the same meaning as before
! =================================================================================================================================
!
  subroutine wrtmf5_eva_lin(nou,jou,mat,mf,ns,np,ep,p,ne,e,theta,u)
    implicit real*8 (a-h, o-z)
    character*80 nou(*)
    dimension ep(*),p(*),e(*),theta(*)
    dimension ilep(np),ile(ne)
    do i=1,np-1
      ilep(i)=2
    enddo
    do i=1,ne-1
      ile(i)=2
    enddo
    call wrtmf5_eva(nou,jou,mat,mf,ns,np,ep,p,ilep,ne,e,theta,ile,u)
    return
  end subroutine wrtmf5_eva_lin
!
! =================================================================================================================================
! subroutine wrtmf5_watt: Write energy-dependent Watt spectrum (lf=11 MF5/MT subsection)
! =================================================================================================================================
!
!   fk(E' <- E) = fk(E',E) = p(E)*(1/I)*exp(E'/a(E))*sinh(sqrt(b(E)*E'))
!
!   where I=c(E,a(E),b(E),u) normalization constant and E' is within in the interval [0,E-u]
!
!     np: number of incident energy points at which p(E) is given
!     ep: incident energy grid for p(E)
!         real*8 array ep=[ep(i),i=1..np]
!      p: fractional probability p(E) of the partial distribution
!         real*8 array p=[p(i),i=1..np]
!   ilep: (p(E),E) interpolation law by intervals
!         integer array ilep=[ilep(i),i=1..np-1]
!    nea: number of incident energy points at which a(E) is given
!     ea: incident energy grid for a(E)
!         real*8 array ea=[ea(i),i=1..nea]
!      a: Effective parameter a(E) in eV
!         real*8 array a=[a(i),i=1..nea]
!   ilea: (a(E),E) interpolation law by intervals
!         integer array ile=[ilea(i),i=1..nea-1]
!    neb: number of incident energy points at which b(E) is given
!     eb: incident energy grid for b(E)
!         real*8 array eb=[eb(i),i=1..neb]
!      b: Effective parameter b(E) in 1/eV
!         real*8 array b=[b(i),i=1..neb]
!   ileb: (b(E),E) interpolation law by intervals
!         integer array ile=[ileb(i),i=1..neb-1]
!      u: Constant that defines the upper energy limit for E' (real*8)
!         E' lies in the closed interval [0,E-u] if u=0, then max(E')=E
!
  subroutine wrtmf5_watt(nou,jou,mat,mt,ns,np,ep,p,ilep,nea,ea,a,ilea,neb,eb,b,ileb,u)
    implicit real*8 (a-h, o-z)
    character*80 nou(*)
    dimension ep(*),p(*),ilep(*),ea(*),a(*),ilea(*),eb(*),b(*),ileb(*)
    dimension nbt(nrmax),ibt(nrmax)
    if (np.gt.ne5max) then
      write(*,*)' MF5/lf=11: Too many incident energies for p(E) np=',np,' ne5max=',ne5max
    endif
    if (nea.gt.ne5max) then
      write(*,*)' MF5/lf=11: Too many incident energies for a(E) ne=',nea,' ne5max=',ne5max
    endif
    if (neb.gt.ne5max) then
      write(*,*)' MF5/lf=11: Too many incident energies for b(E) ne=',neb,' ne5max=',ne5max
    endif
    if (np.gt.npmax.or.nea.gt.npmax.or.neb.gt.npmax) then
      write(*,*)' MF5/lf=11 Error: Too many array elements. np=',np,' ne_a=',nea,' ne_b=',neb,' npmax=',npmax
      stop
    endif
    mf=5
    lf=11
    c1=0.0d0
    c2=0.0d0
    l1=0
    l2=0
    n1=0
    n2=0
    call packibt(np,ilep,nr,nbt,ibt)
    call wrtab1(nou,jou,mat,mf,mt,ns,u,c2,l1,lf,nr,nbt,ibt,np,ep,p)
    call packibt(nea,ilea,nr,nbt,ibt)
    call wrtab1(nou,jou,mat,mf,mt,ns,c1,c2,l1,l2,nr,nbt,ibt,nea,ea,a)
    call packibt(neb,ileb,nr,nbt,ibt)
    call wrtab1(nou,jou,mat,mf,mt,ns,c1,c2,l1,l2,nr,nbt,ibt,neb,eb,b)
    return
  end subroutine wrtmf5_watt
!
! =================================================================================================================================
! subroutine wrtmf5_watt_law: Similar to subroutine wrtmf5_watt, but each interpolation scheme
!                             is represented by only one interpolation law
! =================================================================================================================================
!
!   ilepk: interpolation law for [E, p(E)]
!   ileak: interpolation law for [E, a(E)]
!   ilebk: interpolation law for [E, b(E)]
!   The rest of parameters have the same meaning as before
!
  subroutine wrtmf5_watt_law(nou,jou,mat,mt,ns,np,ep,p,ilepk,nea,ea,a,ileak,neb,eb,b,ilebk,u)
    implicit real*8 (a-h, o-z)
    character*80 nou(*)
    dimension ep(*),p(*),ea(*),a(*),eb(*),b(*)
    dimension ilep(np),ilea(nea),ileb(neb)
    do i=1,np-1
      ilep(i)=ilepk
    enddo
    do i=1,nea-1
      ilea(i)=ileak
    enddo
    do i=1,neb-1
      ileb(i)=ilebk
    enddo
    call wrtmf5_watt(nou,jou,mat,mt,ns,np,ep,p,ilep,nea,ea,a,ilea,neb,eb,b,ileb,u)
    return
  end subroutine wrtmf5_watt_law
!
! =================================================================================================================================
! subroutine wrtmf5_watt_lin: Similar to subroutine wrtmf5_watt, but all interpolation schema
!                             are lin-lin, therefore they are not input parameters.
!                             The rest of parameters have the same meaning as before
! =================================================================================================================================
!
  subroutine wrtmf5_watt_lin(nou,jou,mat,mt,ns,np,ep,p,nea,ea,a,neb,eb,b,u)
    implicit real*8 (a-h, o-z)
    character*80 nou(*)
    dimension ep(*),p(*),ea(*),a(*),eb(*),b(*)
    dimension ilep(np),ilea(nea),ileb(neb)
    do i=1,np-1
      ilep(i)=2
    enddo
    do i=1,nea-1
      ilea(i)=2
    enddo
    do i=1,neb-1
      ileb(i)=2
    enddo
    call wrtmf5_watt(nou,jou,mat,mt,ns,np,ep,p,ilep,nea,ea,a,ilea,neb,eb,b,ileb,u)
    return
  end subroutine wrtmf5_watt_lin
!
! =================================================================================================================================
! subroutine wrtmf5_madland_nix: Write Madland and Nix fission neutron spectrum (lf=12 MF5/MT subsection)
! =================================================================================================================================
!
!   fk(E' <- E) = fk(E',E) = p(E)*0.5*(g(E',tm(E),efl)+g(E',tm(E),efh))
!
!   where g(E',tm,Ef) is an analytical function given by Madland and Nix, and E' is whitin the interval [0, infinity)
!
!     np: number of incident energy points at which p(E) is given
!     ep: incident energy grid for p(E)
!         real*8 array ep=[ep(i),i=1..np]
!      p: fractional probability p(E) of the partial distribution
!         real*8 array p=[p(i),i=1..np]
!   ilep: (p(E),E) interpolation law by intervals
!         integer array ilep=[ilep(i),i=1..np-1]
!     ne: number of incident energy points at which tm(E) is given
!      e: incident energy grid for tm(E)
!         real*8 array e=[e(i),i=1..ne]
!     tm: Effective parameter tm(E) in eV
!         real*8 array tm=[tm(i),i=1..ne]
!    ile: (tm(E),E) interpolation law by intervals
!         integer array ile=[ile(i),i=1..ne-1]
!    efl: average kinetic energy per nucleon of the average
!         light fission fragment (real*8)
!    efh: average kinetic energy per nucleon of the average
!         heavy fission fragment (real*8)
!
  subroutine wrtmf5_madland_nix(nou,jou,mat,mt,ns,np,ep,p,ilep,ne,e,tm,ile,efl,efh)
    implicit real*8 (a-h, o-z)
    character*80 nou(*)
    dimension ep(*),p(*),ilep(*),e(*),tm(*),ile(*)
    dimension nbt(nrmax),ibt(nrmax)
    if (np.gt.ne5max) then
      write(*,*)' MF5/lf=12 Warning: Too many incident energies for p(E) np=',np,' ne5max=',ne5max
    endif
    if (ne.gt.ne5max) then
      write(*,*)' MF5/lf=12 Warning: Too many incident energies for tm(E) ne=',ne,' ne5max=',ne5max
    endif
    if (np.gt.npmax.or.ne.gt.npmax) then
      write(*,*)' MF5/lf=12 Error: Too many array elements. np=',np,' ne=',ne,' npmax=',npmax
      stop
    endif
    mf=5
    lf=12
    c1=0.0d0
    c2=0.0d0
    l1=0
    l2=0
    n1=0
    n2=0
    call packibt(np,ilep,nr,nbt,ibt)
    call wrtab1(nou,jou,mat,mf,mt,ns,c1,c2,l1,lf,nr,nbt,ibt,np,ep,p)
    call packibt(ne,ile,nr,nbt,ibt)
    call wrtab1(nou,jou,mat,mf,mt,ns,efl,efh,l1,l2,nr,nbt,ibt,ne,e,tm)
    return
  end subroutine wrtmf5_madland_nix
!
! =================================================================================================================================
! subroutine wrtmf5_madland_nix_law: Similar to subroutine wrtmf5_madland_nix, but each interpolation scheme
!                                    is represented by only one interpolation law
! =================================================================================================================================
!
!   ilepk: interpolation law for [E, p(E)]
!    ilek: interpolation law for [E, tm(E)]
!   The rest of parameters have the same meaning as before
!
  subroutine wrtmf5_madland_nix_law(nou,jou,mat,mt,ns,np,ep,p,ilepk,ne,e,tm,ilek,efl,efh)
    implicit real*8 (a-h, o-z)
    character*80 nou(*)
    dimension ep(*),p(*),e(*),tm(*)
    dimension ilep(np),ile(ne)
    do i=1,np-1
      ilep(i)=ilepk
    enddo
    do i=1,ne-1
      ile(i)=ilek
    enddo
    call wrtmf5_madland_nix(nou,jou,mat,mt,ns,np,ep,p,ilep,ne,e,tm,ile,efl,efh)
    return
  end subroutine wrtmf5_madland_nix_law
!
! =================================================================================================================================
! subroutine wrtmf5_madland_nix_lin: Similar to subroutine wrtmf5_watt, but all interpolation schema
!                                    are lin-lin, therefore they are not input parameters.
!                                    The rest of parameters have the same meaning as before
! =================================================================================================================================
!
  subroutine wrtmf5_madland_nix_lin(nou,jou,mat,mt,ns,np,ep,p,ne,e,tm,efl,efh)
    implicit real*8 (a-h, o-z)
    character*80 nou(*)
    dimension ep(*),p(*),e(*),tm(*)
    dimension ilep(np),ile(ne)
    do i=1,np-1
      ilep(i)=2
    enddo
    do i=1,ne-1
      ile(i)=2
    enddo
    call wrtmf5_madland_nix(nou,jou,mat,mt,ns,np,ep,p,ilep,ne,e,tm,ile,efl,efh)
    return
  end subroutine wrtmf5_madland_nix_lin
!
! =================================================================================================================================
!
! MF6: Procedures for reading and writing ENDF-6/MF6/MT sub-sections:
! ------------------------------------------------------------------
!
! =================================================================================================================================
! subroutine readmf6mt_head: Read MF6/MT section head
! =================================================================================================================================
!
!   za: ZA number
!  awr: Atomic weight in neutron mass units
!   jp: Integer trigger for treatment of prompt fission neutron and photon
!  lct: Reference system for secondary energy and angles
!   nk: Number of subsections in this section (number of described reaction products)
!
  subroutine readmf6mt_head(nin,jin,za,awr,jp,lct,nk)
    implicit real*8 (a-h, o-z)
    character*80 nin(*)
    call readcont(nin,jin,za,awr,jp,lct,nk,n2,mat,mf,mt,ns)
    if (mt.ne.18.and.jp.gt.0) then
      write(*,*)' Error: jp=',jp,' is not allowed for mt=',mt,' in MF6'
      stop
    elseif (nk.gt.nkmax) then
      write(*,*)' Warning : nk=',nk,' too many subsections in MF6. nkmax=',nkmax
    endif
    return
  end subroutine readmf6mt_head
!
! =================================================================================================================================
! subroutine readmf6_law_form: Get the information of all sub-section (law,ltype,dimensions) in a MF6/MT section knowing jp
! =================================================================================================================================
!
!      jp: law trigger for fission outgoing particles (input integer parameter read by subroutine readmf6_head)
!      nk: number of subsections in this section (input integer parameter read by readmf6_head)
!
!     law: energy-angular distribution representation in MF6 (see below)
!          integer array law=[law(k), k=1..nk]
!   ltype: flag for law-dependent representation (see below)
!          integer array ltype=[ltype(k), k=1..nk]
!    lzap: Product ZA number (1000*Zp+Ap)
!          integer array lzap=[lzap(k), k=1..nk]
!     awp: Product mass in neutron units
!          real*8 array awp=[awp(k), k=1..nk]
!     lip: Product modified flag
!          integer array lip=[lip(k), k=1..nk]
!      np: Number of incident energy points for y(E) (yields,if JP=0) or P(n,E) (distributions for fission particles,if JP>0)
!          integer array np=[np(k), k=1..nk]
!      ne: Number of incident energy points for angle-energy data
!          integer array ne=[ne(k), k=1..nk]
!      na: Angular data dimension
!          integer array na=[na(k), k=1..nk]
!          for law=1, ltype=3 (mixed representation see below)
!          na(k)=100*nah(k)+nal(k), where nah(k) is the dimension for pointwise representation at higher energies and
!          nal(k) is the number of Legendre coefficients at lower energies
!     nep: Secondary energy dimension
!          integer array nep=[nep(k),k=1..nk]
!          for law=1, ltype=3 (mixed representation see below) nep(k) is the dimension of the incident energy data at higher
!          energies. In this case, ne(k) contains the incident energy dimension for the Legendre expansion at lower energies
!
!   law(k) and ltype(k) values:
!
!   For jp=0
!    law=0, ltype=0: None distribution is given, only yields.
!    law=1, ltype=0: Continuum full isotropic energy-angle distribution (na=0, for all incident energies)
!    law=1, ltype=1: Continuum energy-angle distribution (Legendre expansion, lang=1, na>0)
!    law=1, ltype=2: Continuum energy-angle distribution (Kalbach-Mann formalism, lang=2, na=[1,2] energy dependent data)
!    law=1, ltype=3: Continuum energy-angle distribution (Kalbach-Mann formalism, lang=2, na=1 for all incident energies)
!    law=1, ltype=4: Continuum energy-angle distribution (Kalbach-Mann formalism, lang=2, na=2 for all incident energies)
!    law=1, ltype=5: Continuum energy-angle distribution (Tabulated data, lang=11-15)
!    law=2, ltype=1: Two-body reaction angular distribution (Legendre expansion, lang=0)
!    law=2, ltype=2: Two-body reaction angular distribution (Tabulated data, lang=12 or 14)
!    law=2, ltype=3: Two-body reaction angular distribution (Legendre expansion at low E, Tabulated data at high E)
!    law=3, ltype=0: Two-body reaction angular distribution (Isotropic distribution, only yiels)
!    law=4, ltype=0: Recoil distribution for a two-body reaction (To combine with law=2 or 3, only yields)
!    law=5, ltype=1: Charged particle elastic scattering (nuclear amplitud expansion, LTP=1)
!    law=5, ltype=2: Charged particle elastic scattering (residual cross section expansion, LTP=2)
!    law=5, ltype=3: Charged particle elastic scattering (nuclear plus interference tabulation, LTP=12,14 or 15)
!    law=6, ltype=0: n-body phase-space distribution
!    law=7, ltype=0: Laboratory continuum angle-energy distribution
!
!   For fission with jp>0 only
!    law=  0, ltype=0: For fission neutrons or photons (no angle-energy distribution given, only P(0,E))
!    law= -5, ltype=0: For fission neutrons (Average angle and energy distributions given in MF=4/5, only P(n,E))
!    law=-15, ltype=0: For fission photons (Average angle and energy distributions given in MF=14/15, only P(n,E))
!    law>  0, ltype=0: For fission neutrons or photons P(n,E) particle distributions given in MF6 (not coded)
!
  subroutine readmf6_law_form(nin,jin,jp,nk,law,ltype,lzap,awp,lip,np,ne,na,nep)
    implicit real*8 (a-h, o-z)
    parameter (nxmax=max(nep6max,nmu6max), nbmax=(2*nmu6max+2)*nep6max)
    character*80 nin(*)
    dimension law(*),ltype(*),lzap(*),awp(*),lip(*),np(*),ne(*),na(*),nep(*)
    dimension nbt(nrmax),ibt(nrmax),x(nxmax),b(nbmax)
    jin0=jin
    do k=1,nk
      call readtab1(nin,jin,zap,awp(k),lip(k),lawk,nr,nbt,ibt,npk,x,b)
      kzap=nint(zap+1.0d-12)
      if (npk.gt.ne6max) then
        write(*,*)' MF6/law=',lawk,' Warning: zap=',kzap,' np=',npk,' too many incident energy points. ne6max=',ne6max
      endif
      if (nr.gt.nrmax) then
        write(*,*)' MF6/law=',lawk,' Warning: zap=',kzap,' nr=',nr,' too many interpolation ranges. nrmax=',nrmax
      endif
      law(k)=lawk
      lzap(k)=kzap
      np(k)=npk
      if (jp.eq.0.and.(lawk.ge.0.and.lawk.le.7)) then
        if (lawk.eq.1) then
          call readtab2(nin,jin,c1,c2,lang,l2,nr,nbt,ibt,nek)
          if (nek.gt.ne6max) then
             write(*,*)' MF6/law=1 Warning: zap=',kzap,' ne=',nek,' too many incident energy points. ne6max=',ne6max
          endif
          if (nr.gt.nrmax) then
            write(*,*)' MF6/law=',lawk,' Warning: zap=',kzap,' nr=',nr,' too many interpolation ranges. nrmax=',nrmax
            stop
          endif
          if (lang.ge.11.or.lang.le.15) then
            lang=3
          elseif (lang.ne.1.or.lang.ne.2) then
            write(*,*)' MF6/law=1 Error: zap=',kzap,' lang=',lang,' is not allowed'
            stop
          endif
          ne(k)=nek
          call readlist(nin,jin,c1,c2,l1,nak,n1,nepk,b)
          if (lang.eq.3) then
            nak2=nak/2
            if (nak.ne.(2*nak2)) then
              write(*,*)' zap=',kzap,' E=',c2
              write(*,*)' MF6/law=1 lang=11-15 Error: Number of agular parameters is not an even number na=',nak
              stop
            endif
          endif
          nakmax=nak
          nakmin=nak
          nepkmax=nepk
          do i=2,nek
            call readlist(nin,jin,c1,c2,l1,nak,n1,nepk,b)
            if (lang.eq.3) then
              nak2=nak/2
              if (nak.ne.(2*nak2)) then
                write(*,*)' zap=',kzap,' E=',c2
                write(*,*)' MF6/law=1 lang=11-15 Error: Number of agular parameters is not an even number na=',nak
                stop
              endif
            endif
            if (nak.gt.nakmax) then
              nakmax=nak
            elseif (nak.lt.nakmin) then
              nakmin=nak
            endif
            if (nepk.gt.nepkmax) nepkmax=nepk
          enddo
          if (nepkmax.gt.nep6max) then
            write(*,*)' MF6/law=1 Warning: zap=',kzap,' nep=',nepkmax,' too many outgoing energy points. nep6max=',nep6max
          endif
          nep(k)=nepkmax
          if (nakmax.eq.0) then
            na(k)=0
            ltype(k)=0
          else
            if (lang.eq.1) then
              if (nakmax.gt.nlmax) then
                write(*,*)' MF6/law=1 lang=1 Error: Too high Legendre order nl=',nakmax,' for zap=',kzap,' nlmax=',nlmax
                stop
              else
                nakmax=nakmax+1
                na(k)=nakmax
                ltype(k)=1
              endif
            elseif (lang.eq.2) then
              if (nakmax.gt.2) then
                write(*,*)' MF6/law=1 lang=2 Error: Too large na=',nakmax,' for zap=',kzap,' namax= 2'
                stop
              else
                na(k)=nakmax
                if (nakmax.ne.nakmin) then
                  ltype(k)=2
                elseif (nakmax.eq.1) then
                  ltype(k)=3
                else
                  ltype(k)=4
                endif
              endif
            else
              nakmax=nakmax/2
              if (nakmax.gt.nmu6max) then
                write(*,*)' MF6/law=1 lang=11-15 Warning: Too many cosines',nakmax,' for zap=',kzap,' nmu6max=',nmu6max
              endif
              na(k)=nakmax
              ltype(k)=5
            endif
          endif
          if (nakmax.gt.0) then
            nmax=nek*nepkmax*nakmax
          else
            nmax=nek*nepkmax
          endif
          if (nmax.gt.npmax) then
            write(*,*)' MF6/law=1 ltype=',ltype(k),' Error: Arrays too large np=',nmax,' for zap=',kzap,' npmax=',npmax
            stop
          endif
        elseif (lawk.eq.2) then
          call readtab2(nin,jin,c1,c2,l2,l2,nr,nbt,ibt,nek)
          if (nek.gt.ne6max) then
            write(*,*)' MF6/law=2 Warning: zap=',kzap,' ne=',nek,' too many incident energy points. ne6max=',ne6max
          endif
          if (nr.gt.nrmax) then
            write(*,*)' MF6/law=2 Warning: zap=',kzap,' nr=',nr,' too many interpolation ranges. nrmax=',nrmax
            stop
          endif
          call readlist(nin,jin,c1,c2,lang,l2,n1,nli,b)
          if (lang.eq.0) then
            if (n1.ne.nli) then
              write(*,*)' MF6/law=2 Error: zap=',kzap,' format error nw=',n1,' nl=',nli,' lang=',lang
              stop
            endif
            lang=1
          elseif (lang.eq.12.or.lang.eq.14) then
            if (n1.ne.(2*nli)) then
              write(*,*)' MF6/law=2 Error: zap=',kzap,' format error nw=',n1,' nl=',nli,' lang=',lang
              stop
            endif
            lang=2
          else
            write(*,*)' MF6/law=2 Error: zap=',kzap,' lang=',lang,' is not allowed'
            stop
          endif
          langi=lang
          langl=lang
          langh=lang
          if (lang.eq.1) then
            nal=nli
            nah=0
            nel=1
            neh=0
          else
            nal=0
            nah=nli
            nel=0
            neh=1
          endif
          n=0
          do i=2,nek
            call readlist(nin,jin,c1,c2,lang,l2,n1,nli,b)
            if (lang.eq.0) then
             if (n1.ne.nli) then
                write(*,*)' MF6/law=2 Error: zap=',kzap,' format error nw=',n1,' nl=',nli,' lang=',lang
                stop
              endif
              lang=1
              if (nli.gt.nal) nal=nli
              nel=nel+1
            elseif (lang.eq.12.or.lang.eq.14) then
              if (n1.ne.(2*nli)) then
                write(*,*)' MF6/law=2 Error: zap=',kzap,' format error nw=',n1,' nl=',nli,' lang=',lang
                stop
              endif
              lang=2
              if (nli.gt.nah) nah=nli
              neh=neh+1
            else
              write(*,*)' MF6/law=2 Error: lang=',lang,' is not allowed'
              stop
            endif
            if (lang.ne.langi) then
              if (n.eq.0) langh=lang
              langi=lang
              n=n+1
            endif
          enddo
          if (nah.gt.nmu6max) then
            write(*,*)' MF6/law=2 lang=12-14 Warning: Too many cosines nmu=',nah,' for zap=',kzap,' nmu6max=',nmu6max
          endif
          if (nal.gt.nlmax) then
            write(*,*)' MF6/law=2 lang=0 Error: Too high Legendre order nl=',nal,' for zap=',kzap,' nlmax=',nlmax
            stop
          endif
          if (nal.gt.0) nal=nal+1
          if (nel*nal.gt.npmax.or.neh*nah.gt.npmax) then
            write(*,*)' MF6/law=2 Error: Too large arrays for zap=',kzap,' npmax=',npmax
            write(*,*)' ne=',nel+neh,' nel=',nel,' nal=',nal,' neh=',neh,' nah=',nah
            stop
          endif
          if (n.eq.0) then
            ne(k)=max(nel,neh)
            na(k)=max(nal,nah)
            nep(k)=0
            ltype(k)=langl
          elseif (n.eq.1.and.langl.eq.1.and.langh.eq.2) then
            ne(k)=nel
            nep(k)=neh
            na(k)=100*nah+nal
            ltype(k)=3
          else
            write(*,*)' MF6/law=2 Error: Randomly mixed representations not coded. n=',n,' langl=',langl,' langh=',langh
            stop
          endif
        elseif (lawk.eq.5) then
          call readtab2(nin,jin,c1,c2,lidp,l2,nr,nbt,ibt,nek)
          if (nek.gt.ne6max) then
            write(*,*)' MF6/law=5 Warning: zap=',kzap,' ne=',nek,' too many incident energy points. ne6max=',ne6max
          endif
          if (nr.gt.nrmax) then
            write(*,*)' MF6/law=5 Warning: zap=',kzap,' nr=',nr,' too many interpolation ranges. nrmax=',nrmax
            stop
          endif
          ne(k)=nek
          call readlist(nin,jin,c1,c2,ltp,l2,n1,nlk,b)
          if (ltp.gt.2) ltp=3
          ltpk=ltp
          do i=2,nek
            call readlist(nin,jin,c1,c2,ltp,l2,n1,nli,b)
            if (ltp.gt.2) ltp=3
            if (ltp.ne.ltpk) then
              write(*,*)' MF6/law=5 Error: Random mixed representations not coded. ltpk=',ltpk,' ltp=',ltp
              stop
            endif
            if (nli.gt.nlk) nlk=nli
          enddo
          if (ltpk.eq.1.or.ltpk.eq.2) then
            if (nlk.gt.nlmax) then
              write(*,*)' MF6/law=5 ltp=',ltpk,' Error: Too high Legendre order nl=',nlk,' for zap=',kzap,' nlmax=',nlmax
              stop
            endif
            na(k)=nlk+1
            if (ltpk.eq.1.and.lidp.eq.0) then
              nep(k)=2*nlk+1
              if (nep(k)*nek.gt.npmax) then
                write(*,*)' MF6/law=5 Error: Too large arrays for zap=',kzap,' npmax=',npmax
                write(*,*)' ne=',nek,' na=',na(k),' nep=',nep(k),' ne*na=',na(k)*nek,' nep*na=',nep(k)*nek
                stop
              endif
            else
              nep(k)=nlk+1
              if ((na(k)*nek).gt.npmax) then
                write(*,*)' MF6/law=5 Error: Too large arrays for zap=',kzap,' npmax=',npmax
                write(*,*)' ne=',nek,' na=',na(k),' nep=',nep(k),' ne*na=',na(k)*nek,' nep*na=',nep(k)*nek
                stop
              endif
            endif
          else
            na(k)=nlk
            nep(k)=0
            if (nlk.gt.nmu6max) then
              write(*,*)' MF6/law=5 Error: too many cosines values for zap=',kzap,' nmu=',nlk,' nmu6max=',nmu6max
            endif
            if (nlk*nek.gt.npmax) then
              write(*,*)' MF6/law=5 Error: Too large arrays for zap=',kzap,' npmax=',npmax
              write(*,*)' ne=',nek,' na=',nlk,' ne*na=',nlk*nek
              stop
            endif
          endif
          ltype(k)=ltpk
        elseif (lawk.eq.7) then
          call readtab2(nin,jin,c1,c2,l1,l2,nr,nbt,ibt,nek)
          if (nek.gt.ne6max) then
            write(*,*)' MF6/law=7 Warning: Too many incident energy points in MF6. ne=',nek,' ne6max=',ne6max
          endif
          if (nr.gt.nrmax) then
            write(*,*)' MF6/law=7 Error: Too many incident energy interpolation ranges. nr=',nr,' nrmax=',nrmax
            stop
          endif
          ne(k)=nek
          nmuk=0
          nepk=0
          do i=1,nek
            call readtab2(nin,jin,c1,c2,l1,l2,nr,nbt,ibt,nmui)
            if (nmui.gt.nmu6max) then
              write(*,*)' MF6/law=7 Warning: Too many cosines. nmu=',nmui,' nmu6max=',nmu6max
            endif
            if (nr.gt.nrmax) then
              write(*,*)'  MF6/law=7 Error: Too many cosines interpolation ranges. nr=',nr,' nrmax=',nrmax
              stop
            endif
            if (nmui.gt.nmuk) nmuk=nmui
            do j=1,nmui
              call readtab1(nin,jin,c1,c2,l1,l2,nr,nbt,ibt,nepi,x,b)
              if (nepi.gt.nep6max) then
                write(*,*)' MF6/law=7 Warning: Too many secondary energies. nep=',nepi,' nep6max=',nep6max
              endif
              if (nr.gt.nrmax) then
                write(*,*)' MF6/law=7 Error: Too many secondary energy interpolation ranges. nr=',nr,' nrmax=',nrmax
                stop
              endif
              if (nepi.gt.nepk) nepk=nepi
            enddo
          enddo
          if (nek*nepk*nmuk.gt.npmax) then
            write(*,*)' MF6/law=7 Error: Too large arrays for zap=',kzap,' npmax=',npmax
            write(*,*)' ne=',nek,' nmu=',nmuk,' nep=',nepk,' ne*nmu*nep=',nek*nepk*nmuk
            stop
          endif
          na(k)=nmuk
          nep(k)=nepk
          ltype(k)=0
        elseif (lawk.eq.0.or.lawk.eq.3.or.lawk.eq.4.or.lawk.eq.6) then
          ne(k)=0
          na(k)=0
          nep(k)=0
          ltype(k)=0
          if (lawk.eq.6) call readcont(nin,jin,c1,c2,l1,l2,n1,n2,mat,mf,mt,ns)
        else
          write(*,*)' MF6 Fatal Error: Unknown law=',lawk,' JP=',jp,' for zap=',kzap
          stop
        endif
      elseif (jp.gt.0.and.(lawk.eq.0.or.lawk.eq.-5.or.lawk.eq.-15)) then
        ne(k)=0
        na(k)=0
        nep(k)=0
        ltype(k)=0
        if (lawk.eq.0) then
          write(*,*)' MF6/law=0/JP>0 Warning: P(0,E) given in MF6 for zap=',kzap,' lip=',lip(k)
        elseif (lawk.eq.-5) then
          write(*,*)' MF6/law=-5/JP>0 Warning: P(n,E) given in MF6 for zap=',kzap,' lip=',lip(k)
          write(*,*)' Average distributions given in MF=4/5.'
        elseif (lawk.eq.-15) then
          write(*,*)' MF6/law=-15/JP>0 Warning: P(n,E) given in MF6 for zap=',kzap,' lip=',lip(k)
          write(*,*)' Average distributions given in MF=14/15.'
        else
          write(*,*)' MF6 Error: law=',lawk,' JP=',jp,' for zap=',kzap,' not coded yet'
          stop
        endif
      endif
    enddo
    jin=jin0
    return
  end subroutine readmf6_law_form
!
! =================================================================================================================================
! subroutine readmf6law0_3_4_neg: Read data for MF6/LAWs=0,3,4 for jp=0 and 0,-5, -15 for jp>0
! =================================================================================================================================
!
!   Only yields or P(n,E) are provided, no angle-energy distribution
!
!   law: MF6/law (0,3,4,-5,-15)
!   zap: Product ZA identifier 1000*Zp+Ap (real*8)
!   awp: Product mass in neutron units (real*8)
!   lip: Product modifier flag (integer) for isomeric state, multiple emission
!    ny: number of incident energies (integer)
!    ey: incident energies E(i)
!        real*8 array ey=[ey(i),i=1..ny]
!     y: y(E) yields for law=0,3,4 jp=0
!        P(n,E) fission particle distribution for law=0,-5,-15 jp>0
!        real*8 array y=[y(i),i=1..ny]
!   ily: interpolation law for (ey,y(ey)) by intervals
!        integer array ily=[ily(i),i=1,ny-1]
!
  subroutine readmf6_law0_3_4_neg(nin,jin,law,zap,awp,lip,ny,ey,y,ily)
    implicit real*8 (a-h, o-z)
    character*80 nin(*)
    dimension ey(*),y(*),ily(*)
    dimension nbt(nrmax),ibt(nrmax)
    call readtab1(nin,jin,zap,awp,lip,law,nr,nbt,ibt,ny,ey,y)
    call unpackibt(nr,nbt,ibt,ny,ily)
    return
  end subroutine readmf6_law0_3_4_neg
!
! =================================================================================================================================
! subroutine readmf6_law1_iso: Read data for MF6/LAW=1 Legendre expansion (law=1, ltype=0 full isotropic)
! =================================================================================================================================
!
!      zap: Product ZA identifier 1000*Zp+Ap (real*8)
!      awp: Product mass in neutron units (real*8)
!      lip: Product modifier flag (integer) for isomeric state, multiple emission
!       ny: number of incident energy points for y(E) (integer)
!       ey: incident energy points for y(E)
!           real*8 array ey=[ey(i),i=1..ny]
!        y: y(E) yields
!           real*8 array y=[y(i),i=1..ny]
!      ily: interpolation law for (E,y(E)) by intervals
!           integer array ily=[ily(i),i=1,ny-1]
!       ne: number of incident energy points for the Legendre expansion (integer)
!        e: incident energy points for the Legendre expansion
!           real*8 array e=[e(i), i=1..ne]
!      ile: incident energy interpolation law by intervals
!           integer array ile=[ile(i),i=1..ne-1]
!       nd: Number of discret energies
!           integer array nd=[nd(i),i=1..ne], 0 <= nd(i) <= nep(i)
!   nepdim: Maximun number of secindary energies (integer)
!      nep: Number of secondary energies
!           integer array nep=[nep(i),i=1..ne]
!       ep: Secondary/outgoing energies
!           real*8 array ep=[ep(j,i), j=1..nep(i), i=1..ne], ne<=ne6max for MF6
!      lep: interpolation law between secondary energies for j>nd(i) (integer)
!       f0: total emission probability f0(E',E)
!           real*8 array f0=[f0(j,i), j=1..nep(i), i=1..ne]
!
!   for j>nd(i), a continuous distribution is given with interpolation law=lep
!   between secondary energies, for j<=nd(i) discret distribution is provided.
!
  subroutine readmf6_law1_iso(nin,jin,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nd,nepdim,nep,ep,lep,f0)
    implicit real*8 (a-h,o-z)
    parameter (nwmax=2*nep6max)
    character*80 nin(*)
    dimension ey(*),y(*),ily(*),e(*),ile(*),nd(*),nep(*),ep(nepdim,*),f0(nepdim,*)
    dimension nbt(nrmax),ibt(nrmax),b(nwmax)
    call readtab1(nin,jin,zap,awp,lip,law,nr,nbt,ibt,ny,ey,y)
    if (law.ne.1) then
      write(*,*)' MF6/law=1 Error: law=',law
      stop
    endif
    call unpackibt(nr,nbt,ibt,ny,ily)
    call readtab2(nin,jin,c1,c2,lang,lep,nr,nbt,ibt,ne)
    call unpackibt(nr,nbt,ibt,ne,ile)
    do i=1,ne
      call readlist(nin,jin,c1,e(i),ndi,nai,nw,nepi,b)
      if (nai.ne.0) then
        write(*,*)' MF6/law=1 Error: distribution is not isotropic na=',nai,' not equal zero as expected'
        stop
      endif
      if (nw.gt.nwmax) then
        write(*,*)' MF6/law=1 Error: LIST record too large. nep=',nepi,' na=',nai,' nw=',nw,' nwmax=',nwmax
        stop
      endif
      nd(i)=ndi
      nep(i)=nepi
      ncyc=2
      do j=1,nepi
        jb=ncyc*(j-1)+1
        ep(j,i)=b(jb)
        jb=jb+1
        f0(j,i)=b(jb)
      enddo
    enddo
    return
  end subroutine readmf6_law1_iso
!
! =================================================================================================================================
! subroutine readmf6_law1_leg: Read data for MF6/LAW=1 Legendre expansion (law=1, lang=1)
! =================================================================================================================================
!
!      zap: Product ZA identifier 1000*Zp+Ap (real*8)
!      awp: Product mass in neutron units (real*8)
!      lip: Product modifier flag (integer) for isomeric state, multiple emission
!       ny: number of incident energy points for y(E) (integer)
!       ey: incident energy points for y(E)
!           real*8 array ey=[ey(i),i=1..ny]
!        y: y(E) yields
!           real*8 array y=[y(i),i=1..ny]
!      ily: interpolation law for (E,y(E)) by intervals
!           integer array ily=[ily(i),i=1,ny-1]
!       ne: number of incident energy points for the Legendre expansion (integer)
!        e: incident energy points for the Legendre expansion
!           real*8 array e=[e(i), i=1..ne]
!      ile: incident energy interpolation law by intervals
!           integer array ile=[ile(i),i=1..ne-1]
!       nd: Number of discret energies
!           integer array nd=[nd(i),i=1..ne], 0 <= nd(i) <= nep(i)
!    nadim: Maximum number of Legendre coefficients (integer)
!       na: Number of Legendre coefficients
!           integer array na=[na(i),i=1..ne],  1 <= na(i) <= ncmax=nlmax+1
!   nepdim: Maximum number of secondary energies (integer)
!      nep: Number of secondary energies
!           integer array nep=[nep(i),i=1..ne]
!       ep: Secondary/outgoing energies
!           real*8 array ep=[ep(j,i), j=1..nep(i), i=1..ne], ne<=nemax for MF6
!      lep: interpolation law between secondary energies for j>nd(i) (integer)
!        a: Legendre coefficients
!           real*8 array a=[a(k,j,i),k=1..na(i),j=1..nep(i),i=1..ne]
!
!   for j>nd(i), a continuous distribution is given with interpolation law=lep
!   between secondary energies, for j<=nd(i) discret distribution is provided.
!
  subroutine readmf6_law1_leg(nin,jin,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nd,nadim,na,nepdim,nep,ep,lep,a)
    implicit real*8 (a-h,o-z)
    parameter (nwmax=(nlmax+2)*nep6max)
    character*80 nin(*)
    dimension ey(*),y(*),ily(*),e(*),ile(*),nd(*),na(*),nep(*),ep(nepdim,*),a(nadim,nepdim,*)
    dimension nbt(nrmax),ibt(nrmax),b(nwmax)
    call readtab1(nin,jin,zap,awp,lip,law,nr,nbt,ibt,ny,ey,y)
    if (law.ne.1) then
      write(*,*)' MF6/law=1 Error: law=',law
      stop
    endif
    call unpackibt(nr,nbt,ibt,ny,ily)
    call readtab2(nin,jin,c1,c2,lang,lep,nr,nbt,ibt,ne)
    if (lang.ne.1) then
      write(*,*)' MF6/law=1/lang=1 Error: law=',law,' lang=',lang
      stop
    endif
    call unpackibt(nr,nbt,ibt,ne,ile)
    do i=1,ne
      call readlist(nin,jin,c1,e(i),ndi,nai,nw,nepi,b)
      if (nw.gt.nwmax) then
        write(*,*)' MF6/law=1/lang=1 Error: LIST record too large. nep=',nepi,' na=',nai,' nw=',nw,' nwmax=',nwmax
        stop
      endif
      nd(i)=ndi
      na(i)=nai+1
      nep(i)=nepi
      ncyc=nai+2
      do j=1,nepi
        jb=ncyc*(j-1)+1
        ep(j,i)=b(jb)
        jb=jb+1
        a(1,j,i)=b(jb)
        if (nai.gt.0) then
          do k=1,nai
            a(k+1,j,i)=b(jb+k)
          enddo
        endif
      enddo
    enddo
    return
  end subroutine readmf6_law1_leg
!
! =================================================================================================================================
! subroutine readmf6_law1_Kalbach_Mann: Read data for MF6/LAW=1 Kalbach Mann formalism (law=1, lang=2, na=[1,2])
! =================================================================================================================================
!
!     zap: Product ZA identifier 1000*Zp+Ap (real*8)
!     awp: Product mass in neutron units (real*8)
!     lip: Product modifier flag (integer) for isomeric state, multiple emission
!      ny: number of incident energy points for y(E) (integer)
!      ey: incident energy points for y(E)
!          real*8 array ey=[ey(i),i=1..ny]
!       y: y(E) yields
!          real*8 array y=[y(i),i=1..ny]
!     ily: interpolation law for (E,y(E)) by intervals
!          integer array ily=[ily(i),i=1,ny-1]
!      ne: number of incident energy points for the Kalbach-Mann formalism (integer)
!       e: incident energy points
!          real*8 array e=[e(i), i=1..ne]
!     ile: incident energy interpolation law by intervals
!          integer array ile=[ile(i),i=1..ne-1]
!      nd: Number of discret energies
!          integer array nd=[nd(i),i=1..ne], 0 <= nd(i) <= nep(i)
!      na: Number of angular dimensions
!          integer array na=[na(i),i=1..ne],  na(i) <= 2
!  nepdim: Maximum number of secondary energies (integer)
!     nep: Number of secondary energies
!          integer array nep=[nep(i),i=1..ne]
!      ep: Secondary/outgoing energies
!          real*8 array ep=[ep(j,i), j=1..nep(i), i=1..ne], ne<=nemax for MF6
!     lep: interpolation law between secondary energies for j>nd(i) (integer)
!      f0: total emission probability f0(E',E)
!          real*8 array f0=[f0(j,i), j=1..nep(i), i=1..ne]
!       r: pre-compound fraction (ranges from 0.0 to 1.0)
!          real*8 array r=[r(j,i), j=1..nep(i), i=1..ne]
!       a: Kalbach-Mann a-parameter
!          real*8 array a=[a(j,i), j=1..nep(i), i=1..ne]
!
!   for j>nd(i), a continuous distribution is given with interpolation law=lep
!   between secondary energies, for j<=nd(i) discret distribution is provided.
!
  subroutine readmf6_law1_Kalbach_Mann(nin,jin,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nd,na,nepdim,nep,ep,lep,f0,r,a)
    implicit real*8 (a-h,o-z)
    parameter (nwmax=4*nep6max)
    character*80 nin(*)
    dimension ey(*),y(*),ily(*),e(*),ile(*),nd(*),na(*),nep(*),ep(nepdim,*),f0(nepdim,*),r(nepdim,*),a(nepdim,*)
    dimension nbt(nrmax),ibt(nrmax),b(nwmax)
    call readtab1(nin,jin,zap,awp,lip,law,nr,nbt,ibt,ny,ey,y)
    if (law.ne.1) then
      write(*,*)' MF6/law=1 Error: law=',law
      stop
    endif
    call unpackibt(nr,nbt,ibt,ny,ily)
    call readtab2(nin,jin,c1,c2,lang,lep,nr,nbt,ibt,ne)
    if (lang.ne.2) then
      write(*,*)' MF6/law=1/lang=2 Error: law=',law,' lang=',lang
      stop
    endif
    call unpackibt(nr,nbt,ibt,ne,ile)
    do i=1,ne
      call readlist(nin,jin,c1,e(i),ndi,nai,nw,nepi,b)
      if (nw.gt.nwmax) then
        write(*,*)' MF6/law=1/lang=2 Error: LIST record too large. nep=',nepi,' na=',nai,' nw=',nw,' nwmax=',nwmax
        stop
      endif
      nd(i)=ndi
      na(i)=nai
      nep(i)=nepi
      ncyc=nai+2
      do j=1,nepi
        jb=ncyc*(j-1)+1
        ep(j,i)=b(jb)
        f0(j,i)=b(jb+1)
        if (nai.eq.0) then
          r(j,i)=0.0d0
          a(j,i)=0.0d0
        elseif (nai.eq.1) then
          r(j,i)=b(jb+2)
          a(j,i)=0.0d0
        else
          r(j,i)=b(jb+2)
          a(j,i)=b(jb+3)
        endif
      enddo
    enddo
    return
  end subroutine readmf6_law1_Kalbach_Mann
!
! =================================================================================================================================
! subroutine readmf6_law1_Kalbach_Mann_na1: Read data for MF6/LAW=1 Kalbach Mann formalism (law=1, lang=2, na=1)
! =================================================================================================================================
!
!     zap: Product ZA identifier 1000*Zp+Ap (real*8)
!     awp: Product mass in neutron units (real*8)
!     lip: Product modifier flag (integer) for isomeric state, multiple emission
!      ny: number of incident energy points for y(E) (integer)
!      ey: incident energy points for y(E)
!          real*8 array ey=[ey(i),i=1..ny]
!       y: y(E) yields
!          real*8 array y=[y(i),i=1..ny]
!     ily: interpolation law for (E,y(E)) by intervals
!          integer array ily=[ily(i),i=1,ny-1]
!      ne: number of incident energy points for the Kalbach-Mann formalism (integer)
!       e: incident energy points
!          real*8 array e=[e(i), i=1..ne]
!     ile: incident energy interpolation law by intervals
!          integer array ile=[ile(i),i=1..ne-1]
!      nd: Number of discret energies
!          integer array nd=[nd(i),i=1..ne], 0 <= nd(i) <= nep(i)
!  nepdim: Maximum number of secondary energies (integer)
!     nep: Number of secondary energies
!          integer array nep=[nep(i),i=1..ne]
!      ep: Secondary/outgoing energies
!          real*8 array ep=[ep(j,i), j=1..nep(i), i=1..ne], ne<=nemax for MF6
!     lep: interpolation law between secondary energies for j>nd(i) (integer)
!      f0: total emission probability f0(E',E)
!          real*8 array f0=[f0(j,i), j=1..nep(i), i=1..ne]
!       r: pre-compound fraction (ranges from 0.0 to 1.0)
!          real*8 array r=[r(j,i), j=1..nep(i), i=1..ne]
!
!   for j>nd(i), a continuous distribution is given with interpolation law=lep
!   between secondary energies, for j<=nd(i) discret distribution is provided.
!
  subroutine readmf6_law1_Kalbach_Mann_na1(nin,jin,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nd,nepdim,nep,ep,lep,f0,r)
    implicit real*8 (a-h,o-z)
    parameter (ncyc=3, nwmax=ncyc*nep6max)
    character*80 nin(*)
    dimension ey(*),y(*),ily(*),e(*),ile(*),nd(*),nep(*),ep(nepdim,*),f0(nepdim,*),r(nepdim,*)
    dimension nbt(nrmax),ibt(nrmax),b(nwmax)
    call readtab1(nin,jin,zap,awp,lip,law,nr,nbt,ibt,ny,ey,y)
    if (law.ne.1) then
      write(*,*)' MF6/law=1 Error: law=',law
      stop
    endif
    call unpackibt(nr,nbt,ibt,ny,ily)
    call readtab2(nin,jin,c1,c2,lang,lep,nr,nbt,ibt,ne)
    if (lang.ne.2) then
      write(*,*)' MF6/law=1/lang=2 Error: law=',law,' lang=',lang
      stop
    endif
    call unpackibt(nr,nbt,ibt,ne,ile)
    do i=1,ne
      call readlist(nin,jin,c1,e(i),ndi,nai,nw,nepi,b)
      if (nai.ne.1) then
        write(*,*)' MF6/law=1/lang=2 na=',nai,' not equal 1 as expected'
        stop
      endif
      if (nw.gt.nwmax) then
        write(*,*)' MF6/law=1/lang=2 Error: LIST record too large. nep=',nepi,' na=',nai,' nw=',nw,' nwmax=',nwmax
        stop
      endif
      nd(i)=ndi
      nep(i)=nepi
      do j=1,nepi
        jb=ncyc*(j-1)+1
        ep(j,i)=b(jb)
        f0(j,i)=b(jb+1)
        r(j,i)=b(jb+2)
      enddo
    enddo
    return
  end subroutine readmf6_law1_Kalbach_Mann_na1
!
! =================================================================================================================================
! subroutine readmf6_law1_Kalbach_Mann_na2: Read data for MF6/LAW=1 Kalbach Mann formalism (law=1, lang=2, na=2)
! =================================================================================================================================
!
!     zap: Product ZA identifier 1000*Zp+Ap (real*8)
!     awp: Product mass in neutron units (real*8)
!     lip: Product modifier flag (integer) for isomeric state, multiple emission
!      ny: number of incident energy points for y(E) (integer)
!      ey: incident energy points for y(E)
!          real*8 array ey=[ey(i),i=1..ny]
!       y: y(E) yields
!          real*8 array y=[y(i),i=1..ny]
!     ily: interpolation law for (E,y(E)) by intervals
!          integer array ily=[ily(i),i=1,ny-1]
!      ne: number of incident energy points for the Kalbach-Mann formalism (integer)
!       e: incident energy points
!          real*8 array e=[e(i), i=1..ne]
!     ile: incident energy interpolation law by intervals
!          integer array ile=[ile(i),i=1..ne-1]
!      nd: Number of discret energies
!          integer array nd=[nd(i),i=1..ne], 0 <= nd(i) <= nep(i)
!  nepdim: Maximum number of secondary energies (integer)
!     nep: Number of secondary energies
!          integer array nep=[nep(i),i=1..ne]
!      ep: Secondary/outgoing energies
!          real*8 array ep=[ep(j,i), j=1..nep(i), i=1..ne], ne<=nemax for MF6
!     lep: interpolation law between secondary energies for j>nd(i) (integer)
!      f0: total emission probability f0(E',E)
!          real*8 array f0=[f0(j,i), j=1..nep(i), i=1..ne]
!       r: pre-compound fraction (ranges from 0.0 to 1.0)
!          real*8 array r=[r(j,i), j=1..nep(i), i=1..ne]
!       a: Kalbach-Mann a-parameter
!          real*8 array a=[a(j,i), j=1..nep(i), i=1..ne]
!
!   for j>nd(i), a continuous distribution is given with interpolation law=lep
!   between secondary energies, for j<=nd(i) discret distribution is provided.
!
  subroutine readmf6_law1_Kalbach_Mann_na2(nin,jin,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nd,nepdim,nep,ep,lep,f0,r,a)
    implicit real*8 (a-h,o-z)
    parameter (ncyc=4, nwmax=ncyc*nep6max)
    character*80 nin(*)
    dimension ey(*),y(*),ily(*),e(*),ile(*),nd(*),nep(*),ep(nepdim,*),f0(nepdim,*),r(nepdim,*),a(nepdim,*)
    dimension nbt(nrmax),ibt(nrmax),b(nwmax)
    call readtab1(nin,jin,zap,awp,lip,law,nr,nbt,ibt,ny,ey,y)
    if (law.ne.1) then
      write(*,*)' MF6/law=1 Error: law=',law
      stop
    endif
    call unpackibt(nr,nbt,ibt,ny,ily)
    call readtab2(nin,jin,c1,c2,lang,lep,nr,nbt,ibt,ne)
    if (lang.ne.2) then
      write(*,*)' MF6/law=1/lang=2 Error: law=',law,' lang=',lang
      stop
    endif
    call unpackibt(nr,nbt,ibt,ne,ile)
    do i=1,ne
      call readlist(nin,jin,c1,e(i),ndi,nai,nw,nepi,b)
      if (nai.ne.2) then
        write(*,*)' MF6/law=1/lang=2 na=',nai,' not equal 2 as expected'
        stop
      endif
      if (nw.gt.nwmax) then
        write(*,*)' MF6/law=1/lang=2 Error: LIST record too large. nep=',nepi,' na=',nai,' nw=',nw,' nwmax=',nwmax
        stop
      endif
      nd(i)=ndi
      nep(i)=nepi
      do j=1,nepi
        jb=ncyc*(j-1)+1
        ep(j,i)=b(jb)
        f0(j,i)=b(jb+1)
        r(j,i)=b(jb+2)
        a(j,i)=b(jb+3)
      enddo
    enddo
    return
  end subroutine readmf6_law1_Kalbach_Mann_na2
!
! =================================================================================================================================
! subroutine readmf6_law1_tab: Read tabulated distribution (law=1, lang=11-15)
! =================================================================================================================================
!
!      zap: Product ZA identifier 1000*Zp+Ap (real*8)
!      awp: Product mass in neutron units (real*8)
!      lip: Product modifier flag (integer) for isomeric state, multiple emission
!       ny: number of incident energy points for y(E) (integer)
!       ey: incident energy points for y(E)
!           real*8 array ey=[ey(i),i=1..ny]
!        y: y(E) yields
!           real*8 array y=[y(i),i=1..ny]
!      ily: interpolation law for (E,y(E)) by intervals
!           integer array ily=[ily(i),i=1,ny-1]
!       ne: number of incident energy points for tabulated data (integer)
!        e: incident energy points tabulated data
!           real*8 array e=[e(i), i=1..ne]
!      ile: incident energy interpolation law by intervals
!           integer array ile=[ile(i),i=1..ne-1]
!       nd: Number of discret energies
!           integer array nd=[nd(i),i=1..ne], 0 <= nd(i) <= nep(i)
!    nadim: Maximum number of cosines (integer)
!       na: Number of tabulated cosines
!           integer array na=[na(i),i=1..ne], na(i) <= nadim
!   nepdim: Maximum number of outgoing energies (integer)
!      nep: Number of secondary energies
!           integer array nep=[nep(i),i=1..ne], nep(i) <= nepdim
!       ep: Secondary/outgoing energies
!           real*8 array ep=[ep(j,i),j=1..nep(i),i=1..ne]
!      lep: interpolation law between secondary energies for j>nd(i) (integer)
!       f0: total emission probability f0(E',E)
!           real*8 array f0=[f0(j,i),j=1..nep(i),i=1..ne]
!      xmu: cosines values (ranges from -1.0 to 1.0)
!           real*8 array xmu=[xmu(j,k,i), j=1..nep(i), k=1..na(i), i=1..ne]
!      ymu: angular distribution
!           real*8 array ymu=[ymu(j,k,i), j=1..nep(i), k=1..na(i), i=1..ne]
!      lmu: interpolation law between cosines (integer)
!
!   for j>nd(i), a continuous distribution is given with interpolation law=lep
!   between secondary energies, for j<=nd(i) discret distribution is provided.
!
  subroutine readmf6_law1_tab(nin,jin,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nd,nadim,na,nepdim,nep,ep,lep,f0,xmu,ymu,lmu)
    implicit real*8 (a-h,o-z)
    parameter (nbmax=(2*nmu6max+2)*nep6max)
    character*80 nin(*)
    dimension ey(*),y(*),ily(*),e(*),ile(*),nd(*),na(*),nep(*)
    dimension ep(nepdim,*),f0(nepdim,*),xmu(nepdim,nadim,*),ymu(nepdim,nadim,*)
    dimension nbt(nrmax),ibt(nrmax),b(nbmax)
    call readtab1(nin,jin,zap,awp,lip,law,nr,nbt,ibt,ny,ey,y)
    if (law.ne.1) then
      write(*,*)' MF6/law=1 Error: law=',law
      stop
    endif
    call unpackibt(nr,nbt,ibt,ny,ily)
    call readtab2(nin,jin,c1,c2,lang,lep,nr,nbt,ibt,ne)
    if (lang.lt.11.or.lang.gt.15) then
      write(*,*)' Error: law=1/lang=11-15 inconsistency. law=',law,' lang=',lang
      stop
    endif
    call unpackibt(nr,nbt,ibt,ne,ile)
    do i=1,ne
      call readlist(nin,jin,c1,e(i),ndi,nai,nw,nepi,b)
      nai2=nai/2
      lmu=lang-10
      nd(i)=ndi
      na(i)=nai2
      nep(i)=nepi
      ncyc=nai+2
      do j=1,nepi
        jb=ncyc*(j-1)+1
        ep(j,i)=b(jb)
        jb=jb+1
        a0=b(jb)
        f0(j,i)=a0
        if (nai2.gt.0) then
          do k=1,nai2
            jb=jb+1
            xmu(j,k,i)=b(jb)
            jb=jb+1
            ymu(j,k,i)=2.0d0*b(jb)*a0
          enddo
        endif
      enddo
    enddo
    return
  end subroutine readmf6_law1_tab
!
! =================================================================================================================================
! subroutine readmf6_law2_leg: Read data for MF6/LAW=2 Legendre expansion (law=2, lang=0)
! =================================================================================================================================
!
!     zap: Product ZA identifier 1000*Zp+Ap (real*8)
!     awp: Product mass in neutron units (real*8)
!     lip: Product modifier flag (integer) for isomeric state, multiple emission
!      ny: number of incident energy points for y(E) (integer)
!      ey: incident energy points for y(E)
!          real*8 array ey=[ey(i),i=1..ny]
!       y: y(E) yields
!          real*8 array y=[y(i),i=1..ny]
!     ily: interpolation law for (E,y(E)) by intervals
!          integer array ily=[ily(i),i=1,ny-1]
!      ne: number of incident energy points for the Legendre expansion (integer)
!       e: incident energy points for the Legendre expansion
!          real*8 array e=[e(i), i=1..ne]
!     ile: incident energy interpolation law by intervals
!          integer array ile=[ile(i),i=1..ne-1]
!   nadim: Maximum number of Legendre coefficients (integer)
!      na: Number of Legendre coefficients
!          integer array na=[na(i),i=1..ne],  1 <= na(i) <= nadim
!       a: Legendre coefficients
!          real*8 array a=[a(k,i),k=1..na(i),i=1..ne]
!
  subroutine readmf6_law2_leg(nin,jin,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nadim,na,a)
    implicit real*8 (a-h,o-z)
    character*80 nin(*)
    dimension ey(*),y(*),ily(*),e(*),ile(*),na(*),a(nadim,*)
    dimension nbt(nrmax),ibt(nrmax),b(nadim)
    call readtab1(nin,jin,zap,awp,lip,law,nr,nbt,ibt,ny,ey,y)
    if (law.ne.2) then
      write(*,*)' MF6/law=2 Error: law=',law
      stop
    endif
    call unpackibt(nr,nbt,ibt,ny,ily)
    call readtab2(nin,jin,c1,c2,l1,l2,nr,nbt,ibt,ne)
    call unpackibt(nr,nbt,ibt,ne,ile)
    do i=1,ne
      call readlist(nin,jin,c1,e(i),lang,n1,nw,nli,b)
      if (lang.ne.0) then
        write(*,*)' MF6 law=2/lang=0 Error: law=',law,' lang=',lang
        stop
      endif
      na(i)=nli+1
      a(1,i)=1.0d0
      do k=1,nli
        a(k+1,i)=b(k)
      enddo
    enddo
    return
  end subroutine readmf6_law2_leg
!
! =================================================================================================================================
! subroutine readmf6_law2_tab: Read data for MF6/LAW=2 tabular representation (law=2, ltype=2/lang=[12,14])
! =================================================================================================================================
!
!     zap: Product ZA identifier 1000*Zp+Ap (real*8)
!     awp: Product mass in neutron units (real*8)
!     lip: Product modifier flag (integer) for isomeric state, multiple emission
!      ny: number of incident energy points for y(E) (integer)
!      ey: incident energy points for y(E)
!          real*8 array ey=[ey(i),i=1..ny]
!       y: y(E) yields
!          real*8 array y=[y(i),i=1..ny]
!     ily: interpolation law for (E,y(E)) by intervals
!          integer array ily=[ily(i),i=1,ny-1]
!      ne: number of incident energy points for tabulated data (integer)
!       e: incident energy points for tabulated data
!          real*8 array e=[e(i), i=1..ne]
!     ile: incident energy interpolation law by intervals
!          integer array ile=[ile(i),i=1..ne-1]
!   nadim: Maximum number of cosines. Maximum number of angular dimension (integer)
!     nmu: number of cosine by incident energy
!          integer array nmu=[nmu(i),1..ne]
!     xmu: set of cosine values by incident energy
!          real*8 xmu=[xmu(k,i),k=1..nmu(i),i=1..ne]
!     ymu: angle-distribution values by incident energy
!          real*8 ymu=[ymu(k,i),k=1,nmu(i),i=1..ne]
!    ilmu: cosine interpolation law by incident energy
!          integer ilmu=[ilmu(i),i=1..ne]
!
  subroutine readmf6_law2_tab(nin,jin,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nadim,nmu,xmu,ymu,ilmu)
    implicit real*8 (a-h,o-z)
    parameter (nbmax=2*nmu6max)
    character*80 nin(*)
    dimension ey(*),y(*),ily(*),e(*),ile(*),nmu(*),xmu(nadim,*),ymu(nadim,*),ilmu(*)
    dimension nbt(nrmax),ibt(nrmax),b(nbmax)
    call readtab1(nin,jin,zap,awp,lip,law,nr,nbt,ibt,ny,ey,y)
    if (law.ne.2) then
      write(*,*)' MF6/law=2 Error: law=',law
      stop
    endif
    call unpackibt(nr,nbt,ibt,ny,ily)
    call readtab2(nin,jin,c1,c2,l1,l2,nr,nbt,ibt,ne)
    call unpackibt(nr,nbt,ibt,ne,ile)
    do i=1,ne
      call readlist(nin,jin,c1,e(i),lang,n1,nw,nmui,b)
      if (lang.ne.12.or.lang.ne.14) then
        write(*,*)' MF6/law=2 lang=',lang,' Error: law=',law,' lang=',lang
        stop
      endif
      nmu(i)=nmui
      ilmu(i)=lang-10
      jb=0
      do k=1,nmui
        jb=jb+1
        xmu(k,i)=b(jb)
        jb=jb+1
        ymu(k,i)=b(jb)
      enddo
    enddo
    return
  end subroutine readmf6_law2_tab
!
! =================================================================================================================================
! subroutine readmf6_law2_mix: Read one MF6 subsection for law=2/ltype=3 (mix)
! =================================================================================================================================
!
!    1. Legendre expansion at lower energies (lang=0)
!    2. Tabulated distributions at higher energies (lang=12 or 14)
!
!    zap: Product ZA identifier 1000*Zp+Ap (real*8)
!    awp: Product mass in neutron units (real*8)
!    lip: Product modifier flag (integer) for isomeric state, multiple emission
!     ny: number of incident energy points for y(E) (integer)
!     ey: incident energy points for y(E)
!         real*8 array ey=[ey(i),i=1..ny]
!      y: y(E) yields
!         real*8 array y=[y(i),i=1..ny]
!    ily: interpolation law for (E,y(E)) by intervals
!         integer array ily=[ily(i),i=1,ny-1]
!    nel: number of incident energy points for the Legendre expansion (integer)
!     el: incident energy points for the Legendre expansion at lower energies
!         real*8 array el=[el(i), i=1..nel]
!   ilel: incident energy interpolation law by intervals at lower energies
!         integer array ilel=[ilel(i),i=1..nel-1]
!    nal: Maximum number of Legendre coefficients
!     na: Number of Legendre coefficients (integer)
!         integer array na=[na(i),i=1..nel],  1 <= na(i) <= nal
!      a: Legendre coefficients
!         real*8 array a=[a(k,i),k=1..na(i),i=1..nel]
!    neh: number of incident energy points for tabulated data (integer)
!     eh: incident energy points for tabulated data at higher energies
!         real*8 array eh=[eh(i), i=1..neh]
!   ileh: incident energy interpolation law by intervals at higher energies
!         integer array ileh=[ileh(i),i=1..neh-1]
!    nah: Maximum number of cosines (integer)
!    nmu: number of cosine by incident energy
!         integer array nmu=[nmu(i),1..neh], 1 < nmu(i) <= nah
!    xmu: set of cosine values by incident energy
!         real*8 xmu=[xmu(k,i),k=1..nmu(i),i=1..neh]
!    ymu: angle-distribution values by incident energy
!         real*8 ymu=[ymu(k,i),k=1,nmu(i),i=1..neh]
!   ilmu: cosine interpolation law by incident energy
!         integer ilmu=[ilmu(i),i=1..neh]
!
  subroutine readmf6_law2_mix(nin,jin,zap,awp,lip,ny,ey,y,ily,nel,el,ilel,nal,na,a,neh,eh,ileh,nah,nmu,xmu,ymu,ilmu)
    implicit real*8 (a-h,o-z)
    parameter (eps=1.0d-6, nbmax=max(ncmax,2*nmu6max))
    character*80 nin(*)
    dimension ey(*),y(*),ily(*)
    dimension el(*),ilel(*),na(*),a(nal,*)
    dimension eh(*),ileh(*),nmu(*),xmu(nah,*),ymu(nah,*),ilmu(*)
    dimension nbt(nrmax),ibt(nrmax),jbt(ne6max),b(nbmax)
    call readtab1(nin,jin,zap,awp,lip,law,nr,nbt,ibt,ny,ey,y)
    if (law.ne.2) then
      write(*,*)' MF6/law=2 Error: law=',law
      stop
    endif
    call unpackibt(nr,nbt,ibt,ny,ily)
    call readtab2(nin,jin,c1,c2,l1,l2,nr,nbt,ibt,ne)
    call unpackibt(nr,nbt,ibt,ne,jbt)
    lang0=0
    nel=0
    do while (lang0.eq.0.and.nel.lt.ne)
      jin0=jin
      call readlist(nin,jin,c1,ei,lang,n1,nw,nli,b)
      if (lang.eq.0) then
        nel=nel+1
        el(nel)=ei
        na(nel)=nli+1
        a(1,nel)=1.0d0
        do k=1,nli
          a(k+1,nel)=b(k)
        enddo
      elseif (lang.eq.12.or.lang.eq.14) then
        do i=1,nel-1
          ilel(i)=jbt(i)
        enddo
      else
        write(*,*)' MF6/law=2 Error: Inconsistency found for ltype=3(mix) at low energies. lang=',lang
        stop
      endif
      lang0=lang
    enddo
    if (nel.lt.ne) then
      lang0=lang
      neh=0
      jin=jin0
      do while ((lang0.eq.12.or.lang0.eq.14).and.(neh+nel).lt.ne)
        call readlist(nin,jin,c1,ei,lang,n1,nw,nmui,b)
        if (lang.eq.12.or.lang.eq.14) then
          neh=neh+1
          eh(neh)=ei
          nmu(neh)=nmui
          ilmu(neh)=lang-10
          jb=0
          do k=1,nmui
            jb=jb+1
            xmu(k,neh)=b(jb)
            jb=jb+1
            ymu(k,neh)=b(jb)
          enddo
        else
          write(*,*)' MF6/law=2 Error: Inconsistency found for ltype=3(mix) at low energies. lang=',lang
          stop
        endif
        lang0=lang
      enddo
      do i=1,neh-1
        ileh(i)=jbt(nel+i)
      enddo
      if (nel.eq.0) then
        write(*,*)' Warning: Not a mixed representation, only tabular representation is used'
        write(*,*)'          Legendre expansion data and arrays have no meaning'
      elseif (abs(eh(1)-el(nel)).gt.(eps*el(nel))) then
        write(*,*)' Warning: Energy discontinuity between Legendre expansion and tabulated data'
      endif
    else
      write(*,*)' Warning: Not a mixed representation, only Legendre expansion is used'
      write(*,*)'          Tabulated data and arrays have no meaning'
    endif
    return
  end subroutine readmf6_law2_mix
!
! =================================================================================================================================
! subroutine readmf6_law6_phase_space: Read one MF6 subsection for law=6 (n-body phase-space)
! =================================================================================================================================
!
!    zap: Product ZA identifier 1000*Zp+Ap (real*8)
!    awp: Product mass in neutron units (real*8)
!    lip: Product modifier flag (integer) for isomeric state, multiple emission
!     ny: number of incident energy points for y(E) (integer)
!     ey: incident energy points for y(E)
!         real*8 array ey=[ey(i),i=1..ny]
!      y: y(E) yields
!         real*8 array y=[y(i),i=1..ny]
!    ily: interpolation law for (E,y(E)) by intervals
!         integer array ily=[ily(i),i=1,ny-1]
!   npsx: number of particles n distributed according to the phase-space law (integer)
!   apsx: total mass M in neutron unit of the n particles (real*8)
!
  subroutine readmf6_law6_phase_space(nin,jin,zap,awp,lip,ny,ey,y,ily,npsx,apsx)
    implicit real*8 (a-h,o-z)
    character*80 nin(*)
    dimension ey(*),y(*),ily(*)
    dimension nbt(nrmax),ibt(nrmax)
    call readtab1(nin,jin,zap,awp,lip,law,nr,nbt,ibt,ny,ey,y)
    if (law.ne.6) then
      write(*,*)' MF6/law=6 Error: law=',law
      stop
    endif
    call unpackibt(nr,nbt,ibt,ny,ily)
    call readcont(nin,jin,apsx,c2,l1,l2,n1,npsx,mat,mf,mt,ns)
    return
  end subroutine readmf6_law6_phase_space
!
! =================================================================================================================================
! subroutine readmf6_law7_lab: Read one MF6 subsection for law=7 (Laboratory Angle-Energy Distribution)
! =================================================================================================================================
!
!      zap: Product ZA identifier 1000*Zp+Ap (real*8)
!      awp: Product mass in neutron units (real*8)
!      lip: Product modifier flag (integer) for isomeric state, multiple emission
!       ny: number of incident energy points for y(E) (integer)
!       ey: incident energy points for y(E)
!           real*8 array ey=[ey(i),i=1..ny]
!        y: y(E) yields
!           real*8 array y=[y(i),i=1..ny]
!      ily: interpolation law for (E,y(E)) by intervals
!           integer array ily=[ily(i),i=1,ny-1]
!       ne: number of incident energy points for law7 (integer)
!        e: incident energy points
!           real*8 array e=[e(i), i=1..ne]
!      ile: incident energy interpolation law by intervals
!           integer array ile=[ile(i),i=1..ne-1]
!    nadim: Maximum number of cosines (integer)
!      nmu: Number of tabulated cosines
!           integer array nmu=[nmu(i),i=1..ne],  1 < nmu(i) <= nadim
!      xmu: cosines values by incident energy (ranges from -1.0 to 1.0)
!           real*8 array xmu=[xmu(k,i),j=1..nmu(i), i=1..ne]
!     ilmu: cosine interpolation law by interval
!           real*8 array ilmu=[ilmu(j,i),j=1..nmu(i)-1, i=1..ne]
!   nepdim: maximum number of outgoing energies (integer)
!      nep: Number of secondary energies by cosine and incident energy
!           integer array nep=[nep(j,i),j=1..nmu(i), i=1..ne]
!       ep: Secondary/outgoing energies
!           real*8 array ep=[ep(k,j,i),k=1..nep(j,i),j=1..nmu(i),i=1..ne]
!        f: Angle-energy distribution
!           real*8 array f=[f(k,j,i),k=1..nep(j,i),j=1..nmu(i),i=1..ne]
!     ilef: secondary energy interpolation law
!           integer array ilef=[ilef(k,j,i),k=1..nep(j,i),j=1..nmu(i),i=1..ne]
!
  subroutine readmf6_law7_lab(nin,jin,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nadim,nmu,xmu,ilmu,nepdim,nep,ep,f,ilef)
    implicit real*8 (a-h,o-z)
    character*80 nin(*)
    dimension ey(*),y(*),ily(*),e(*),ile(*),nmu(*),xmu(nadim,*),ilmu(nadim,*)
    dimension nep(nadim,*),ep(nadim,nepdim,*),f(nadim,nepdim,*),ilef(nadim,nepdim,*)
    dimension nbt(nrmax),ibt(nrmax),x1(nepdim),y1(nepdim),jbtep(nepdim),jbtna(nadim)
    call readtab1(nin,jin,zap,awp,lip,law,nr,nbt,ibt,ny,ey,y)
    if (law.ne.7) then
      write(*,*)' Error: law=7 inconsistency. law=',law
      stop
    endif
    call unpackibt(nr,nbt,ibt,ny,ily)
    call readtab2(nin,jin,c1,c2,l1,l2,nr,nbt,ibt,ne)
    call unpackibt(nr,nbt,ibt,ne,ile)
    do i=1,ne
      call readtab2(nin,jin,c1,e(i),l1,l2,nr,nbt,ibt,nmui)
      nmu(i)=nmui
      call unpackibt(nr,nbt,ibt,nmui,jbtna)
      do j=1,nmui-1
        ilmu(j,i)=jbtna(j)
      enddo
      do j=1,nmui
        call readtab1(nin,jin,c1,xmui,l1,l2,nr,nbt,ibt,nepi,x1,y1)
        xmu(j,i)=xmui
        nep(j,i)=nepi
        call unpackibt(nr,nbt,ibt,nepi,jbtep)
        do k=1,nepi-1
          ep(k,j,i)=x1(k)
          f(k,j,i)=y1(k)
          ilef(k,j,i)=jbtep(k)
        enddo
        ep(nepi,j,i)=x1(nepi)
        f(nepi,j,i)=y1(nepi)
      enddo
    enddo
    return
  end subroutine readmf6_law7_lab
!
! =================================================================================================================================
! subroutine wrtmf6mt_head: Read MF6/MT section head
! =================================================================================================================================
!    za: ZA number
!   awr: Atomic weight in neutron mass units
!    jp: Integer trigger for treatment of prompt fission neutron and photon
!   lct: Reference system for secondary energy and angles
!    nk: Number of subsections in this section (number of described reaction products)
!
  subroutine wrtmf6mt_head(nou,jou,mat,mt,ns,za,awr,jp,lct,nk)
    implicit real*8 (a-h, o-z)
    character*80 nou(*)
    if (mt.ne.18.and.jp.gt.0) then
      write(*,*)' Error: jp=',jp,' is not allowed for mt=',mt,' in MF6'
      stop
    elseif (nk.gt.nkmax) then
      write(*,*)' Error : nk=',nk,' too many subsections in MF6. nkmax=',nkmax
      stop
    endif
    mf=6
    n2=0
    call wrtcont(nou,jou,mat,mf,mt,ns,za,awr,jp,lct,nk,n2)
    return
  end subroutine wrtmf6mt_head
!
! =================================================================================================================================
! subroutine wrtmf6law0_3_4_neg: Read data for MF6/LAWs=0,3,4 for jp>0 and 0,-5, -15 for jp<0
!                                Only yields or P(n,E) are provided, no distribution.
! =================================================================================================================================
!   law: MF6/law (0,3,4,-5,-15)
!   zap: Product ZA identifier 1000*Zp+Ap (real*8)
!   awp: Product mass in neutron units (real*8)
!   lip: Product modifier flag (integer) for isomeric state, multiple emission
!    ny: number of incident energies (integer)
!    ey: incident energies E(i)
!        real*8 array ey=[ey(i),i=1..ny]
!     y: y(E) yields for law=0,3,4 jp=0
!        P(n,E) fission particle distribution for law=0,-5,-15 jp>0
!        real*8 array y=[y(i),i=1..ny]
!   ily: interpolation law for (ey,y(ey)) by intervals
!        integer array ily=[ily(i),i=1,ny-1]
!
  subroutine wrtmf6_law0_3_4_neg(nou,jou,mat,mt,ns,law,zap,awp,lip,ny,ey,y,ily)
    implicit real*8 (a-h, o-z)
    character*80 nou(*)
    dimension ey(*),y(*),ily(*)
    dimension nbt(nrmax),ibt(nrmax)
    if (law.ne.0.or.law.ne.3.or.law.ne.4.or.law.ne.-5.or.law.ne.-15) then
      write(*,*)' MF6 Error: law=',law,' not equal 0,3,4,-5 or -15'
      stop
    elseif (ny.gt.ne6max) then
      write(*,*)' MF6/law=',law,' Error: Too many incident energy points in MF6. n=',ny,' nmax=',ne6max
      stop
    endif
    call packibt(ny,ily,nr,nbt,ibt)
    mf=6
    call wrtab1(nou,jou,mat,mf,mt,ns,zap,awp,lip,law,nr,nbt,ibt,ny,ey,y)
    return
  end subroutine wrtmf6_law0_3_4_neg
!
! =================================================================================================================================
! subroutine wrtmf6_law1_leg: Write data for MF6/LAW=1 Legendre expansion (law=1, lang=1)
! =================================================================================================================================
!
!      zap: Product ZA identifier 1000*Zp+Ap (real*8)
!      awp: Product mass in neutron units (real*8)
!      lip: Product modifier flag (integer) for isomeric state, multiple emission
!       ny: number of incident energy points for y(E) (integer)
!       ey: incident energy points for y(E)
!           real*8 array ey=[ey(i),i=1..ny]
!        y: y(E) yields
!           real*8 array y=[y(i),i=1..ny]
!      ily: interpolation law for (E,y(E)) by intervals
!           integer array ily=[ily(i),i=1,ny-1]
!       ne: number of incident energy points for the Legendre expansion (integer)
!        e: incident energy points for the Legendre expansion
!           real*8 array e=[e(i), i=1..ne]
!      ile: incident energy interpolation law by intervals
!           integer array ile=[ile(i),i=1..ne-1]
!       nd: Number of discret energies
!           integer array nd=[nd(i),i=1..ne], 0 <= nd(i) <= nep(i)
!    nadim: Maximum number of Legendre coefficients (integer)
!       na: Number of Legendre coefficients
!           integer array na=[na(i),i=1..ne],  1 <= na(i) <= ncmax=nlmax+1
!   nepdim: Maximum number of secondary energies (integer)
!      nep: Number of secondary energies
!           integer array nep=[nep(i),i=1..ne]
!       ep: Secondary/outgoing energies
!           real*8 array ep=[ep(j,i), j=1..nep(i), i=1..ne], ne<=nemax for MF6
!      lep: interpolation law between secondary energies for j>nd(i) (integer)
!        a: Legendre coefficients
!           real*8 array a=[a(k,j,i),k=1..na(i),j=1..nep(i),i=1..ne]
!
!   for j>nd(i), a continuous distribution is given with interpolation law=lep
!   between secondary energies, for j<=nd(i) discret distribution is provided.
!
  subroutine wrtmf6_law1_leg(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nd,nadim,na,nepdim,nep,ep,lep,a)
    implicit real*8 (a-h,o-z)
    parameter (nwmax=(nlmax+2)*nep6max)
    character*80 nou(*)
    dimension ey(*),y(*),ily(*),e(*),ile(*),nd(*),na(*),nep(*),ep(nepdim,*),a(nadim,nepdim,*)
    dimension nbt(nrmax),ibt(nrmax),b(nwmax)
    if (ny.gt.ne6max) then
      write(*,*)' MF6/law=1/lang=1 Warning: Too many incident energy points. ny=',ny,' ne6max=',ne6max
    endif
    if (ne.gt.ne6max) then
      write(*,*)' MF6/law=1/lang=1 Warning: Too many incident energy points. ne=',ne,' ne6max=',ne6max
    endif
    if (nepdim.gt.nep6max) then
      write(*,*)' MF6/law=1/lang=1 Warning: Too many outgoing energies. nep=',nepdim,' nep6max=',nep6max
    endif
    if (nadim.gt.ncmax) then
      write(*,*)' MF6/law=1/lang=1 Error: Too many Legendre coefficients. nc=',nadim,' ncmax=',ncmax
      stop
    endif
    if ((nadim+1)*nepdim.gt.nwmax) then
      write(*,*)' MF6/law=1/lang=1 Error: Too large LIST record. nwmax=',nwmax
      write(*,*)' na=',nadim,' nl=',nadim-1,' nep=',nepdim,' nw=',(nadim+1)*nepdim
      stop
    endif
    if (ne*nadim*nepdim.gt.npmax) then
      write(*,*)' MF6/law=1/lang=1 Error: Too large arrays. npmax=',npmax
      write(*,*)' ne=',ne,' na=',nadim,' nep=',nepdim,' np=',ne*nadim*nepdim
      stop
    endif
    if (lep.lt.1.or.lep.gt.5) then
      write(*,*)' MF6/law=1/lang=1 Error: Incorrect secondary energy interpolation law (lep)=',lep,' (1-5 allowed)'
      stop
    endif
    call packibt(ny,ily,nr,nbt,ibt)
    mf=6
    law=1
    lang=1
    c1=0.0d0
    c2=0.0d0
    l1=0
    l2=0
    n1=0
    n2=0
    call wrtab1(nou,jou,mat,mf,mt,ns,zap,awp,lip,law,nr,nbt,ibt,ny,ey,y)
    call packibt(ne,ile,nr,nbt,ibt)
    call wrtab2(nou,jou,mat,mf,mt,ns,c1,c2,lang,lep,nr,nbt,ibt,ne)
    do i=1,ne
      ndi=nd(i)
      nai=na(i)
      nli=nai-1
      nepi=nep(i)
      ncyc=nli+2
      nw=nepi*ncyc
      do j=1,nepi
        jb=ncyc*(j-1)+1
        b(jb)=ep(j,i)
        do k=1,nai
          b(jb+k)=a(k,j,i)
        enddo
      enddo
      call wrtlist(nou,jou,mat,mf,mt,ns,c1,e(i),ndi,nli,nw,nepi,b)
    enddo
    return
  end subroutine wrtmf6_law1_leg
!
! =================================================================================================================================
! subroutine wrtmf6_law1_leg_law: Write data for MF6/LAW=1 Legendre expansion (law=1, lang=1)
! =================================================================================================================================
!
!   Similar to wrtmf6_law1_leg, but each interpolation scheme is represented by only one interpolation law.
!    lawy: interpolation law for [E, y(E)]
!    lawa: interpolation law for [E, a(E',E]]
!   The rest of parameters have the same meaning as before
!
  subroutine wrtmf6_law1_leg_law(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,lawy,ne,e,lawa,nd,nadim,na,nepdim,nep,ep,lep,a)
    implicit real*8 (a-h,o-z)
    character*80 nou(*)
    dimension ey(*),y(*),e(*),nd(*),na(*),nep(*),ep(nepdim,*),a(nadim,nepdim,*)
    dimension ily(ny),ile(ne)
    do i=1,ny-1
      ily(i)=lawy
    enddo
    do i=1,ne-1
      ile(i)=lawa
    enddo
    call wrtmf6_law1_leg(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nd,nadim,na,nepdim,nep,ep,lep,a)
    return
  end subroutine wrtmf6_law1_leg_law
!
! =================================================================================================================================
! subroutine wrtmf6_law1_leg_lin: Write data for MF6/LAW=1 Legendre expansion (law=1, lang=1)
! =================================================================================================================================
!
!   Similar to wrtmf6_law1_leg, but all interpolation schema are lin-lin, therefore they are not input parameters.
!   The rest of parameters have the same meaning as before
!
  subroutine wrtmf6_law1_leg_lin(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ne,e,nd,nadim,na,nepdim,nep,ep,lep,a)
    implicit real*8 (a-h,o-z)
    character*80 nou(*)
    dimension ey(*),y(*),e(*),nd(*),na(*),nep(*),ep(nepdim,*),a(nadim,nepdim,*)
    dimension ily(ny),ile(ne)
    do i=1,ny-1
      ily(i)=2
    enddo
    do i=1,ne-1
      ile(i)=2
    enddo
    call wrtmf6_law1_leg(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nd,nadim,na,nepdim,nep,ep,lep,a)
    return
  end subroutine wrtmf6_law1_leg_lin
!
! =================================================================================================================================
! subroutine wrtmf6_law1_iso: Write data for MF6/LAW=1 Legendre expansion, but full isotropic (law=1, lang=1, na=0)
! =================================================================================================================================
!
!      zap: Product ZA identifier 1000*Zp+Ap (real*8)
!      awp: Product mass in neutron units (real*8)
!      lip: Product modifier flag (integer) for isomeric state, multiple emission
!       ny: number of incident energy points for y(E) (integer)
!       ey: incident energy points for y(E)
!           real*8 array ey=[ey(i),i=1..ny]
!        y: y(E) yields
!           real*8 array y=[y(i),i=1..ny]
!      ily: interpolation law for (E,y(E)) by intervals
!           integer array ily=[ily(i),i=1,ny-1]
!       ne: number of incident energy points for the Legendre expansion (integer)
!        e: incident energy points for the Legendre expansion
!           real*8 array e=[e(i), i=1..ne]
!      ile: incident energy interpolation law by intervals
!           integer array ile=[ile(i),i=1..ne-1]
!       nd: Number of discret energies
!           integer array nd=[nd(i),i=1..ne], 0 <= nd(i) <= nep(i)
!   nepdim: Maximun number of secindary energies (integer)
!      nep: Number of secondary energies
!           integer array nep=[nep(i),i=1..ne]
!       ep: Secondary/outgoing energies
!           real*8 array ep=[ep(j,i), j=1..nep(i), i=1..ne], ne<=ne6max for MF6
!      lep: interpolation law between secondary energies for j>nd(i) (integer)
!       f0: total emission probability f0(E',E)
!           real*8 array f0=[f0(j,i), j=1..nep(i), i=1..ne]
!
!   for j>nd(i), a continuous distribution is given with interpolation law=lep
!   between secondary energies, for j<=nd(i) discret distribution is provided.
!
  subroutine wrtmf6_law1_iso(nou,jou,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nd,nepdim,nep,ep,lep,f0)
    implicit real*8 (a-h,o-z)
    parameter (nwmax=2*nep6max)
    character*80 nou(*)
    dimension ey(*),y(*),ily(*),e(*),ile(*),nd(*),nep(*),ep(nepdim,*),f0(nepdim,*)
    dimension nbt(nrmax),ibt(nrmax),b(nwmax)
    if (ny.gt.ne6max) then
      write(*,*)' MF6/law=1 (iso) Warning: Too many incident energy points. ny=',ny,' ne6max=',ne6max
    endif
    if (ne.gt.ne6max) then
      write(*,*)' MF6/law=1 (iso) Warning: Too many incident energy points. ne=',ne,' ne6max=',ne6max
    endif
    if (nepdim.gt.nep6max) then
      write(*,*)' MF6/law=1 (iso) Error: Too many outgoing energies. nep=',nepdim,' nep6max=',nep6max
      stop
    endif
    if (ne*nepdim.gt.npmax) then
      write(*,*)' MF6/law=1 (iso) Error: Arrays too large. npmax=',npmax
      write(*,*)' ne=',ne,' nep=',nepdim,' ne*nep=',ne*nepdim
      stop
    endif
    if (lep.lt.1.or.lep.gt.5) then
      write(*,*)' MF6/law=1 (iso) Error: Incorrect secondary energy interpolation law (lep)=',lep,' (1-5 allowed)'
      stop
    endif
    call packibt(ny,ily,nr,nbt,ibt)
    mf=6
    law=1
    lang=1
    na=0
    ncyc=2
    c1=0.0d0
    c2=0.0d0
    l1=0
    l2=0
    n1=0
    n2=0
    call wrtab1(nou,jou,mat,mf,mt,ns,zap,awp,lip,law,nr,nbt,ibt,ny,ey,y)
    call packibt(ne,ile,nr,nbt,ibt)
    call wrtab2(nou,jou,mat,mf,mt,ns,c1,c2,lang,lep,nr,nbt,ibt,ne)
    do i=1,ne
      ndi=nd(i)
      nepi=nep(i)
      nw=nepi*ncyc
      do j=1,nepi
        jb=ncyc*(j-1)+1
        b(jb)=ep(j,i)
        b(jb+1)=f0(j,i)
      enddo
      call wrtlist(nou,jou,mat,mf,mt,ns,c1,e(i),ndi,na,nw,nepi,b)
    enddo
    return
  end subroutine wrtmf6_law1_iso
!
! =================================================================================================================================
! subroutine wrtmf6_law1_iso_law: Write data for MF6/LAW=1 Legendre expansion (law=1, isotropic)
! =================================================================================================================================
!
!   Similar to wrtmf6_law1_iso, but each interpolation scheme is represented by only one interpolation law.
!     lawy: interpolation law for [E, y(E)]
!    lawf0: interpolation law for [E, f0(E',E]]
!   The rest of parameters have the same meaning as before
!
  subroutine wrtmf6_law1_iso_law(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,lawy,ne,e,lawf0,nd,nepdim,nep,ep,lep,f0)
    implicit real*8 (a-h,o-z)
    character*80 nou(*)
    dimension ey(*),y(*),e(*),nd(*),nep(*),ep(nepdim,*),f0(nepdim,*)
    dimension ily(ny),ile(ne)
    do i=1,ny-1
      ily(i)=lawy
    enddo
    do i=1,ne-1
      ile(i)=lawf0
    enddo
    call wrtmf6_law1_iso(nou,jou,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nd,nepdim,nep,ep,lep,f0)
    return
  end subroutine wrtmf6_law1_iso_law
!
! =================================================================================================================================
! subroutine wrtmf6_law1_iso_lin: Write data for MF6/LAW=1 Legendre expansion (law=1, isotropic)
! =================================================================================================================================
!
!   Similar to wrtmf6_law1_leg_iso, but all interpolation schema are lin-lin, therefore they are not input parameters.
!   The rest of parameters have the same meaning as before
!
  subroutine wrtmf6_law1_iso_lin(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ne,e,nd,nepdim,nep,ep,lep,f0)
    implicit real*8 (a-h,o-z)
    character*80 nou(*)
    dimension ey(*),y(*),e(*),nd(*),nep(*),ep(nepdim,*),f0(nepdim,*)
    dimension ily(ny),ile(ne)
    do i=1,ny-1
      ily(i)=2
    enddo
    do i=1,ne-1
      ile(i)=2
    enddo
    call wrtmf6_law1_iso(nou,jou,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nd,nepdim,nep,ep,lep,f0)
    return
  end subroutine wrtmf6_law1_iso_lin
!
! =================================================================================================================================
! subroutine wrtmf6_law1_Kalbach_Mann: Write data for MF6/LAW=1 Kalbach Mann formalism (law=1, lang=2)
! =================================================================================================================================
!
!     zap: Product ZA identifier 1000*Zp+Ap (real*8)
!     awp: Product mass in neutron units (real*8)
!     lip: Product modifier flag (integer) for isomeric state, multiple emission
!      ny: number of incident energy points for y(E) (integer)
!      ey: incident energy points for y(E)
!          real*8 array ey=[ey(i),i=1..ny]
!       y: y(E) yields
!          real*8 array y=[y(i),i=1..ny]
!     ily: interpolation law for (E,y(E)) by intervals
!          integer array ily=[ily(i),i=1,ny-1]
!      ne: number of incident energy points for the Kalbach-Mann formalism (integer)
!       e: incident energy points
!          real*8 array e=[e(i), i=1..ne]
!     ile: incident energy interpolation law by intervals
!          integer array ile=[ile(i),i=1..ne-1]
!      nd: Number of discret energies
!          integer array nd=[nd(i),i=1..ne], 0 <= nd(i) <= nep(i)
!      na: Number of angular dimensions
!          integer array na=[na(i),i=1..ne],  na(i) <= 2
!  nepdim: Maximum number of secondary energies (integer)
!     nep: Number of secondary energies
!          integer array nep=[nep(i),i=1..ne]
!      ep: Secondary/outgoing energies
!          real*8 array ep=[ep(j,i), j=1..nep(i), i=1..ne], ne<=nemax for MF6
!     lep: interpolation law between secondary energies for j>nd(i) (integer)
!      f0: total emission probability f0(E',E)
!          real*8 array f0=[f0(j,i), j=1..nep(i), i=1..ne]
!       r: pre-compound fraction (ranges from 0.0 to 1.0)
!          real*8 array r=[r(j,i), j=1..nep(i), i=1..ne]
!       a: Kalbach-Mann a-parameter
!          real*8 array a=[a(j,i), j=1..nep(i), i=1..ne]
!
!   for j>nd(i), a continuous distribution is given with interpolation law=lep
!   between secondary energies, for j<=nd(i) discret distribution is provided.
!
  subroutine wrtmf6_law1_Kalbach_Mann(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nd,na,nepdim,nep,ep,lep,f0,r,a)
    implicit real*8 (a-h,o-z)
    parameter (nwmax=4*nep6max)
    character*80 nou(*)
    dimension ey(*),y(*),ily(*),e(*),ile(*),nd(*),na(*),nep(*),ep(nepdim,*),f0(nepdim,*),r(nepdim,*),a(nepdim,*)
    dimension nbt(nrmax),ibt(nrmax),b(nwmax)
    if (ny.gt.ne6max) then
      write(*,*)' MF6/law=1/lang=2 Warning: Too many incident energy points. ny=',ny,' ne6max=',ne6max
    endif
    if (ne.gt.ne6max) then
      write(*,*)' MF6/law=1/lang=2 Warning: Too many incident energy points. ne=',ne,' ne6max=',ne6max
    endif
    if (nepdim.gt.nep6max) then
      write(*,*)' MF6/law=1/lang=2 Error: Too many outgoing energies. nep=',nepdim,' nep6max=',nep6max
      stop
    endif
    if (ne*nepdim.gt.npmax) then
      write(*,*)' MF6/law=1/lang=2 Error: Arrays too large. npmax=',npmax
      write(*,*)' ne=',ne,' nep=',nepdim,' ne*nep=',ne*nepdim
      stop
    endif
    if (lep.lt.1.or.lep.gt.5) then
      write(*,*)' MF6/law=1 lang=2 Error: Incorrect secondary energy interpolation law (lep)=',lep,' (1-5 allowed)'
      stop
    endif
    call packibt(ny,ily,nr,nbt,ibt)
    mf=6
    law=1
    lang=2
    c1=0.0d0
    c2=0.0d0
    l1=0
    l2=0
    n1=0
    n2=0
    call wrtab1(nou,jou,mat,mf,mt,ns,zap,awp,lip,law,nr,nbt,ibt,ny,ey,y)
    call packibt(ne,ile,nr,nbt,ibt)
    call wrtab2(nou,jou,mat,mf,mt,ns,c1,c2,lang,lep,nr,nbt,ibt,ne)
    do i=1,ne
      ndi=nd(i)
      nai=na(i)
      nepi=nep(i)
      ncyc=nai+2
      nw=nepi*ncyc
      if (nai.gt.2) then
        write(*,*)' MF6/law=1 lang=2 Error: Number of angular parameters na>2 for Kalbach-Mann formalism'
        stop
      endif
      do j=1,nepi
        jb=ncyc*(j-1)+1
        b(jb)=ep(j,i)
        b(jb+1)=f0(j,i)
        if (nai.eq.1) then
          b(jb+2)=r(j,i)
        elseif (nai.eq.2) then
          b(jb+2)=r(j,i)
          b(jb+3)=a(j,i)
        endif
      enddo
      call wrtlist(nou,jou,mat,mf,mt,ns,c1,e(i),ndi,nai,nw,nepi,b)
    enddo
    return
  end subroutine wrtmf6_law1_Kalbach_Mann
!
! =================================================================================================================================
! subroutine wrtmf6_law1_Kalbach_Mann_law: Write data for MF6/LAW=1 Legendre expansion (law=1, lang=2)
! =================================================================================================================================
!
!   Similar to wrtmf6_law1_Kalbach_Mann, but each interpolation scheme is represented by only one interpolation law.
!    lawy: interpolation law for [E, y(E)]
!    lawf: interpolation law for [E, (f0(E',E),r(E',E),a(E',E))]
!   The rest of parameters have the same meaning as before
!
  subroutine wrtmf6_law1_Kalbach_Mann_law(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,lawy,ne,e,lawf,nd,na,nepdim,nep,ep,lep,f0,r,a)
    implicit real*8 (a-h,o-z)
    character*80 nou(*)
    dimension ey(*),y(*),e(*),nd(*),na(*),nep(*),ep(nepdim,*),f0(nepdim,*),r(nepdim,*),a(nepdim,*)
    dimension ily(ny),ile(ne)
    do i=1,ny-1
      ily(i)=lawy
    enddo
    do i=1,ne-1
      ile(i)=lawf
    enddo
    call wrtmf6_law1_Kalbach_Mann(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nd,na,nepdim,nep,ep,lep,f0,r,a)
    return
  end subroutine wrtmf6_law1_Kalbach_Mann_law

! =================================================================================================================================
! subroutine wrtmf6_law1_Kalbach_Mann_lin: Write data for MF6/LAW=1 Legendre expansion (law=1, lang=1, isotropic)
! =================================================================================================================================
!
!   Similar to wrtmf6_law1_Kalbach_Mann, but all interpolation schema are lin-lin, therefore they are not input parameters.
!   The rest of parameters have the same meaning as before
!
  subroutine wrtmf6_law1_Kalbach_Mann_lin(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ne,e,nd,na,nepdim,nep,ep,lep,f0,r,a)
    implicit real*8 (a-h,o-z)
    character*80 nou(*)
    dimension ey(*),y(*),e(*),nd(*),na(*),nep(*),ep(nepdim,*),f0(nepdim,*),r(nepdim,*),a(nepdim,*)
    dimension ily(ny),ile(ne)
    do i=1,ny-1
      ily(i)=2
    enddo
    do i=1,ne-1
      ile(i)=2
    enddo
    lep=2
    call wrtmf6_law1_Kalbach_Mann(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nd,na,nepdim,nep,ep,lep,f0,r,a)
    return
  end subroutine wrtmf6_law1_Kalbach_Mann_lin
!
! =================================================================================================================================
! subroutine wrtmf6_law1_Kalbach_Mann_na1: Write data for MF6/LAW=1 Kalbach Mann formalism (law=1, lang=2, na=1)
! =================================================================================================================================
!
!     zap: Product ZA identifier 1000*Zp+Ap (real*8)
!     awp: Product mass in neutron units (real*8)
!     lip: Product modifier flag (integer) for isomeric state, multiple emission
!      ny: number of incident energy points for y(E) (integer)
!      ey: incident energy points for y(E)
!          real*8 array ey=[ey(i),i=1..ny]
!       y: y(E) yields
!          real*8 array y=[y(i),i=1..ny]
!     ily: interpolation law for (E,y(E)) by intervals
!          integer array ily=[ily(i),i=1,ny-1]
!      ne: number of incident energy points for the Kalbach-Mann formalism (integer)
!       e: incident energy points
!          real*8 array e=[e(i), i=1..ne]
!     ile: incident energy interpolation law by intervals
!          integer array ile=[ile(i),i=1..ne-1]
!      nd: Number of discret energies
!          integer array nd=[nd(i),i=1..ne], 0 <= nd(i) <= nep(i)
!  nepdim: Maximum number of secondary energies (integer)
!     nep: Number of secondary energies
!          integer array nep=[nep(i),i=1..ne]
!      ep: Secondary/outgoing energies
!          real*8 array ep=[ep(j,i), j=1..nep(i), i=1..ne], ne<=nemax for MF6
!     lep: interpolation law between secondary energies for j>nd(i) (integer)
!      f0: total emission probability f0(E',E)
!          real*8 array f0=[f0(j,i), j=1..nep(i), i=1..ne]
!       r: pre-compound fraction (ranges from 0.0 to 1.0)
!          real*8 array r=[r(j,i), j=1..nep(i), i=1..ne]
!
!   for j>nd(i), a continuous distribution is given with interpolation law=lep
!   between secondary energies, for j<=nd(i) discret distribution is provided.
!
  subroutine wrtmf6_law1_Kalbach_Mann_na1(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nd,nepdim,nep,ep,lep,f0,r)
    implicit real*8 (a-h,o-z)
    parameter (ncyc=3, nwmax=ncyc*nep6max)
    character*80 nou(*)
    dimension ey(*),y(*),ily(*),e(*),ile(*),nd(*),nep(*),ep(nepdim,*),f0(nepdim,*),r(nepdim,*)
    dimension nbt(nrmax),ibt(nrmax),b(nwmax)
    if (ny.gt.ne6max) then
      write(*,*)' MF6/law=1/lang=2 Warning: Too many incident energy points. ny=',ny,' ne6max=',ne6max
    endif
    if (ne.gt.ne6max) then
      write(*,*)' MF6/law=1/lang=2 Warning: Too many incident energy points. ne=',ne,' ne6max=',ne6max
    endif
    if (nepdim.gt.nep6max) then
      write(*,*)' MF6/law=1/lang=2 Error: Too many outgoing energies. nep=',nepdim,' nep6max=',nep6max
      stop
    endif
    if (ne*nepdim.gt.npmax) then
      write(*,*)' MF6/law=1/lang=2 Error: Arrays too large. npmax=',npmax
      write(*,*)' ne=',ne,' nep=',nepdim,' ne*nep=',ne*nepdim
      stop
    endif
    if (lep.lt.1.or.lep.gt.5) then
      write(*,*)' MF6/law=1 lang=2 Error: Incorrect secondary energy interpolation law (lep)=',lep,' (1-5 allowed)'
      stop
    endif
    call packibt(ny,ily,nr,nbt,ibt)
    mf=6
    law=1
    lang=2
    nai=1
    c1=0.0d0
    c2=0.0d0
    l1=0
    l2=0
    n1=0
    n2=0
    call wrtab1(nou,jou,mat,mf,mt,ns,zap,awp,lip,law,nr,nbt,ibt,ny,ey,y)
    call packibt(ne,ile,nr,nbt,ibt)
    call wrtab2(nou,jou,mat,mf,mt,ns,c1,c2,lang,lep,nr,nbt,ibt,ne)
    do i=1,ne
      ndi=nd(i)
      nepi=nep(i)
      nw=nepi*ncyc
      do j=1,nepi
        jb=ncyc*(j-1)+1
        b(jb)=ep(j,i)
        b(jb+1)=f0(j,i)
        b(jb+2)=r(j,i)
      enddo
      call wrtlist(nou,jou,mat,mf,mt,ns,c1,e(i),ndi,nai,nw,nepi,b)
    enddo
    return
  end subroutine wrtmf6_law1_Kalbach_Mann_na1
!
! =================================================================================================================================
! subroutine wrtmf6_law1_Kalbach_Mann_na1_law: Write data for MF6/LAW=1 Kalbach Mann formalism (law=1, lang=2, na=1)
! =================================================================================================================================
!
!   Similar to wrtmf6_law1_Kalbach_Mann_na1, but each interpolation scheme is represented by only one interpolation law.
!    lawy: interpolation law for [E, y(E)]
!    lawf: interpolation law for [E, (f0(E',E),r(E',E)]
!   The rest of parameters have the same meaning as before
!
  subroutine wrtmf6_law1_Kalbach_Mann_na1_law(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,lawy,ne,e,lawf,nd,nepdim,nep,ep,lep,f0,r)
    implicit real*8 (a-h,o-z)
    character*80 nou(*)
    dimension ey(*),y(*),e(*),nd(*),nep(*),ep(nepdim,*),f0(nepdim,*),r(nepdim,*)
    dimension ily(ny),ile(ne)
    do i=1,ny-1
      ily(i)=lawy
    enddo
    do i=1,ne-1
      ile(i)=lawf
    enddo
    call wrtmf6_law1_Kalbach_Mann_na1(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nd,nepdim,nep,ep,lep,f0,r)
    return
  end subroutine wrtmf6_law1_Kalbach_Mann_na1_law
!
! =================================================================================================================================
! subroutine wrtmf6_law1_Kalbach_Mann_na1_lin: Write data for MF6/LAW=1 Kalbach Mann formalism (law=1, lang=2, na=1)
! =================================================================================================================================
!
!   Similar to wrtmf6_law1_Kalbach_Mann_na1, but all interpolation schema are lin-lin
!   The rest of parameters have the same meaning as before
!
  subroutine wrtmf6_law1_Kalbach_Mann_na1_lin(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ne,e,nd,nepdim,nep,ep,lep,f0,r)
    implicit real*8 (a-h,o-z)
    character*80 nou(*)
    dimension ey(*),y(*),e(*),nd(*),nep(*),ep(nepdim,*),f0(nepdim,*),r(nepdim,*)
    dimension ily(ny),ile(ne)
    do i=1,ny-1
      ily(i)=2
    enddo
    do i=1,ne-1
      ile(i)=2
    enddo
    call wrtmf6_law1_Kalbach_Mann_na1(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nd,nepdim,nep,ep,lep,f0,r)
    return
  end subroutine wrtmf6_law1_Kalbach_Mann_na1_lin
!
! =================================================================================================================================
! subroutine wrtmf6_law1_Kalbach_Mann_na2: Write data for MF6/LAW=1 Kalbach Mann formalism (law=1, lang=2, na=2)
! =================================================================================================================================
!
!     zap: Product ZA identifier 1000*Zp+Ap (real*8)
!     awp: Product mass in neutron units (real*8)
!     lip: Product modifier flag (integer) for isomeric state, multiple emission
!      ny: number of incident energy points for y(E) (integer)
!      ey: incident energy points for y(E)
!          real*8 array ey=[ey(i),i=1..ny]
!       y: y(E) yields
!          real*8 array y=[y(i),i=1..ny]
!     ily: interpolation law for (E,y(E)) by intervals
!          integer array ily=[ily(i),i=1,ny-1]
!      ne: number of incident energy points for the Kalbach-Mann formalism (integer)
!       e: incident energy points
!          real*8 array e=[e(i), i=1..ne]
!     ile: incident energy interpolation law by intervals
!          integer array ile=[ile(i),i=1..ne-1]
!      nd: Number of discret energies
!          integer array nd=[nd(i),i=1..ne], 0 <= nd(i) <= nep(i)
!  nepdim: Maximum number of secondary energies (integer)
!     nep: Number of secondary energies
!          integer array nep=[nep(i),i=1..ne]
!      ep: Secondary/outgoing energies
!          real*8 array ep=[ep(j,i), j=1..nep(i), i=1..ne], ne<=nemax for MF6
!     lep: interpolation law between secondary energies for j>nd(i) (integer)
!      f0: total emission probability f0(E',E)
!          real*8 array f0=[f0(j,i), j=1..nep(i), i=1..ne]
!       r: pre-compound fraction (ranges from 0.0 to 1.0)
!          real*8 array r=[r(j,i), j=1..nep(i), i=1..ne]
!       a: Kalbach-Mann a-parameter
!          real*8 array a=[a(j,i), j=1..nep(i), i=1..ne]
!
!   for j>nd(i), a continuous distribution is given with interpolation law=lep
!   between secondary energies, for j<=nd(i) discret distribution is provided.
!
  subroutine wrtmf6_law1_Kalbach_Mann_na2(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nd,nepdim,nep,ep,lep,f0,r,a)
    implicit real*8 (a-h,o-z)
    parameter (ncyc=4, nwmax=ncyc*nep6max)
    character*80 nou(*)
    dimension ey(*),y(*),ily(*),e(*),ile(*),nd(*),nep(*),ep(nepdim,*),f0(nepdim,*),r(nepdim,*),a(nepdim,*)
    dimension nbt(nrmax),ibt(nrmax),b(nwmax)
    if (ny.gt.ne6max) then
      write(*,*)' MF6/law=1/lang=2 Warning: Too many incident energy points. ny=',ny,' ne6max=',ne6max
    endif
    if (ne.gt.ne6max) then
      write(*,*)' MF6/law=1/lang=2 Warning: Too many incident energy points. ne=',ne,' ne6max=',ne6max
    endif
    if (nepdim.gt.nep6max) then
      write(*,*)' MF6/law=1/lang=2 Error: Too many outgoing energies. nep=',nepdim,' nep6max=',nep6max
      stop
    endif
    if (ne*nepdim.gt.npmax) then
      write(*,*)' MF6/law=1/lang=2 Error: Arrays too large. npmax=',npmax
      write(*,*)' ne=',ne,' nep=',nepdim,' ne*nep=',ne*nepdim
      stop
    endif
    if (lep.lt.1.or.lep.gt.5) then
      write(*,*)' MF6/law=1 lang=2 Error: Incorrect secondary energy interpolation law (lep)=',lep,' (1-5 allowed)'
      stop
    endif
    call packibt(ny,ily,nr,nbt,ibt)
    mf=6
    law=1
    lang=2
    nai=2
    c1=0.0d0
    c2=0.0d0
    l1=0
    l2=0
    n1=0
    n2=0
    call wrtab1(nou,jou,mat,mf,mt,ns,zap,awp,lip,law,nr,nbt,ibt,ny,ey,y)
    call packibt(ne,ile,nr,nbt,ibt)
    call wrtab2(nou,jou,mat,mf,mt,ns,c1,c2,lang,lep,nr,nbt,ibt,ne)
    do i=1,ne
      ndi=nd(i)
      nepi=nep(i)
      nw=nepi*ncyc
      do j=1,nepi
        jb=ncyc*(j-1)+1
        b(jb)=ep(j,i)
        b(jb+1)=f0(j,i)
        b(jb+2)=r(j,i)
        b(jb+3)=a(j,i)
      enddo
      call wrtlist(nou,jou,mat,mf,mt,ns,c1,e(i),ndi,nai,nw,nepi,b)
    enddo
    return
  end subroutine wrtmf6_law1_Kalbach_Mann_na2
!
! =================================================================================================================================
! subroutine wrtmf6_law1_Kalbach_Mann_na2_law: Write data for MF6/LAW=1 Kalbach Mann formalism (law=1, lang=2, na=2)
! =================================================================================================================================
!
!   Similar to wrtmf6_law1_Kalbach_Mann_na2, but each interpolation scheme is represented by only one interpolation law.
!    lawy: interpolation law for [E, y(E)]
!    lawf: interpolation law for [E, (f0(E',E),r(E',E),a(E',E))]
!   The rest of parameters have the same meaning as before
!
  subroutine wrtmf6_law1_Kalbach_Mann_na2_law(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,lawy,ne,e,lawf,nd,nepdim,nep,ep,lep,f0,r,a)
    implicit real*8 (a-h,o-z)
    character*80 nou(*)
    dimension ey(*),y(*),e(*),nd(*),nep(*),ep(nepdim,*),f0(nepdim,*),r(nepdim,*),a(nepdim,*)
    dimension ily(ny),ile(ne)
    do i=1,ny-1
      ily(i)=lawy
    enddo
    do i=1,ne-1
      ile(i)=lawf
    enddo
    call wrtmf6_law1_Kalbach_Mann_na2(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nd,nepdim,nep,ep,lep,f0,r,a)
    return
  end subroutine wrtmf6_law1_Kalbach_Mann_na2_law
!
! =================================================================================================================================
! subroutine wrtmf6_law1_Kalbach_Mann_na2_lin: Write data for MF6/LAW=1 Kalbach Mann formalism (law=1, lang=2, na=2)
! =================================================================================================================================
!
!   Similar to wrtmf6_law1_Kalbach_Mann_na2, but all interpolation schema are lin-lin
!   The rest of parameters have the same meaning as before
!
  subroutine wrtmf6_law1_Kalbach_Mann_na2_lin(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ne,e,nd,nepdim,nep,ep,lep,f0,r,a)
    implicit real*8 (a-h,o-z)
    character*80 nou(*)
    dimension ey(*),y(*),e(*),nd(*),nep(*),ep(nepdim,*),f0(nepdim,*),r(nepdim,*),a(nepdim,*)
    dimension ily(ny),ile(ne)
    do i=1,ny-1
      ily(i)=2
    enddo
    do i=1,ne-1
      ile(i)=2
    enddo
    call wrtmf6_law1_Kalbach_Mann_na2(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nd,nepdim,nep,ep,lep,f0,r,a)
    return
  end subroutine wrtmf6_law1_Kalbach_Mann_na2_lin
!
! =================================================================================================================================
! subroutine wrtmf6_law1_tab: Write tabulated distribution (law=1, lang=11-15 )
! =================================================================================================================================
!
!      zap: Product ZA identifier 1000*Zp+Ap (real*8)
!      awp: Product mass in neutron units (real*8)
!      lip: Product modifier flag (integer) for isomeric state, multiple emission
!       ny: number of incident energy points for y(E) (integer)
!       ey: incident energy points for y(E)
!           real*8 array ey=[ey(i),i=1..ny]
!        y: y(E) yields
!           real*8 array y=[y(i),i=1..ny]
!      ily: interpolation law for (E,y(E)) by intervals
!           integer array ily=[ily(i),i=1,ny-1]
!       ne: number of incident energy points for tabulated data (integer)
!        e: incident energy points tabulated data
!           real*8 array e=[e(i), i=1..ne]
!      ile: incident energy interpolation law by intervals
!           integer array ile=[ile(i),i=1..ne-1]
!       nd: Number of discret energies
!           integer array nd=[nd(i),i=1..ne], 0 <= nd(i) <= nep(i)
!    nadim: Maximum number of cosines (integer)
!       na: Number of tabulated cosines (integer)
!           integer array na=[na(i),i=1..ne], na(i) <= nadim
!   nepdim: Maximum number of outgoing energies (integer)
!      nep: Number of secondary energies (integer)
!           integer array nep=[nep(i),i=1..ne], nep(i) <= nepdim
!       ep: Secondary/outgoing energies
!           real*8 array ep=[ep(j,i),j=1..nep(i),i=1..ne]
!      lep: interpolation law between secondary energies for j>nd(i) (integer)
!       f0: total emission probability f0(E',E)
!           real*8 array f0=[f0(j,i),j=1..nep(i),i=1..ne]
!      xmu: cosines values (ranges from -1.0 to 1.0)
!           real*8 array xmu=[xmu(j,k,i), j=1..nep(i), k=1..na(i), i=1..ne]
!      ymu: angular distribution
!           real*8 array ymu=[ymu(j,k,i), j=1..nep(i), k=1..na(i), i=1..ne]
!      lmu: interpolation law between cosines (integer)
!
!   for j>nd(i), a continuous distribution is given with interpolation law=lep
!   between secondary energies, for j<=nd(i) discret distribution is provided.
!
  subroutine wrtmf6_law1_tab(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nd,nadim,na,nepdim,nep,ep,lep,f0,xmu,ymu,lmu)
    implicit real*8 (a-h,o-z)
    parameter (nbmax=(2*nmu6max+2)*nep6max)
    character*80 nou(*)
    dimension ey(*),y(*),ily(*),e(*),ile(*),nd(*),na(*),nep(*)
    dimension ep(nepdim,*),f0(nepdim,*),xmu(nepdim,nadim,*),ymu(nepdim,nadim,*)
    dimension nbt(nrmax),ibt(nrmax),b(nbmax)
    if (ny.gt.ne6max) then
      write(*,*)' MF6/law=1/lang=11-15 Warning: Too many incident energy points. ny=',ny,' ne6max=',ne6max
    endif
    if (ne.gt.ne6max) then
      write(*,*)' MF6/law=1/lang=11-15 Warning: Too many incident energy points. ne=',ne,' ne6max=',ne6max
    endif
    if (nadim.gt.nmu6max) then
      write(*,*)' MF6/law=1/lang=11-15 Warning: Too many cosines. nmu=',nadim,' nmu6max=',nmu6max
    endif
    if (nepdim.gt.nep6max) then
      write(*,*)' MF6/law=1/lang=11-15 Warning: Too many cosines. nep=',nepdim,' nep6max=',nep6max
    endif
    if (lmu.lt.1.or.lmu.gt.5) then
      write(*,*)' MF6/law=1/lang=11-15 Error: Incorrect angular distribution interpolation law=',lmu,' (1-5 allowed)'
      stop
    endif
    if (lep.lt.1.or.lep.gt.5) then
      write(*,*)' MF6/law=1/lang=11-15 Error: Incorrect secondary energy interpolation law=',lep,' (1-5 allowed)'
      stop
    endif
    if ((2*nadim+2)*nepdim.gt.nbmax) then
      write(*,*)' MF6/law=1/lang=11-15 Error: Too large LIST record nbmax=',nbmax
      write(*,*)' na=',nadim,' nep=',nepdim,' nb=',(2*nadim+2)*nepdim
      stop
    endif
    if (ne*nadim*nepdim.gt.npmax) then
      write(*,*)' MF6/law=1/lang=11-15 Error: Too large arrays for zap=',nint(zap),' npmax=',npmax
      write(*,*)' ne=',ne,' na=',nadim,' nep=',nepdim,' ne*na*nep=',ne*nadim*nepdim
      stop
    endif
    call packibt(ny,ily,nr,nbt,ibt)
    mf=6
    law=1
    lang=lmu+10
    c1=0.0d0
    c2=0.0d0
    l1=0
    l2=0
    n1=0
    n2=0
    call wrtab1(nou,jou,mat,mf,mt,ns,zap,awp,lip,law,nr,nbt,ibt,ny,ey,y)
    call packibt(ne,ile,nr,nbt,ibt)
    call wrtab2(nou,jou,mat,mf,mt,ns,c1,c2,lang,lep,nr,nbt,ibt,ne)
    do i=1,ne
      ndi=nd(i)
      nai2=na(i)
      nai=2*nai2
      nepi=nep(i)
      ncyc=nai+2
      nw=nepi*ncyc
      do j=1,nepi
        jb=ncyc*(j-1)+1
        b(jb)=ep(j,i)
        a0=f0(j,i)
        jb=jb+1
        b(jb)=a0
        if (nai2.gt.0) then
          do k=1,nai2
            jb=jb+1
            b(jb)=xmu(j,k,i)
            jb=jb+1
            if (a0.ne.0.0d0) then
              b(jb)=0.5d0*ymu(j,k,i)/a0
            else
              b(jb)=0.25d0
            endif
          enddo
        endif
      enddo
      call wrtlist(nou,jou,mat,mf,mt,ns,c1,e(i),ndi,nai,nw,nepi,b)
    enddo
    return
  end subroutine wrtmf6_law1_tab
!
! =================================================================================================================================
! subroutine wrtmf6_law1_tab_law: Write data for MF6/LAW=1 tabulated distribution (law=1, lang=11-15)
! =================================================================================================================================
!
!   Similar to wrtmf6_law1_tab, but each interpolation scheme is represented by only one interpolation law.
!    lawy: interpolation law for [E, y(E)]
!    lawf: interpolation law between incident energies
!   The rest of parameters have the same meaning as before
!
  subroutine wrtmf6_law1_tab_law(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,lawy,ne,e,lawf,nd,nadim,na,nepdim,nep,ep,lep,f0,xmu,ymu,lmu)
    implicit real*8 (a-h,o-z)
    character*80 nou(*)
    dimension ey(*),y(*),e(*),nd(*),na(*),nep(*)
    dimension ep(nepdim,*),f0(nepdim,*),xmu(nepdim,nadim,*),ymu(nepdim,nadim,*)
    dimension ily(ny),ile(ne)
    do i=1,ny-1
      ily(i)=lawy
    enddo
    do i=1,ne-1
      ile(i)=lawf
    enddo
    call wrtmf6_law1_tab(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nd,nadim,na,nepdim,nep,ep,lep,f0,xmu,ymu,lmu)
    return
  end subroutine wrtmf6_law1_tab_law
!
! =================================================================================================================================
! subroutine wrtmf6_law1_tab_lin: Write data for MF6/LAW=1 tabulated distribution (law=1, lang=11-15)
! =================================================================================================================================
!
!   Similar to wrtmf6_law1_tab, but all interpolation schema are lin-lin, therefore they are not input parameters.
!   lep and lmu are also assumed equal 2 (lin-lin)
!   The rest of parameters have the same meaning as before
!
  subroutine wrtmf6_law1_tab_lin(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ne,e,nd,nadim,na,nepdim,nep,ep,f0,xmu,ymu)
    implicit real*8 (a-h,o-z)
    character*80 nou(*)
    dimension ey(*),y(*),e(*),nd(*),na(*),nep(*)
    dimension ep(nepdim,*),f0(nepdim,*),xmu(nepdim,nadim,*),ymu(nepdim,nadim,*)
    dimension ily(ny),ile(ne)
    do i=1,ny-1
      ily(i)=2
    enddo
    do i=1,ne-1
      ile(i)=2
    enddo
    lep=2
    lmu=2
    call wrtmf6_law1_tab(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nd,nadim,na,nepdim,nep,ep,lep,f0,xmu,ymu,lmu)
    return
  end subroutine wrtmf6_law1_tab_lin
!
! =================================================================================================================================
! subroutine wrtmf6_law2_leg: Write data for MF6/LAW=2 Legendre expansion (law=2, lang=0)
! =================================================================================================================================
!
!     zap: Product ZA identifier 1000*Zp+Ap (real*8)
!     awp: Product mass in neutron units (real*8)
!     lip: Product modifier flag (integer) for isomeric state, multiple emission
!      ny: number of incident energy points for y(E) (integer)
!      ey: incident energy points for y(E)
!          real*8 array ey=[ey(i),i=1..ny]
!       y: y(E) yields
!          real*8 array y=[y(i),i=1..ny]
!     ily: interpolation law for (E,y(E)) by intervals
!          integer array ily=[ily(i),i=1,ny-1]
!      ne: number of incident energy points for the Legendre expansion (integer)
!       e: incident energy points for the Legendre expansion
!          real*8 array e=[e(i), i=1..ne]
!     ile: incident energy interpolation law by intervals
!          integer array ile=[ile(i),i=1..ne-1]
!   nadim: Maximum number of Legendre coefficients (integer)
!      na: Number of Legendre coefficients
!          integer array na=[na(i),i=1..ne],  1 <= na(i) <= nadim
!       a: Legendre coefficients
!          real*8 array a=[a(k,i),k=1..na(i),i=1..ne]
!
  subroutine wrtmf6_law2_leg(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nadim,na,a)
    implicit real*8 (a-h,o-z)
    character*80 nou(*)
    dimension ey(*),y(*),ily(*),e(*),ile(*),na(*),a(nadim,*)
    dimension nbt(nrmax),ibt(nrmax),b(nadim)
    if (ny.gt.ne6max) then
      write(*,*)' MF6/law=2/lang=0 Warning: Too many incident energy points. ny=',ny,' ne6max=',ne6max
    endif
    if (ne.gt.ne6max) then
      write(*,*)' MF6/law=2/lang=0 Warning: Too many incident energy points. ne=',ne,' ne6max=',ne6max
    endif
    if (nadim.gt.ncmax) then
      write(*,*)' MF6/law=2/lang=0 Error: Too many Legendre coefficients. nc=',nadim,' ncmax=',ncmax
      stop
    endif
    if (ne*nadim.gt.npmax)then
      write(*,*)' MF6/law=2/lang=0 Error: Too large arrays. npmax=',npmax
      write(*,*)' ne=',ne,' na=',nadim
      stop
    endif
    call packibt(ny,ily,nr,nbt,ibt)
    mf=6
    law=2
    lang=0
    c=1.0d0
    c1=0.0d0
    c2=0.0d0
    l1=0
    l2=0
    n1=0
    n2=0
    call wrtab1(nou,jou,mat,mf,mt,ns,zap,awp,lip,law,nr,nbt,ibt,ny,ey,y)
    call packibt(ne,ile,nr,nbt,ibt)
    call wrtab2(nou,jou,mat,mf,mt,ns,c1,c2,l1,l2,nr,nbt,ibt,ne)
    do i=1,ne
      nai=na(i)
      nli=nai-1
      do k=1,nai
        b(k)=a(k,i)
      enddo
      call renorm_leg(nai,b,c,f)
      do k=1,nli
        b(k)=b(k+1)
      enddo
      call wrtlist(nou,jou,mat,mf,mt,ns,c1,e(i),lang,n1,nli,nli,b)
    enddo
    return
  end subroutine wrtmf6_law2_leg
!
! =================================================================================================================================
! subroutine wrtmf6_law2_leg_law: Write data for MF6/LAW=2 Legendre expansion (law=2, lang=0)
! =================================================================================================================================
!
!   Similar to wrtmf6_law2_leg, but each interpolation scheme is represented by only one interpolation law.
!     lawy: interpolation law for [E, y(E)]
!     lawa: interpolation law for [E,a(E',E)]
!   The rest of parameters have the same meaning as before
!
  subroutine wrtmf6_law2_leg_law(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,lawy,ne,e,lawa,nadim,na,a)
    implicit real*8 (a-h,o-z)
    character*80 nou(*)
    dimension ey(*),y(*),e(*),na(*),a(nadim,*)
    dimension ily(ny),ile(ne)
    do i=1,ny-1
      ily(i)=lawy
    enddo
    do i=1,ne-1
      ile(i)=lawa
    enddo
    call wrtmf6_law2_leg(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nadim,na,a)
    return
  end subroutine wrtmf6_law2_leg_law
!
! =================================================================================================================================
! subroutine wrtmf6_law2_leg_lin: Write data for MF6/LAW=2 Legendre expansion (law=2, lang=0)
! =================================================================================================================================
!
!   Similar to wrtmf6_law2_leg, but all interpolation schema are lin-lin, therefore they are not input parameters.
!   The rest of parameters have the same meaning as before
!
  subroutine wrtmf6_law2_leg_lin(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ne,e,nadim,na,a)
    implicit real*8 (a-h,o-z)
    character*80 nou(*)
    dimension ey(*),y(*),e(*),na(*),a(nadim,*)
    dimension ily(ny),ile(ne)
    do i=1,ny-1
      ily(i)=2
    enddo
    do i=1,ne-1
      ile(i)=2
    enddo
    call wrtmf6_law2_leg(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nadim,na,a)
    return
  end subroutine wrtmf6_law2_leg_lin
!
! =================================================================================================================================
! subroutine wrtmf6_law2_tab: Write data for MF6/LAW=2 tabular representation (law=2, lang=[12 or 14] )
! =================================================================================================================================
!
!     zap: Product ZA identifier 1000*Zp+Ap (real*8)
!     awp: Product mass in neutron units (real*8)
!     lip: Product modifier flag (integer) for isomeric state, multiple emission
!      ny: number of incident energy points for y(E) (integer)
!      ey: incident energy points for y(E)
!          real*8 array ey=[ey(i),i=1..ny]
!       y: y(E) yields
!          real*8 array y=[y(i),i=1..ny]
!     ily: interpolation law for (E,y(E)) by intervals
!          integer array ily=[ily(i),i=1,ny-1]
!      ne: number of incident energy points for tabulated data (integer)
!       e: incident energy points for tabulated data
!          real*8 array e=[e(i), i=1..ne]
!     ile: incident energy interpolation law by intervals
!          integer array ile=[ile(i),i=1..ne-1]
!   nadim: Maximum number of cosines. Maximum number of angular dimension (integer)
!     nmu: number of cosine by incident energy
!          integer array nmu=[nmu(i),1..ne]
!     xmu: set of cosine values by incident energy
!          real*8 xmu=[xmu(k,i),k=1..nmu(i),i=1..ne]
!     ymu: angle-distribution values by incident energy
!          real*8 ymu=[ymu(k,i),k=1,nmu(i),i=1..ne]
!    ilmu: cosine interpolation law by incident energy
!          integer ilmu=[ilmu(i),i=1..ne]
!
  subroutine wrtmf6_law2_tab(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nadim,nmu,xmu,ymu,ilmu)
    implicit real*8 (a-h,o-z)
    parameter (nbmax=2*nmu6max)
    character*80 nou(*)
    dimension ey(*),y(*),ily(*),e(*),ile(*),nmu(*),xmu(nadim,*),ymu(nadim,*),ilmu(*)
    dimension nbt(nrmax),ibt(nrmax),x1(nadim),y1(nadim),b(nbmax),jbt(nadim)
    if (ny.gt.ne6max) then
      write(*,*)' MF6/law=2 (tabulated) Warning: Too many incident energy points. ny=',ny,' ne6max=',ne6max
    endif
    if (ne.gt.ne6max) then
      write(*,*)' MF6/law=2 (tabulated) Warning: Too many incident energy points. ne=',ne,' ne6max=',ne6max
    endif
    if (nadim.gt.nmu6max) then
      write(*,*)' MF6/law=2 (tabulated) Error: Too many cosines. nmu=',nadim,' nmu6max=',nmu6max
      stop
    endif
    if (ne*nadim.gt.npmax) then
      write(*,*)' MF6/law=2 (tabulated) Error: Too large arrays. npmax=',npmax
      write(*,*)' ne=',ne,' na=',nadim,' ne*na=',ne*nadim
      stop
    endif
    call packibt(ny,ily,nr,nbt,ibt)
    mf=6
    law=2
    c=1.0d0
    c1=0.0d0
    c2=0.0d0
    l1=0
    l2=0
    n1=0
    n2=0
    call wrtab1(nou,jou,mat,mf,mt,ns,zap,awp,lip,law,nr,nbt,ibt,ny,ey,y)
    call packibt(ne,ile,nr,nbt,ibt)
    call wrtab2(nou,jou,mat,mf,mt,ns,c1,c2,l1,l2,nr,nbt,ibt,ne)
    do i=1,ne
      ilaw=ilmu(i)
      lang=ilaw+10
      nmui=nmu(i)
      nw=2*nmui
      if (lang.ne.12.or.lang.ne.14) then
        write(*,*)' MF6/law=2 (tabulated/lang=12 or 14) Error:  law=',law,' lang=',lang
        stop
      endif
      do k=1,nmui
        x1(k)=xmu(k,i)
        y1(k)=ymu(k,i)
        jbt(k)=ilaw
      enddo
      call renorm_tab(nmui,x1,y1,jbt,c,f)
      jb=0
      do k=1,nmui
        jb=jb+1
        b(jb)=x1(k)
        jb=jb+1
        b(jb)=y1(k)
      enddo
      call wrtlist(nou,jou,mat,mf,mt,ns,c1,e(i),lang,n1,nw,nmui,b)
    enddo
    return
  end subroutine wrtmf6_law2_tab
!
! =================================================================================================================================
! subroutine wrtmf6_law2_tab_law: Write data for MF6/LAW=2 tabular representation (law=2, lang=[12 or 14])
! =================================================================================================================================
!
!   Similar to wrtmf6_law2_tab, but each interpolation scheme is represented by only one interpolation law.
!     lawy: interpolation law for [E, y(E)]
!     lawf: interpolation law between incident energies
!    lawmu: interpolation law between cosines
!   The rest of parameters have the same meaning as before
!
  subroutine wrtmf6_law2_tab_law(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,lawy,ne,e,lawf,nadim,nmu,xmu,ymu,lawmu)
    implicit real*8 (a-h,o-z)
    character*80 nou(*)
    dimension ey(*),y(*),e(*),nmu(*),xmu(nadim,*),ymu(nadim,*)
    dimension ily(ny),ile(ne),ilmu(ne)
    do i=1,ny-1
      ily(i)=lawy
    enddo
    do i=1,ne-1
      ile(i)=lawf
      ilmu(i)=lawmu
    enddo
    call wrtmf6_law2_tab(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nadim,nmu,xmu,ymu,ilmu)
    return
  end subroutine wrtmf6_law2_tab_law
!
! =================================================================================================================================
! subroutine wrtmf6_law2_tab_lin: Write data for MF6/LAW=2 tabular representation (law=2, lang=[12 or 14])
! =================================================================================================================================
!
!   Similar to wrtmf6_law2_tab, but all interpolation schema are lin-lin, therefore they are not input parameters.
!   The rest of parameters have the same meaning as before
!
  subroutine wrtmf6_law2_tab_lin(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ne,e,nadim,nmu,xmu,ymu)
    implicit real*8 (a-h,o-z)
    character*80 nou(*)
    dimension ey(*),y(*),e(*),nmu(*),xmu(nadim,*),ymu(nadim,*)
    dimension ily(ny),ile(ne),ilmu(ne)
    do i=1,ny-1
      ily(i)=2
    enddo
    do i=1,ne-1
      ile(i)=2
      ilmu(i)=2
    enddo
    call wrtmf6_law2_tab(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nadim,nmu,xmu,ymu,ilmu)
    return
  end subroutine wrtmf6_law2_tab_lin
!
! =================================================================================================================================
! subroutine wrtmf6_law2_mix: Write one MF6 subsection for law=2/ltype=3 (mix)
! =================================================================================================================================
!
!    1. Legendre expansion at lower energies (lang=0)
!    2. Tabulated distributions at higher energies (lang=12 or 14)
!
!    zap: Product ZA identifier 1000*Zp+Ap (real*8)
!    awp: Product mass in neutron units (real*8)
!    lip: Product modifier flag (integer) for isomeric state, multiple emission
!     ny: number of incident energy points for y(E) (integer)
!     ey: incident energy points for y(E)
!         real*8 array ey=[ey(i),i=1..ny]
!      y: y(E) yields
!         real*8 array y=[y(i),i=1..ny]
!    ily: interpolation law for (E,y(E)) by intervals
!         integer array ily=[ily(i),i=1,ny-1]
!    nel: number of incident energy points for the Legendre expansion (integer)
!     el: incident energy points for the Legendre expansion at lower energies
!         real*8 array el=[el(i), i=1..nel]
!   ilel: incident energy interpolation law by intervals at lower energies
!         integer array ilel=[ilel(i),i=1..nel-1]
!    nal: Maximum number of Legendre coefficients
!     na: Number of Legendre coefficients (integer)
!         integer array na=[na(i),i=1..nel],  1 <= na(i) <= nal
!      a: Legendre coefficients
!         real*8 array a=[a(k,i),k=1..na(i),i=1..nel]
!    neh: number of incident energy points for tabulated data (integer)
!     eh: incident energy points for tabulated data at higher energies
!         real*8 array eh=[eh(i), i=1..neh]
!   ileh: incident energy interpolation law by intervals at higher energies
!         integer array ileh=[ileh(i),i=1..neh-1]
!    nah: Maximum number of cosines (integer)
!    nmu: number of cosine by incident energy
!         integer array nmu=[nmu(i),1..neh], 1 < nmu(i) <= nah
!    xmu: set of cosine values by incident energy
!         real*8 xmu=[xmu(k,i),k=1..nmu(i),i=1..neh]
!    ymu: angle-distribution values by incident energy
!         real*8 ymu=[ymu(k,i),k=1,nmu(i),i=1..neh]
!   ilmu: cosine interpolation law by incident energy
!         integer ilmu=[ilmu(i),i=1..neh]
!
  subroutine wrtmf6_law2_mix(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ily,nel,el,ilel,nal,na,a,neh,eh,ileh,nah,nmu,xmu,ymu,ilmu)
    implicit real*8 (a-h,o-z)
    parameter (eps=1.0d-6, nbmax=max(ncmax,2*nmu6max))
    character*80 nou(*)
    dimension ey(*),y(*),ily(*)
    dimension el(*),ilel(*),na(*),a(nal,*)
    dimension eh(*),ileh(*),nmu(*),xmu(nah,*),ymu(nah,*),ilmu(*)
    dimension nbt(nrmax),ibt(nrmax),nbth(nrmax),ibth(nrmax),b(nbmax),x1(nah),y1(nah),jbt(nah)
    ne=nel+neh
    if (ny.gt.ne6max) then
      write(*,*)' MF6/law=2 (mixed) Warning: Too many incident energy points. ny=',ny,' ne6max=',ne6max
    endif
    if (ne.gt.ne6max) then
      write(*,*)' MF6/law=2 (mixed) Warning: Too many incident energy points. ne=',ne,' ne6max=',ne6max
    endif
    if (nal.gt.ncmax) then
      write(*,*)' MF6/law=2 (mixed) Warning: Too many Legendre coefficients. nc=',nal,' ncmax=',ncmax
    endif
    if (nah.gt.nmu6max) then
      write(*,*)' MF6/law=2 (mixed) Warning: Too many cosines. nmu=',nah,' nmu6max=',nmu6max
    endif
    if (nel.le.0.or.neh.le.0) then
      write(*,*)' MF6/law=2 (mixed) Error: Not mixed representation. nel=',nel,' neh=',neh
      stop
    endif
    if (max(nal,2*nah).gt.nbmax) then
      write(*,*)' MF6/law=2 (mixed) Error: Too large arrays for zap=',nint(zap),' nbmax=',nbmax
      write(*,*)' nal=',nal,' 2*nah=',2*nah
      stop
    endif
    if (nal*nel.gt.npmax.or.nah*neh.gt.npmax) then
      write(*,*)' MF6/law=2 (mixed) Error: Too large arrays for zap=',nint(zap),' npmax=',npmax
      write(*,*)' nal*nel=',nal*nel,' nah*neh=',nah*neh
      stop
    endif
    if (abs(eh(1)-el(nel)).gt.(eps*el(nel))) then
      write(*,*)' MF6/law=2 (mixed) Error: Energy discontinuity between Legendre and tabulated data el/eh=',el(nel),eh(1)
      stop
    endif
    call packibt(ny,ily,nr,nbt,ibt)
    mf=6
    law=2
    c=1.0d0
    c1=0.0d0
    c2=0.0d0
    l1=0
    l2=0
    n1=0
    n2=0
    call wrtab1(nou,jou,mat,mf,mt,ns,zap,awp,lip,law,nr,nbt,ibt,ny,ey,y)
    call packibt(nel,ilel,nrl,nbt,ibt)
    call packibt(neh,ileh,nrh,nbth,ibth)
    nr=nrl+nrh
    if (nr.gt.nrmax) then
      write(*,*)' MF6/law=2 (mixed) Error: Too many incident energy interpolation ranges. nr=',nr,' nrmax=',nrmax
      stop
    else
      do i=1,nrh
        k=nrl+i
        nbt(k)=nbth(i)+nel
        ibt(k)=ibth(i)
      enddo
    endif
    call wrtab2(nou,jou,mat,mf,mt,ns,c1,c2,l1,l2,nr,nbt,ibt,ne)
    lang=0
    do i=1,nel
      nai=na(i)
      nli=nai-1
      do k=1,nai
        b(k)=a(k,i)
      enddo
      call renorm_leg(nai,b,c,f)
      do k=1,nli
        b(k)=b(k+1)
      enddo
      call wrtlist(nou,jou,mat,mf,mt,ns,c1,el(i),lang,n1,nli,nli,b)
    enddo
    do i=1,neh
      ilaw=ilmu(i)
      lang=ilaw+10
      nmui=nmu(i)
      nw=2*nmui
      if (lang.ne.12.or.lang.ne.14) then
        write(*,*)' MF6/law=2 (mixed) Error: law=2/lang=',lang,' not allowed at higher energies'
        stop
      endif
      do k=1,nmui
        x1(k)=xmu(k,i)
        y1(k)=ymu(k,i)
        jbt(k)=ilaw
      enddo
      call renorm_tab(nmui,x1,y1,jbt,c,f)
      jb=0
      do k=1,nmui
        jb=jb+1
        b(jb)=x1(k)
        jb=jb+1
        b(jb)=y1(k)
      enddo
      call wrtlist(nou,jou,mat,mf,mt,ns,c1,eh(i),lang,n1,nw,nmui,b)
    enddo
    return
  end subroutine wrtmf6_law2_mix
!
! =================================================================================================================================
! subroutine wrtmf6_law2_mix_law: Write one MF6 subsection for law=2/ltype=3 (mix)
! =================================================================================================================================
!
!   1. Legendre expansion at lower energies (lang=0)
!   2. Tabulated distributions at higher energies (lang=12 or 14)
!
!   Similar to wrtmf6_law2_mix, but each interpolation scheme is represented by only one interpolation law.
!     lawy: interpolation law for [E, y(E)]
!    lawel: interpolation law between incident low energies
!    laweh: interpolation law between incident high energies
!      lmu: interpolation law between cosines
!   The rest of parameters have the same meaning as before
!
  subroutine wrtmf6_law2_mix_law(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,lawy,nel,el,lawel,nal,na,a,neh,eh,laweh,nah,nmu,xmu,ymu,lmu)
    implicit real*8 (a-h,o-z)
    character*80 nou(*)
    dimension ey(*),y(*)
    dimension el(*),na(*),a(nal,*)
    dimension eh(*),nmu(*),xmu(nah,*),ymu(nah,*)
    dimension ily(ny),ilel(nel),ileh(neh),ilmu(neh)
    do i=1,ny-1
      ily(i)=lawy
    enddo
    do i=1,nel-1
      ilel(i)=lawel
    enddo
    do i=1,neh-1
      ileh(i)=laweh
      ilmu(i)=lmu
    enddo
    call wrtmf6_law2_mix(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ily,nel,el,ilel,nal,na,a,neh,eh,ileh,nah,nmu,xmu,ymu,ilmu)
    return
  end subroutine wrtmf6_law2_mix_law
!
! =================================================================================================================================
! subroutine wrtmf6_law2_mix_lin: Write one MF6 subsection for law=2/ltype=3 (mix)
! =================================================================================================================================
!   1. Legendre expansion at lower energies (lang=0)
!   2. Tabulated distributions at higher energies (lang=12 or 14)
!
!   Similar to wrtmf6_law2_mix, but all interpolation schema are lin-lin, therefore they are not input parameters.
!   The rest of parameters have the same meaning as before
!
  subroutine wrtmf6_law2_mix_lin(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,nel,el,nal,na,a,neh,eh,nah,nmu,xmu,ymu)
    implicit real*8 (a-h,o-z)
    character*80 nou(*)
    dimension ey(*),y(*)
    dimension el(*),na(*),a(nal,*)
    dimension eh(*),nmu(*),xmu(nah,*),ymu(nah,*)
    dimension ily(ny),ilel(nel),ileh(neh),ilmu(neh)
    do i=1,ny-1
      ily(i)=2
    enddo
    do i=1,nel-1
      ilel(i)=2
    enddo
    do i=1,neh-1
      ileh(i)=2
      ilmu(i)=2
    enddo
    call wrtmf6_law2_mix(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ily,nel,el,ilel,nal,na,a,neh,eh,ileh,nah,nmu,xmu,ymu,ilmu)
    return
  end subroutine wrtmf6_law2_mix_lin
!
! =================================================================================================================================
! subroutine wrtmf6_law6_phase_space: Write one MF6 subsection for law=6 (n-body phase-space)
! =================================================================================================================================
!
!    zap: Product ZA identifier 1000*Zp+Ap (real*8)
!    awp: Product mass in neutron units (real*8)
!    lip: Product modifier flag (integer) for isomeric state, multiple emission
!     ny: number of incident energy points for y(E) (integer)
!     ey: incident energy points for y(E)
!         real*8 array ey=[ey(i),i=1..ny]
!      y: y(E) yields
!         real*8 array y=[y(i),i=1..ny]
!    ily: interpolation law for (E,y(E)) by intervals
!         integer array ily=[ily(i),i=1,ny-1]
!   npsx: number of particles n distributed according to the phase-space law (integer)
!   apsx: total mass M in neutron unit of the n particles (real*8)
!
  subroutine wrtmf6_law6_phase_space(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ily,npsx,apsx)
    implicit real*8 (a-h,o-z)
    character*80 nou(*)
    dimension ey(*),y(*),ily(*)
    dimension nbt(nrmax),ibt(nrmax)
    if (ny.gt.ne6max) then
      write(*,*)' MF6/law=6 Error: Too many incident energy points. ny=',ny,' ne6max=',ne6max
      stop
    endif
    call packibt(ny,ily,nr,nbt,ibt)
    mf=6
    law=6
    c2=0.0d0
    l1=0
    l2=0
    n1=0
    call wrtab1(nou,jou,mat,mf,mt,ns,zap,awp,lip,law,nr,nbt,ibt,ny,ey,y)
    call wrtcont(nou,jou,mat,mf,mt,ns,apsx,c2,l1,l2,n1,npsx)
    return
  end subroutine wrtmf6_law6_phase_space
!
! =================================================================================================================================
! subroutine wrtmf6_law7_lab: Write one MF6 subsection for law=7 (Laboratory Angle-Energy Distribution)
! =================================================================================================================================
!
!      zap: Product ZA identifier 1000*Zp+Ap (real*8)
!      awp: Product mass in neutron units (real*8)
!      lip: Product modifier flag (integer) for isomeric state, multiple emission
!       ny: number of incident energy points for y(E) (integer)
!       ey: incident energy points for y(E)
!           real*8 array ey=[ey(i),i=1..ny]
!        y: y(E) yields
!           real*8 array y=[y(i),i=1..ny]
!      ily: interpolation law for (E,y(E)) by intervals
!           integer array ily=[ily(i),i=1,ny-1]
!       ne: number of incident energy points for law7 (integer)
!        e: incident energy points
!           real*8 array e=[e(i), i=1..ne]
!      ile: incident energy interpolation law by intervals
!           integer array ile=[ile(i),i=1..ne-1]
!    nadim: Maximum number of cosines (integer)
!      nmu: Number of tabulated cosines
!           integer array nmu=[nmu(i),i=1..ne],  1 < nmu(i) <= nadim
!      xmu: cosines values by incident energy (ranges from -1.0 to 1.0)
!           real*8 array xmu=[xmu(k,i),j=1..nmu(i), i=1..ne]
!     ilmu: cosine interpolation law by interval
!           real*8 array ilmu=[ilmu(j,i),j=1..nmu(i)-1, i=1..ne]
!   nepdim: maximum number of outgoing energies (integer)
!      nep: Number of secondary energies by cosine and incident energy
!           integer array nep=[nep(j,i),j=1..nmu(i), i=1..ne]
!       ep: Secondary/outgoing energies
!           real*8 array ep=[ep(k,j,i),k=1..nep(j,i),j=1..nmu(i),i=1..ne]
!        f: Angle-energy distribution
!           real*8 array f=[f(k,j,i),k=1..nep(j,i),j=1..nmu(i),i=1..ne]
!     ilef: secondary energy interpolation law
!           integer array ilef=[ilef(k,j,i),k=1..nep(j,i),j=1..nmu(i),i=1..ne]
!
  subroutine wrtmf6_law7_lab(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nadim,nmu,xmu,ilmu,nepdim,nep,ep,f,ilef)
    implicit real*8 (a-h,o-z)
    character*80 nou(*)
    dimension ey(*),y(*),ily(*),e(*),ile(*),nmu(*),xmu(nadim,*),ilmu(nadim,*)
    dimension nep(nadim,*),ep(nadim,nepdim,*),f(nadim,nepdim,*),ilef(nadim,nepdim,*)
    dimension nbt(nrmax),ibt(nrmax),x1(nepdim),y1(nepdim),jbtep(nepdim),jbtna(nadim)
    if (ny.gt.ne6max) then
      write(*,*)' MF6/law=7 Warning: Too many incident energy points. ny=',ny,' ne6max=',ne6max
    endif
    if (ne.gt.ne6max) then
      write(*,*)' MF6/law=7 Warning: Too many incident energy points. ne=',ne,' ne6max=',ne6max
    endif
    if (nadim.gt.nmu6max) then
      write(*,*)' MF6/law=7 Warning: Too many cosines. nmu=',nadim,' nmu6max=',nmu6max
    endif
    if (nepdim.gt.nep6max) then
      write(*,*)' MF6/law=7 Warning: Too many outgoing energies. nep=',nepdim,' nep6max=',nep6max
    endif
    if (ne*nadim*nepdim.gt.npmax) then
      write(*,*)' MF6/law=7 Error: Too large arrays for zap=',nint(zap),' npmax=',npmax
      write(*,*)' ne=',ne,' na=',nadim,' nep=',nepdim,' ne*na*nep=',ne*nadim*nepdim
      stop
    endif
    call packibt(ny,ily,nr,nbt,ibt)
    mf=6
    law=7
    c1=0.0d0
    c2=0.0d0
    l1=0
    l2=0
    n1=0
    n2=0
    call wrtab1(nou,jou,mat,mf,mt,ns,zap,awp,lip,law,nr,nbt,ibt,ny,ey,y)
    call packibt(ne,ile,nr,nbt,ibt)
    call wrtab2(nou,jou,mat,mf,mt,ns,c1,c2,l1,l2,nr,nbt,ibt,ne)
    do i=1,ne
      nmui=nmu(i)
      do j=1,nmui-1
        jbtna(j)=ilmu(j,i)
      enddo
      call packibt(nmui,jbtna,nr,nbt,ibt)
      call wrtab2(nou,jou,mat,mf,mt,ns,c1,e(i),l1,l2,nr,nbt,ibt,nmui)
      do j=1,nmui
        xmui=xmu(j,i)
        nepi=nep(j,i)
        do k=1,nepi-1
          x1(k)=ep(k,j,i)
          y1(k)=f(k,j,i)
          jbtep(k)=ilef(k,j,i)
        enddo
        x1(nepi)=ep(nepi,j,i)
        y1(nepi)=f(nepi,j,i)
        call packibt(nepi,jbtep,nr,nbt,ibt)
        call wrtab1(nou,jou,mat,mf,mt,ns,c1,xmui,l1,l2,nr,nbt,ibt,nepi,x1,y1)
      enddo
    enddo
    return
  end subroutine wrtmf6_law7_lab
!
! =================================================================================================================================
! subroutine wrtmf6_law7_lab_law: Write one MF6 subsection for law=7 (Laboratory Angle-Energy Distribution)
! =================================================================================================================================
!
!   Similar to wrtmf6_law7_lab, but each interpolation scheme is represented by only one interpolation law.
!     lawy: interpolation law for [E, y(E)]
!     lawf: interpolation law between incident energies
!      lmu: interpolation law between cosines
!      lep: interpolation law between outgoing energies
!   The rest of parameters have the same meaning as before
!
  subroutine wrtmf6_law7_lab_law(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,lawy,ne,e,lawf,nadim,nmu,xmu,lmu,nepdim,nep,ep,f,lep)
    implicit real*8 (a-h,o-z)
    character*80 nou(*)
    dimension ey(*),y(*),e(*),nmu(*),xmu(nadim,*)
    dimension nep(nadim,*),ep(nadim,nepdim,*),f(nadim,nepdim,*)
    dimension ily(ny),ile(ne),ilmu(nadim,ne),ilef(nadim,nepdim,ne)
    do i=1,ny-1
      ily(i)=lawy
    enddo
    do i=1,ne-1
      ile(i)=lawf
    enddo
    do i=1,ne
      nmui=nmu(i)
      do k=1,nmui-1
        ilmu(k,i)=lmu
      enddo
      do k=1,nmui
        nepj=nep(k,i)
        do j=1,nepj-1
          ilef(j,k,i)=lep
        enddo
      enddo
    enddo
    call wrtmf6_law7_lab(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nadim,nmu,xmu,ilmu,nepdim,nep,ep,f,ilef)
    return
  end subroutine wrtmf6_law7_lab_law
!
! =================================================================================================================================
! subroutine wrtmf6_law7_lab_lin
! =================================================================================================================================
!
!   Similar to wrtmf6_law7_lab, but all interpolation schema are lin-lin, therefore they are not input parameters.
!   The rest of parameters have the same meaning as before
!
  subroutine wrtmf6_law7_lab_lin(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ne,e,nadim,nmu,xmu,nepdim,nep,ep,f)
    implicit real*8 (a-h,o-z)
    character*80 nou(*)
    dimension ey(*),y(*),e(*),nmu(*),xmu(nadim,*)
    dimension nep(nadim,*),ep(nadim,nepdim,*),f(nadim,nepdim,*)
    dimension ily(ny),ile(ne),ilmu(nadim,ne),ilef(nadim,nepdim,ne)
    do i=1,ny-1
      ily(i)=2
    enddo
    do i=1,ne-1
      ile(i)=2
    enddo
    do i=1,ne
      nmui=nmu(i)
      do k=1,nmui-1
        ilmu(k,i)=2
      enddo
      do k=1,nmui
        nepj=nep(k,i)
        do j=1,nepj-1
          ilef(j,k,i)=2
        enddo
      enddo
    enddo
    call wrtmf6_law7_lab(nou,jou,mat,mt,ns,zap,awp,lip,ny,ey,y,ily,ne,e,ile,nadim,nmu,xmu,ilmu,nepdim,nep,ep,f,ilef)
    return
  end subroutine wrtmf6_law7_lab_lin
!
! =================================================================================================================================
!
! PROC: Processing routines
!
! =================================================================================================================================
! subroutine renorm_leg(na,a,c,f): renormalize Legendre expansion to a constant value c or return the value of the integral(a0)
! =================================================================================================================================
!
!   na: number of Legendre coefficient (order=na-1)
!    a: Legendre coefficients array
!       real*8 a=[a(k),k=1,na]
!    c: normalization constant, if c<0 or c=0 no normalization applied
!    f: on return if c>0, normalization factor applied to tabulated data
!        otherwise the value of the integral
!
  subroutine renorm_leg(na,a,c,f)
   implicit real*8 (a-h, o-z)
   parameter (zero=0.0d0)
   dimension a(*)
   if (c.gt.zero) then
     a0=a(1)
     if (a0.ne.zero) then
       f=c/a0
       do k=1,na
         a(k)=a(k)*f
       enddo
       if (a0.lt.0) then
         write(*,*)' Warning: negative 0-th order Legendre coefficient found a0=',a0
       endif
     else
       write(*,*)' Warning: null 0-th order Legendre coefficient found a0=',a0,' . Set to ',c
       a(1)=c
       f=c
     endif
   else
     f=a(1)
   endif
   return
  end subroutine renorm_leg
!
! =================================================================================================================================
! subroutine renorm_tab(np,x,y,law,c,f): renormalize tabulated data x,y with interpolation law by interval to a constant
!                                        value c or return the value of the integral.
! =================================================================================================================================
!
!   np: number of tabulated points
!    x: array of x-abscissa
!        real*8 x=[x(i),i=1,np]
!    y: array of y-ordinate
!       real*8 y=[y(i),i=1,np]
!  law: array of interpolation law by interval
!        integer law=[law(i),i=1,np-1]
!    c: normalization constant, if c<0 or c=0 no normalization applied
!    f: on return if c>0, normalization factor applied to tabulated data
!       otherwise the value of the integral
!
!   ilaw=law(i)-(law(i)/10)*10 for interval [x(i+1),x(i)]
!
!           y     x
!   ilaw=1 histogram
!   ilaw=2 lin - lin
!   ilaw=3 lin - log
!   ilaw=4 log - lin
!   ilaw=5 log - log
!
  subroutine renorm_tab(np,x,y,law,c,f)
   implicit real*8 (a-h, o-z)
   parameter (eps=1.0d-6, zero=0.0d0)
   dimension x(*),y(*),law(*)
   sum=0.0d0
   do j=2,np
     j1=j-1
     jlaw=law(j1)
     ilaw=jlaw-(jlaw/10)*10
     x1=x(j1)
     x2=x(j)
     dx=x2-x1
     if (dx.gt.zero) then
       if (ilaw.eq.1) then
         sum=sum+dx*y(j1)
       elseif (ilaw.eq.2) then
         sum=sum+0.5d0*dx*(y(j)+y(j1))
       elseif (ilaw.eq.3) then
         y1=y(j1)
         y2=y(j)
         ax1=abs(x1)
         if (x1*x2.gt.zero.and.dx.gt.eps*ax1) then
           sum=sum+x2*y2-x1*y1-dx*(y2-y1)/log(x2/x1)
         else
           sum=sum+0.5d0*dx*(y1+y2)
         endif
       elseif (ilaw.eq.4) then
         y1=y(j1)
         y2=y(j)
         dy=y2-y1
         ay1=abs(y1)
         ady=abs(dy)
         if (y1*y2.gt.zero.and.ady.gt.eps*ay1) then
           sum=sum+dx*dy/log(y2/y1)
         else
           sum=sum+0.5d0*dx*(y1+y2)
         endif
       elseif (ilaw.eq.5) then
         y1=y(j1)
         y2=y(j)
         dy=y2-y1
         ax1=abs(x1)
         ay1=abs(y1)
         ady=abs(dy)
         x12=x1*x2
         y12=y1*y2
         if (x12.gt.zero.and.dx.gt.eps*ax1.and.y12.gt.zero.and.ady.gt.eps*ay1) then
           d=log(y2/y1)/log(x2/x1)+1.0d0
           sum=sum+y1*x1*(((x2/x1)**d)-1.0d0)/d
         elseif (y12.gt.zero.and.ady.gt.eps*ay1) then
           sum=sum+dx*dy/log(y2/y1)
         elseif (x12.gt.zero.and.dx.gt.eps*ax1) then
           sum=sum+x2*y2-x1*y1-dx*dy/log(x2/x1)
         else
           sum=sum+0.5d0*dx*(y1+y2)
         endif
       else
         write(*,*)' ERROR in renorm: ilaw=',ilaw,' not coded'
         stop
       endif
     endif
   enddo
   if (c.gt.zero) then
     if (sum.ne.zero) then
       f=c/sum
       do j=1,np
         y(j)=y(j)*f
       enddo
     else
       write(*,*)' Warning: Integral equal zero. No normalization applied'
       f=zero
     endif
   else
     f=sum
   endif
   return
  end subroutine renorm_tab
!
! end of module endf6
!
end module endf6
